/*
  ------------------------------------------------------------------------------

    Copyright © 2005-2011, Tommi T. Järvi
                           tommi . jarvi @ iki . fi (remove spaces)


    This file is part of one of my programs.

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2,
    as published by the Free Software Foundation.

    This is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  ------------------------------------------------------------------------------
*/


#include <iostream>
#include <fstream>
#include <cassert>
#include "atomsystem.h"
#include "mathsubs.h"
#include "licence.h"


using namespace std;

/*
  Program for creating nanocrystalline samples using
  the Voronoi-construction. See below for online help.
*/
int main(int argc, char *argv[]) {

  // --- variables

  Atom_System *points;            // grain centers
  Atom_System *grain;             // created grain
  Atom_System *temp;              // temp

  string pfile;                   // grain centers
  string tfile;                   // grain template file
  valarray<double> box(3);        // box
  int seed;                       // rng seed
  double minr;                    // remove atoms so that none are closer to each other than this
  double maxgd;                   // max grain center distance for cutting

  bool rotate_axis = false;       // rotate only around a certain axis
  double phi, theta;              // axis

  bool rotate_y = false;          // rotate only around y


  // --- help

  if(argc==1) {

    cout << "" << endl;
    cout << " Voronoi" << endl;
    cout << "   svn r276 (2013-04-19)" << endl;
    cout << "   (C) 2005-2013 Tommi T. Järvi" << endl;
    cout << "       (mainly, see option 'gpl')" << endl;
    cout << "" << endl;
    cout << " A program for constructing a nanocrystalline system" << endl;
    cout << " using the Voronoi construction. For licence information" << endl;
    cout << " type \"" << argv[0] << " gpl\"." << endl;
    cout << "" << endl;
    cout << "" << endl;
    cout << " Usage: " << argv[0] << " [options] file" << endl;
    cout << "" << endl;
    cout << " The file contains everything needed to make the" << endl;
    cout << " Voronoi construction. The format is almost xyz:" << endl;
    cout << "" << endl;
    cout << "   --------------------" << endl;
    cout << "   N" << endl;
    cout << "   basis box seed minr maxgd" << endl;
    cout << "   P x1 y1 z1" << endl;
    cout << "   P x2 y2 z2" << endl;
    cout << "   ..." << endl;
    cout << "   P xN yN zN" << endl;
    cout << "   --------------------" << endl;
    cout << "" << endl;
    cout << " Options:" << endl;
    cout << "   -r x y z      Rotate randomly only around axis [x,y,z], e.g., \"0 1 0\" for y-axis" << endl;
    cout << "   -ry           Rotate randomly only around the y-axis (allows cutting from disc instead of sphere)"  << endl;
    cout << "" << endl;
    cout << " Parameters:" << endl;
    cout << "   N        number of grains" << endl;
    cout << "   basis    Euler basis file name for creating crystals" << endl;
    cout << "   box      size of created cell (x y z)" << endl;
    cout << "   seed     RNG seed" << endl;
    cout << "   minr     remove atoms closer than minr to each other" << endl;
    cout << "   maxgd    consider grains closer than maxgd when cutting," << endl;
    cout << "            also determines the size of the template crystal" << endl;
    cout << "            from which the grains are cut" << endl;
    cout << "" << endl;
    cout << " For the crystal basis, Euler basis mode 1 is assumed, as well" << endl;
    cout << " as a scaling factor of 1.0." << endl;
    cout << "" << endl;
    cout << " Notes:" << endl;
    cout << "   - For systems with too few grains in one box direction," << endl;
    cout << "     artificial boundaries may appear." << endl;
    cout << "   - For too small maxgd, either overlapping grains or voids" << endl;
    cout << "     may appear." << endl;
    cout << "" << endl;

    exit(0);
  }


  // --- licence
  if(argc==2) {
    pfile = argv[1];
    if(pfile=="gpl") {
      cout << "************************" << endl;
      cout << " This program is licenced under the GNU General Public Licence, version 2." << endl;
      cout << "The complete licence is shown below. All files include a copyright" << endl;
      cout << "notice and all but the files associated with the Mersenne Twister RNG" << endl;
      cout << "are (C) Tommi Järvi except if stated otherwise. The licence for" << endl;
      cout << "Mersenne Twister is shown in the respective files (mt19937ar.*) and" << endl;
      cout << "below after the GNU GPL." << endl;
      cout << "************************" << endl;
      Licence::gpl();
      cout << "************************" << endl;
      Licence::mt();
      cout << "************************" << endl;
      exit(0);
    }
  }


  // --- read command line options
  {
    string dummy;

    for(int i=1; i<argc; i++) {
      string d = argv[i];

      // rotation along one axis only
      if(d=="-r") {
	if(argc-1 < i+3) {
	  cerr << "Not enough parameters for -r" << endl;
	  exit(-1);
	}
	double x = atof(argv[i+1]);
	double y = atof(argv[i+2]);
	double z = atof(argv[i+3]);
	i = i + 4;

	rotate_axis = true;

	double r = sqrt(x*x + y*y + z*z);
	theta = acos(z/r);
	phi = atan(y/x);

	cout << "***" << endl;
	cout << "*** Only rotating around axis [" << x << ", " << y << ", " << z << "] corresp. to (phi, theta) = (" << phi << ", " << theta << ")." << endl;
	cout << "***" << endl;
      }

      // rotation around y only
      else if(d=="-ry") {
	if(argc-1 < i+0) {
	  cerr << "Not enough parameters for -ry" << endl;
	  exit(-1);
	}
	i = i + 1;

	rotate_y = true;

	theta = M_PI/2.0;
	phi = M_PI/2.0;

	cout << "***" << endl;
	cout << "*** Only rotating around axis [" << 0 << ", " << 1 << ", " << 0 << "] corresp. to (phi, theta) = (" << phi << ", " << theta << ")." << endl;
	cout << "***" << endl;
      }
    }  // end of loop over command line arguments

    // checks
    if(rotate_y && rotate_axis) {
      cerr << "Error: Rotating both around y and around some other axis." << endl;
      exit(-1);
    }
  }


  // --- read input parameters

  {
    ifstream f;
    string dummy;

    pfile = argv[argc-1];

    cout << "***" << endl;
    cout << "*** Reading input parameters from file " << pfile << endl;
    cout << "***" << endl;

    f.open(pfile.c_str());
    assert(f);

    // skip first line
    getline(f, dummy, '\n');

    // read parameters
    f >> tfile;
    f >> box[0];
    f >> box[1];
    f >> box[2];
    f >> seed;
    f >> minr;
    f >> maxgd;

    cout << "***  Crystal basis: " << tfile << endl;
    cout << "***  Box:           " << box[0] << " " << box[1] << " " << box[2] << endl;
    cout << "***  RNG seed:      " << seed << endl;
    cout << "***  minr:          " << minr << endl;
    cout << "***  maxgd:         " << maxgd << endl;
    cout << "***" << endl;

    f.close();
  }


  // --- check

  // maxgd
  if(rotate_y) {
    if((maxgd > box[0]/2 || maxgd > box[2]/2)) {
      cerr << "Error: maxgd larger than half of box size so atoms would wrap on themselves over periodics. Decrease it." << endl;
      exit(-1);
    }
  }
  else {
    if(maxgd > box.min()/2) {
      if(rotate_axis) {
	cerr << "Warning: maxgd larger than half of smallest box size so atoms" << endl;
	cerr << "         possibly wrap on themselves over periodics, depending" << endl;
	cerr << "         on the way you rotate." << endl;
      }
      else {
	cerr << "Error: maxgd larger than half of smallest box size so atoms would wrap on themselves over periodics. Decrease it." << endl;
	exit(-1);
      }
    }
  }


  // --- init

  sgenrand(seed);


  // --- create template grain

  cout << "***" << endl;
  cout << "*** Creating template grain..." << endl;
  cout << "***" << endl;

  // get number of cells to be created by checking unit cell size
  {
    // number of cells
    valarray<int> cells(3);
    cells[0] = cells[1] = cells[2] = 1;

    temp = new Atom_System();
    temp->read_unit_cell_and_init(tfile, cells, 1.0, 1, 1);

    if(rotate_y) {
      // size in x, z
      double d = temp->box[0];
      if(temp->box[2] > d) d = temp->box[2];

      cells[0] = 2*int(floor(maxgd/d)) + 1;
      cells[2] = cells[0];

      // for y, adjust box
      {
	int i1, i2;
	i1 = floor(box[1]/temp->box[1]);
	i2 = ceil(box[1]/temp->box[1]);
	if(abs(box[1] - temp->box[1]*i1) < abs(box[1] - temp->box[1]*i2)) {
	  cells[1] = i1;
	}
	else {
	  cells[1] = i2;
	}
      }
      cout << "***" << endl;
      cout << "***   Adjusting box y dimension from " << box[1];
      box[1] = temp->box[1]*cells[1];
      cout << " to " << box[1] << endl;
      cout << "***" << endl;

      cout << "***" << endl;
      cout << "***   Creating with ncell = " << cells[0] << " " << cells[1] << " " << cells[2] << endl;
      cout << "***" << endl;
    }
    else {
      // size
      for(int i=0;i<3;i++) {
	cells[i] = 2*int(floor(maxgd/temp->box[i])) + 1;
      }

      cout << "***" << endl;
      cout << "***   Creating with ncell = " << cells[0] << " " << cells[1] << " " << cells[2] << endl;
      cout << "***" << endl;
    }
    delete temp;

    // create template crystal
    temp = new Atom_System();
    temp->read_unit_cell_and_init(tfile, cells, 1.0, 1, 1);
    {
      // periodics off
      valarray<int> pbc(3);           // pbc's
      pbc[0] = pbc[1] = pbc[2] = 0;
      temp->setpbc(pbc);
    }
    temp->center();  // center on origin

    // make spherical
    {
      valarray<double> c(3);
      c[0] = c[1] = c[2] = 0.0;

      cout << "***" << endl;
      cout << "***   Making spherical and reporting contents..." << endl;
      cout << "***" << endl;

      for(int i=0;i<temp->atoms();i++) {
	if(temp->dr(i, c) > maxgd) {
	  temp->removeat(i);
	  i--;
	}
      }
    }

    temp->report();
  }


  // --- read grain coordinates

  cout << "***" << endl;
  cout << "*** Reading grain data..." << endl;
  cout << "***" << endl;

  points = new Atom_System();
  {
    // periodics on
    valarray<int> pbc(3);           // pbc's
    pbc[0] = pbc[1] = pbc[2] = 1;
    points->setpbc(pbc);
  }
  points->setbox(box);
  points->readxyz(pfile, 0, 1);

  // neighbour list for points
  points->init_nlist(maxgd, maxgd+0.001);
  points->update_symmnlist();


  // --- create grains

  cout << "***" << endl;
  cout << "*** Creating grains..." << endl;
  cout << "***" << endl;

  // loop over grains
  for(int i=0; i<points->atoms(); i++) {
    cout << "*** Grain " << i << endl;

    // grain center
    valarray<double> rg = points->ri(i);

    // create grain
    grain = new Atom_System(*temp);

    // rotate randomly
    if(rotate_axis || rotate_y) {
      // random rotation around given axis
      double alpha = 2.0*M_PI*genrand();
      grain->rotate_axis(phi, theta, alpha);
    }
    else {
      // random euler rotation
      double alpha = 2.0*M_PI*genrand();
      double beta = acos(1.0 - 2.0*genrand());
      double gamma = 2.0*M_PI*genrand();
      grain->rotate_euler(alpha, beta, gamma);
    }

    // translate to where center should be
    grain->translate(rg);

    // remove atoms that will not be needed in any case
    for(int j=0; j<3; j++) {
      grain->remove(j, 0, -box[j]);
      grain->remove(j, 1, box[j]);
    }

    // wrap into periodics
    {
      // periodics on
      valarray<int> pbc(3);           // pbc's
      pbc[0] = pbc[1] = pbc[2] = 1;
      grain->setpbc(pbc);
    }
    grain->setbox(box);
    {
      // wrap to within periodic boundaries
      valarray<double> tt(3);
      tt[0] = tt[1] = tt[2] = 0.0;
      grain->translate(tt);
    }

    // remove atoms closer to other grain centers
    // -> loop over neighbouring grains
    int irem = 0;  // removed atoms
    for(int j=1; j<=points->symmnlist[i][0];j++) {
      int g2 = points->symmnlist[i][j];  // index of other grain
      valarray<double> rg2 = points->ri(g2);  // center of other grain

      // loop over atoms
      for(int k=0;k<grain->atoms();k++) {
	double r1 = grain->dr(k, rg);   // distance from grain
	double r2 = grain->dr(k, rg2);  // distance from other grain
	if(r2 < r1) {
	  grain->removeat(k);
	  k--;
	  irem++;
	}
      }
    }
    cout << "***   Removed " << irem << " atoms because of other grains." << endl;

    // save to file
    {
      string file = "grain_" + stringify(i) + ".xyz";
      grain->savexyz(file, 0, true);
    }

    // clean
    delete grain;
  }  // end of loop over grains

  delete temp;  // delete template grain


  // --- remove atoms too close to each other

  cout << "***" << endl;
  cout << "*** Removing atoms too close to each other..." << endl;
  cout << "***   For each pair, removing from first grain." << endl;
  cout << "***   Overwriting the original grain file." << endl;
  cout << "***" << endl;

  // loop over grains
  for(int i=0; i<points->atoms(); i++) {
    // removed atoms
    int irem = 0;

    // load grain
    {
      // periodics on
      valarray<int> pbc(3);           // pbc's
      pbc[0] = pbc[1] = pbc[2] = 1;

      string file = "grain_" + stringify(i) + ".xyz";

      grain = new Atom_System();
      grain->setpbc(pbc);
      grain->setbox(box);
      grain->readxyz(file, 0, 1);
    }

    // loop over nearby grains
    for(int j=1; j<=points->symmnlist[i][0];j++) {
      int g2 = points->symmnlist[i][j];       // index of other grain

      if(i >= g2) continue;

      cout << "*** Grains " << i << " " << g2 << endl;

      // load other grain
      {
	// periodics on
	valarray<int> pbc(3);           // pbc's
	pbc[0] = pbc[1] = pbc[2] = 1;

	string file = "grain_" + stringify(g2) + ".xyz";

	temp = new Atom_System();
	temp->setpbc(pbc);
	temp->setbox(box);
	temp->readxyz(file, 0, 1);
	temp->init_nlist(minr, minr+0.001);
      }

      // loop over atoms in grain
      for(int k=0;k<grain->atoms();k++) {
	// find closest atom in other grain
	double ra;
	int ia = temp->find_closest_atom(grain->ri(k), ra);

	// remove if needed
	if(ia != -1 && ra < minr) {
	  grain->removeat(k);
	  k--;
	  irem++;
	}
      }  // end of loop over atoms in grain

      delete temp;

    }  // end of loop over other grains

    cout << "***   Removed " << irem << " atoms because too close to atom in another grain." << endl;

    {
      string file = "grain_" + stringify(i) + ".xyz";
      grain->savexyz(file, 0, true);
    }

    delete grain;

  }  // end of loop over grains
}
