/*
  ------------------------------------------------------------------------------

    Copyright © 2005-2013, Tommi T. Järvi
                           firstname . jarvi @ iki . fi (remove spaces)


    This file is part of one of my programs.

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2,
    as published by the Free Software Foundation.

    This is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  ------------------------------------------------------------------------------
*/


#include "atomsystem.h"


// --- destructor


void Atom_System::delete_arrays() {

  /*
    Used to delete all arrays in Atom_System. Always use this.
    Here you can also see the size of each array and which variable
    stores it.
  */


  // --- arrays of size dim
  if(dim != 0) {
    delete [] pbc;
    delete [] box;
    delete [] lsubcell;
    delete [] nsubcell;
  }
  dim = 0;


  // --- arrays of size x0size
  if(x0size != 0) {
    delete [] x0;
    delete [] dx0;
    delete [] x1;
    delete [] F;
    delete [] fixx0;
  }
  x0size = 0;


  // --- nlist
  if(nlistsize != 0) delete [] nlist;
  nlistsize = 0;
  if(cells > 0) delete [] linkedlist_head;
  cells = 0;


  // --- arrays of size atypesize
  if(atypesize != 0) {
    delete [] V;
    delete [] K;
    delete [] linkedlist_list;
    delete [] atype;
    delete [] block;
  }
  atypesize = 0;
}


Atom_System::~Atom_System() {

  delete_arrays();
}


// --- constructors


Atom_System::Atom_System() {

  /*
    Default constructor
  */


  // --- variables
  int i;  // loops


#if defined(DEBUG)
  cout << "Atom_System::def. constr.: AS...";
#endif


  // --- set basic defaults
  dim = 3;


  // --- default to empty system
  natoms = 0;
  x0size = 0;
  atypesize = 0;
  nlistsize = 0;
  cells = 0;
  box = new double[dim];
  pbc = new int[dim];
  lsubcell = new double[dim];
  nsubcell = new int[dim];
  for(i=0;i<dim;i++) {
    box[i] = 0;
    pbc[i] = 0;
  }


  // --- set pointers to null to make valgrind happy
  x0 = NULL;
  x1 = NULL;
  F = NULL;
  fixx0 = NULL;
  atype = NULL;
  block = NULL;
  V = NULL;
  K = NULL;
  dx0 = NULL;
  nlist = NULL;
  linkedlist_head = NULL;
  linkedlist_list = NULL;

  comment = "";
  maxneighbours = 0;
  T = 0;
  a = 0;
  nlist_cut = 0;
  nlist_potcut = 0;

#if defined(DEBUG)
  cout << " ...done(AS)." << endl;
#endif
}


Atom_System::Atom_System(const Atom_System &syst) {

  /*
    Copy constructor
  */


  // --- variables
  int i;


#if defined(DEBUG)
  cout << "Atom_System::copy constructor..." << endl;
#endif


  // --- copy stuff
  dim = syst.dim;

  natoms = syst.natoms;
  x0size = syst.x0size;
  atypesize = syst.atypesize;
  nlistsize = syst.nlistsize;
  maxneighbours = syst.maxneighbours;
  cells = syst.cells;

  comment = syst.comment;
  a = syst.a;
  T = syst.T;

  nlist_cut = syst.nlist_cut;
  nlist_potcut = syst.nlist_potcut;

  atypes = syst.atypes;


  // --- arrays

  // arrays of size x0size

#if defined(DEBUG)
  cout << " (p:x0 x0size " << x0 << " " << x0size << ") ";
#endif

  if(x0size != 0) {
    x0 = new double[x0size];
    dx0 = new double[x0size];
    x1 = new double[x0size];
    F = new double[x0size];
    fixx0 = new int[x0size];
  }

  for(i=0;i<x0size;i++) {
    x0[i] = syst.x0[i];
    dx0[i] = syst.dx0[i];
    x1[i] = syst.x1[i];
    F[i] = syst.F[i];
    fixx0[i] = syst.fixx0[i];
  }

  // arrays of size dim

#if defined(DEBUG)
  cout << " (p:box dim " << box << " " << dim << ") ";
#endif

  if(dim != 0) {
    pbc = new int[dim];
    box = new double[dim];
    lsubcell = new double[dim];
    nsubcell = new int[dim];
  }

  for(i=0;i<dim;i++) {
    pbc[i] = syst.pbc[i];
    box[i] = syst.box[i];
    lsubcell[i] = syst.lsubcell[i];
    nsubcell[i] = syst.nsubcell[i];
  }

  // neighbour list arrays

#if defined(DEBUG)
  cout << " (p:nlist nlistsize " << nlist << " " << nlistsize << ") ";
#endif

  if(nlistsize != 0) {
    nlist = new int[nlistsize];
  }

  for(i=0;i<nlistsize;i++) {
    nlist[i] = syst.nlist[i];
  }

#if defined(DEBUG)
  cout << " (p:llhead cells " << linkedlist_head << " " << cells << ") ";
#endif

  if(cells != 0) {
    linkedlist_head = new int[cells];
  }

  for(i=0;i<cells;i++) {
    linkedlist_head[i] = syst.linkedlist_head[i];
  }

  // arrays of size atypesize

#if defined(DEBUG)
  cout << " (p:atype atypesize " << atype << " " << atypesize << ") ";
#endif
  
  if(atypesize != 0) {
    V = new double[atypesize];
    K = new double[atypesize];
    linkedlist_list = new int[atypesize];
    atype = new int[atypesize];
    block = new int[atypesize];
  }

  for(i=0;i<atypesize;i++) {
    V[i] = syst.V[i];
    K[i] = syst.K[i];
    linkedlist_list[i] = syst.linkedlist_list[i];
    atype[i] = syst.atype[i];
    block[i] = syst.block[i];
  }

  // blocknat

  if(syst.blocknat.size() >= 1) {
    blocknat.resize(syst.blocknat.size());
    blocknat = syst.blocknat;
  }

#if defined(DEBUG)
  cout << "  ...done." << endl;
#endif
}


// --- operators


Atom_System& Atom_System::operator=(const Atom_System &syst) {

  /*
    Operator =
  */


  // --- variables
  int i;  // loops


  // --- =

  // don't copy self
  if(this == &syst) return *this;

  // reset arrays
  delete_arrays();

  dim = syst.dim;

  natoms = syst.natoms;
  x0size = syst.x0size;
  atypesize = syst.atypesize;
  nlistsize = syst.nlistsize;
  maxneighbours = syst.maxneighbours;
  cells = syst.cells;

  comment = syst.comment;
  a = syst.a;
  T = syst.T;

  nlist_cut = syst.nlist_cut;
  nlist_potcut = syst.nlist_potcut;

#if defined(DEBUG)
  cout << "Atom_System::Operator=..." << endl;
  cout << "this: x0size, x0 dx0 x1 F fixx0 " << x0size << " " << x0 << " " << dx0 << " " << x1 << " " << F << " " << fixx0 << endl;
  cout << "syst: x0size, x0 dx0 x1 F fixx0 " << syst.x0size << " " << syst.x0 << " " << syst.dx0 << " " << syst.x1 << " " << syst.F << " " << syst.fixx0 << endl;
#endif

  atypes = syst.atypes;

  // arrays

  // arrays of size x0size

  if(x0size != 0) {
    x0 = new double[x0size];
    dx0 = new double[x0size];
    x1 = new double[x0size];
    F = new double[x0size];
    fixx0 = new int[x0size];
  }

#if defined(DEBUG)
  cout << " (x0 x0size " << x0 << " " << x0size << ")...";
#endif

  for(i=0;i<x0size;i++) {
    x0[i] = syst.x0[i];
    dx0[i] = syst.dx0[i];
    x1[i] = syst.x1[i];
    F[i] = syst.F[i];
    fixx0[i] = syst.fixx0[i];
  }

  // arrays of size dim

  if(dim != 0) {
    pbc = new int[dim];
    box = new double[dim];
    lsubcell = new double[dim];
    nsubcell = new int[dim];
  }

  for(i=0;i<dim;i++) {
    pbc[i] = syst.pbc[i];
    box[i] = syst.box[i];
    lsubcell[i] = syst.lsubcell[i];
    nsubcell[i] = syst.nsubcell[i];
  }

  // neighbour list arrays

  if(nlistsize != 0) {
    nlist = new int[nlistsize];
  }

  for(i=0;i<nlistsize;i++) {
    nlist[i] = syst.nlist[i];
  }

  if(cells != 0) {
    linkedlist_head = new int[cells];
  }

  for(i=0;i<cells;i++) {
    linkedlist_head[i] = syst.linkedlist_head[i];
  }

  // arrays of size atypesize

  if(atypesize != 0) {
    V = new double[atypesize];
    K = new double[atypesize];
    linkedlist_list = new int[atypesize];
    atype = new int[atypesize];
    block = new int[atypesize];
  }

  for(i=0;i<atypesize;i++) {
    V[i] = syst.V[i];
    K[i] = syst.K[i];
    linkedlist_list[i] = syst.linkedlist_list[i];
    atype[i] = syst.atype[i];
    block[i] = syst.block[i];
  }

  // blocknat

  if(syst.blocknat.size() >= 1) {
    blocknat.resize(syst.blocknat.size());
    blocknat = syst.blocknat;
  }

#if defined(DEBUG)
  cout << " ...done(AS=)." << endl;
#endif

  return *this;
}


// --- setting stuff


void Atom_System::zero_arrays() {

  /*
    Zero position, velocity etc. (Not box, atypes etc.)
  */


  // --- variables
  int i;  // loops


  // --- zero

  // arrays of size x0size
  for(i=0;i<x0size;i++) {
    x0[i] = 0.0;
    dx0[i] = 0.0;
    x1[i] = 0.0;
    F[i] = 0.0;
    fixx0[i] = 0;
  }

  // arrays of size atypesize
  for(i=0;i<atypesize;i++) {
    V[i] = 0.0;
    K[i] = 0.0;
    //linkedlist_list[i] = 0;
    //atype[i] = 0;
    block[i] = 0;
  }
}


void Atom_System::set_lattice_constant(double value) {
  a = value;
}


void Atom_System::setcomment(string com) {
  comment = com;
}


void Atom_System::setpbc(valarray<int> cond) {

  int i;

  if(cond.size() != dim) {
    cerr << "Atom_System::setpbc(): Array size mismatch." << endl;
    exit(-1);
  }

  for(i=0;i<dim;i++) {
    pbc[i] = cond[i];
  }
}


void Atom_System::setbox(valarray<double> size) {

  int i;

  if(size.size() != dim) {
    cerr << "Atom_System::setbox(): Array size mismatch." << endl;
    exit(-1);
  }

  for(i=0;i<dim;i++) {
    box[i] = size[i];
  }
}


// --- getting stuff


int Atom_System::atoms() const {
  return natoms;
}


int Atom_System::dimension() const {
  return dim;
}


void Atom_System::init_with_ucell(int crystal, const valarray<int> &size, valarray<int> atom_types, vector<string> atom_names, int offsetmode, int mode) {

  /*
    Init system with a unit cell so that lattice constant
    same in all directions.

    Take this.a as lattice constant.

    See other function for details.
  */


  // --- variables
  int i;  // loops
  valarray<double> alat(dim);  // lattice constants


  // --- create system
  for(i=0;i<dim;i++) {
    alat[i] = a;
  }

  init_with_ucell(crystal, size, alat, atom_types, atom_names, offsetmode, mode);
}


void Atom_System::read_unit_cell_and_init(string filename, valarray<int> size, double scalef, int vecmode, int readmode) {

  /*
    Reads the basis from an xyz-like file with the following format.

    nbasis
    basis vectors
    first atom
    .
    .
    .
    last atom

    For example for Mo (bcc) with lattice constant 3.15 Å.

    2
    3.15 0.0 0.0 0.0 3.15 0.0 0.0 0.0 3.15
    Mo 0.0 0.0 0.0 1
    Mo 0.5 0.5 0.5 1


    parameters
    ----------
    scalef                  scale vectors by this factor (in mode 3, also coordinates scaled)
    vecmode                 1 - read unit cell vectors
                            2 - monoclinic, read a, b, c, alpha, beta, gamma
			    3 - same as one but coordinates not relative to basis vectors (that is, in Å, for example)
    readmode                1 - normal
                            2 - three additional columns: the velocities

    Note! All angle inputs are in degrees!


    unit cells in detail (scaling factor omitted)
    --------------------
    monoclinic
      v1 = a*X
      v2 = b*( X cos(gamma) + Y sin(gamma) )
      v3 = c*( X cos(beta)  + Y cos(alpha) + Z sin(beta) sin(alpha) )
   */


  // --- variables
  int i, j;
  double alpha, beta, gamma, a, b, c;  // helps for reading in unit cells
  Atom_System basis;
  valarray<double> basis_vectors;
  valarray<double> basis_coordinates;
  valarray<double> basis_velocities;
  valarray<int> basis_atypes;
  vector<string> basis_anames;
  string temp;
  int mode;  // mode for unit cell init function


  // --- read basis
  {
    int rm;
    if(readmode==1) rm = 1;
    else if(readmode==2) rm = 4;
    else {
      cerr << "Atom_System::read_unit_cell_and_init(): Unknown read mode " << readmode << endl;
      exit(-1);
    }
    basis.readxyz(filename, rm, 1);
  }
  stringstream str(basis.comment);


  // --- make arrays
  basis_vectors.resize(basis.dimension()*basis.dimension());
  basis_coordinates.resize(basis.atoms()*basis.dimension());
  basis_velocities.resize(basis.atoms()*basis.dimension());
  basis_atypes.resize(basis.atoms());
  basis_anames.resize(basis.atoms());

  // --- atom data
  for(i=0;i<basis.atoms();i++) {
    basis_atypes[i] = basis.atype[i];
    basis_anames[i] = basis.atypes.name(basis.atype[i]);
  }
  for(i=0;i<basis.atoms()*basis.dimension();i++) {
    basis_coordinates[i] = basis.x0[i];

    // velocities, or for read mode without velocities, set to zero
    if(readmode==2) basis_velocities[i] = basis.x1[i];
    else basis_velocities[i] = 0.0;
  }

  if(vecmode==2 && basis.dimension()!=3) {
    cout << "Atom_System::read_unit_cell_and_init(): Error: vector mode 2 requires dimension 3, now " << dim << endl;
    exit(-1);
  }

  // unit cell vectors
  if(vecmode==1 || vecmode==3) {
    cout << "Atom_System::read_unit_cell_and_init(): Reading unit cell vectors directly as components." << endl;
    for(i=0;i<basis.dimension()*basis.dimension();i++) {
      str >> temp;
      basis_vectors[i] = atof(temp.c_str());
    }
  }
  else if(vecmode==2) {
    cout << "Atom_System::read_unit_cell_and_init(): Reading in unit cell as a, b, c, <(a,b), <(b,c), <(a,c)." << endl;

    str >> temp;
    a = atof(temp.c_str());
    str >> temp;
    b = atof(temp.c_str());
    str >> temp;
    c = atof(temp.c_str());
    str >> temp;
    alpha = atof(temp.c_str())*M_PI/180;
    str >> temp;
    beta = atof(temp.c_str())*M_PI/180;
    str >> temp;
    gamma = atof(temp.c_str())*M_PI/180;

    cout << "a:     " << a << endl;
    cout << "b:     " << b << endl;
    cout << "c:     " << c << endl;
    cout << "alpha: " << alpha*180/M_PI << "   <(b,c)" << endl;
    cout << "beta:  " << beta*180/M_PI << "   <(a,c)" << endl;
    cout << "gamma: " << gamma*180/M_PI << "   <(a,b)" << endl;

    basis_vectors[0] = a;
    basis_vectors[1] = 0;
    basis_vectors[2] = 0;

    basis_vectors[3] = b*cos(gamma);
    basis_vectors[4] = b*sin(gamma);
    basis_vectors[5] = 0;

    basis_vectors[6] = c*cos(beta);
    basis_vectors[7] = c*cos(alpha);
    basis_vectors[8] = c*sin(beta)*sin(alpha);
  }
  else {
    cerr << "Atom_System::read_unit_cell_and_init(): Error: Unknown readmode for unit cell vectors: " << vecmode << endl;
    exit(-1);
  }

  // apply scaling factor
  cout << "Atom_System::read_unit_cell_and_init(): Applying scaling factor " << scalef << endl;
  for(i=0;i<basis.dimension()*basis.dimension();i++) {
    basis_vectors[i] *= scalef;
  }
  if(vecmode==3) {
    for(i=0;i<basis.atoms()*basis.dimension();i++) {
	basis_coordinates[i] *= scalef;
    }
  }

  // output unit cell vectors
  cout << "Atom_System::read_unit_cell_and_init(): Unit cell vectors:" << endl;
  for(i=0;i<basis.dimension();i++) {
    for(j=0;j<basis.dimension();j++) {
      cout << basis_vectors[basis.dimension()*i + j] << " ";
    }
    cout << endl;
  }

  // set atom coordinate mode
  if(vecmode==3) mode = 2;
  else mode = 1;

  // make system
  init_unit_cell(size, basis_vectors, basis_coordinates, basis_velocities, basis_atypes, basis_anames, mode);
}


void Atom_System::init_unit_cell(const valarray<int> &size, const valarray<double> &basis_vectors, const valarray<double> &basis_coordinates, const valarray<double> &basis_velocities, valarray<int> basis_atypes, vector<string> basis_anames, int mode) {

  /*
    Create a system using an arbitrary unit cell. Since other than
    rectangular pbc's aren't supported elsewhere in Atom_System yet,
    the elements in the box-vector are set as the lengths of the
    unit cell vectors.

    parameter            size             explanation
    -------------------------------------------------
    size                 dim              number of cells to relicate in each direction
    basis_vectors        dim*dim          unit cell vectors [x0, y0, z0, x1, ...]
    basis_coordinates    dim*N            coordinates [x0, y0, z0, x1, ...] with respect to the unit cell vectors
    basis_velocities     dim*N            velocities  [x0, y0, z0, x1, ...] with respect to the unit cell vectors (not tested much!)
    basis_atypes         N                atom types (integers)
    basis_anames         N                atom types (names, e.g. "Au")
    mode                 integer          1 - normal, 2 - coordinates not with respect to the unit cell vectors
  */


  // --- variables
  int i, j, n;  // loops
  valarray<int> ind(dim);  // loops
  int nbasis;  // number of atoms in basis
  int ncreated;  // created atoms
  valarray<double> offset;  // offsets


  // --- checks
  nbasis = basis_atypes.size();
  if(size.size() != dim || basis_vectors.size() != dim*dim || basis_coordinates.size() != dim*nbasis || basis_atypes.size() != nbasis || basis_anames.size() != nbasis || nbasis==0) {
    cerr << "Atom_System::init_unit_cell(): Something wrong with input vector sizes. Printing vector, size" << endl;
    cerr << "size: " << size.size() << endl;
    cerr << "basis_vectors: " << basis_vectors.size() << endl;
    cerr << "basis_coordinates: " << basis_coordinates.size() << endl;
    cerr << "basis_atypes: " << basis_atypes.size() << endl;
    cerr << "basis_anames: " << basis_anames.size() << endl;
    exit(-1);
  }


  // --- report
  cout << "Atom_System::init_unit_cell(): Initializing with " << nbasis << " atoms in the basis." << endl;


  // --- prepare

  // allocate arrays
  natoms = nbasis;
  for(i=0;i<dim;i++) {
    natoms *= size[i];
  }
  reset(natoms);

  // offsets
  // before allowing non-zero offset, debug it properly!
  offset.resize(dim);
  for(i=0;i<dim;i++) {
    offset[i] = 0.0;
  }

  // add atom types
  atypes.reset();
  for(i=0;i<nbasis;i++) {
    atypes.add(basis_anames[i], basis_atypes[i]);
  }

  // set box size
  for(i=0;i<dim;i++) {
    box[i] = 0;
    for(j=0;j<dim;j++) {
      box[i] += basis_vectors[3*i + j]*basis_vectors[3*i + j];
    }
    box[i] = sqrt(box[i])*size[i];
  }


  // --- create lattice
  ncreated = 0;

  // loop over unit cells
  for(ind[0]=0;ind[0]<size[0];ind[0]++) {
    for(ind[1]=0;ind[1]<size[1];ind[1]++) {
      for(ind[2]=0;ind[2]<size[2];ind[2]++) {

	// loop over atoms in unit cell
        for(n=0;n<nbasis;n++) {

	  // check
	  if(ncreated >= natoms) {
	    cerr << "Atom_System::init_unit_cell(): Error / Bug: Creating too many atoms!" << endl;
	    exit(-1);
	  }

	  // --- zero coordinates
	  for(j=0;j<dim;j++) {
	    x0[dim*ncreated + j] = 0.0;
	    x1[dim*ncreated + j] = 0.0;
	  }

	  // --- add offset, cell position

	  // loop over coordinates
	  for(j=0;j<dim;j++) {
	    // loop over unit cell vectors
	    for(i=0;i<dim;i++) {
	      x0[dim*ncreated + j] += (ind[i] + offset[j])*basis_vectors[dim*i + j];
	    }  // end of loop over coordinates
	  }  // end of loop over unit cell vectors

	  // --- add atom position

	  if(mode==1) {
	    // loop over coordinates
	    for(j=0;j<dim;j++) {
	      // loop over unit cell vectors
	      for(i=0;i<dim;i++) {
		x0[dim*ncreated + j] += basis_coordinates[dim*n + i]*basis_vectors[dim*i + j];
		x1[dim*ncreated + j] += basis_velocities[dim*n + i]*basis_vectors[dim*i + j];
	      }  // end of loop over coordinates
	    }  // end of loop over unit cell vectors
	  }
	  else if(mode==2) {
	    // loop over atom coordinates
	    for(j=0;j<dim;j++) {
	      x0[dim*ncreated + j] += basis_coordinates[dim*n + j];
	      x1[dim*ncreated + j] += basis_velocities[dim*n + j];
	    }  // end of loop over atom coordinates
	  }
	  else {
	    cerr << "Atom_System::init_unit_cell(): Unknown mode " << mode << endl;
	    exit(-1);
	  }

          // set atom type
	  atype[ncreated] = basis_atypes[n];
	  atypes.natype[atype[ncreated]]++;

	  ncreated++;
        }
      }
    }
  }

  if(ncreated != natoms) {
    cerr << "Atom_System::init_unit_cell(): --- Error with number of atoms." << endl;
    exit(-1);
  }

}


void Atom_System::init_with_ucell(int crystal, const valarray<int> &size, const valarray<double> &alat, valarray<int> atom_types, vector<string> atom_names, int offsetmode, int mode) {

  /*
    Init system with a unit cell.

    Sets:
    - box size

    Parameters:
    crystal       crystal type
    size          number of unit cells
    alat          lattice constant in each direction
    atom_types
    atom_names
    offsetmode
    mode


    crystal types
    -------------
    1            (100) FCC
    2            (111) FCC
    3            (111) HCP
    4            (100) BCC
    5            (100) SC
    6            (100) DIA
    7            A16 (alpha-S)

    offset
    ------
    0            0, 0, 0
    1            default

    mode
    ----
    0            box: [-x/2,x/2] (default)
    1            box: [0,x]
  */


  // --- variables

  int i, j, k, n;  // loops
  valarray<int> ind(dim);  // loops
  int nucell;  // number of atoms in unit cell
  double *x0ucell;  // positions of the atoms
  double *offset;  // offset of the unit cell
  int ncreated;  // number of atoms created

  // unit cell database
  vector< valarray<double> > cells;  // unit cell database
  int ncells;  // number of unit cell types
  vector< valarray<double> > offsets;  // default offsets


  // --- definitions of different unit cells
  ncells = 7;    // number of unit cells
  cells.resize(ncells + 1);
  offsets.resize(ncells + 1);
  for(i=0;i<ncells+1;i++) {
    offsets[i].resize(3);
  }


  // fcc (100)
  cells[1].resize(4*3);

  cells[1][0] = 0.0;
  cells[1][1] = 0.0;
  cells[1][2] = 0.0;

  cells[1][3] = 0.5;
  cells[1][4] = 0.5;
  cells[1][5] = 0.0;

  cells[1][6] = 0.0;
  cells[1][7] = 0.5;
  cells[1][8] = 0.5;

  cells[1][9] = 0.5;
  cells[1][10] = 0.0;
  cells[1][11] = 0.5;

  offsets[1][0] = 0.25;
  offsets[1][1] = 0.25;
  offsets[1][2] = 0.25;


  // fcc (111)
  cells[2].resize(6*3);

  cells[2][0] = 0.0;
  cells[2][1] = 0.0;
  cells[2][2] = 0.0;

  cells[2][3] = 0.5;
  cells[2][4] = 0.5;
  cells[2][5] = 0.0;

  cells[2][6] = 0.0;
  cells[2][7] = 0.3333333333333;
  cells[2][8] = 0.3333333333333;

  cells[2][9] = 0.5;
  cells[2][10] = 0.8333333333333;
  cells[2][11] = 0.3333333333333;

  cells[2][12] = 0.5;
  cells[2][13] = 0.1666666666667;
  cells[2][14] = 0.6666666666667;

  cells[2][15] = 0.0;
  cells[2][16] = 0.6666666666667;
  cells[2][17] = 0.6666666666667;

  offsets[2][0] = 0.25;
  offsets[2][1] = 0.083333333333;
  offsets[2][2] = 0.166666666667;

  // hcp (111)
  cells[3].resize(4*3);

  cells[ 3 ][ 0 ] = 0.0;
  cells[ 3 ][ 1 ] = 0.0;
  cells[ 3 ][ 2 ] = 0.0;

  cells[ 3 ][ 3 ] = 0.5;
  cells[ 3 ][ 4 ] = 0.5;
  cells[ 3 ][ 5 ] = 0.0;

  cells[ 3 ][ 6 ] = 0.0;
  cells[ 3 ][ 7 ] = 0.3333333;
  cells[ 3 ][ 8 ] = 0.5;

  cells[ 3 ][ 9 ] = 0.5;
  cells[ 3 ][ 10 ] = 0.8333333;
  cells[ 3 ][ 11 ] = 0.5;

  offsets[3][0] = 0.125;
  offsets[3][1] = 0.083333333333;
  offsets[3][2] = 0.25;

  // bcc (100)
  cells[4].resize(2*3);

  cells[4][0] = 0.0;
  cells[4][1] = 0.0;
  cells[4][2] = 0.0;

  cells[4][3] = 0.5;
  cells[4][4] = 0.5;
  cells[4][5] = 0.5;

  offsets[4][0] = 0.25;
  offsets[4][1] = 0.25;
  offsets[4][2] = 0.25;

  // sc (100)
  cells[5].resize(1*3);

  cells[5][0] = 0.0;
  cells[5][1] = 0.0;
  cells[5][2] = 0.0;

  offsets[5][0] = 0.5;
  offsets[5][1] = 0.5;
  offsets[5][2] = 0.5;

  // dia (100)
  cells[6].resize(8*3);

  cells[6][0] = 0.0;
  cells[6][1] = 0.0;
  cells[6][2] = 0.0;

  cells[6][3] = 0.25;
  cells[6][4] = 0.25;
  cells[6][5] = 0.25;

  cells[6][6] = 0.5;
  cells[6][7] = 0.5;
  cells[6][8] = 0.0;

  cells[6][9] = 0.75;
  cells[6][10] = 0.75;
  cells[6][11] = 0.25;

  cells[6][12] = 0.5;
  cells[6][13] = 0.0;
  cells[6][14] = 0.5;

  cells[6][15] = 0.75;
  cells[6][16] = 0.25;
  cells[6][17] = 0.75;

  cells[6][18] = 0.0;
  cells[6][19] = 0.5;
  cells[6][20] = 0.5;

  cells[6][21] = 0.25;
  cells[6][22] = 0.75;
  cells[6][23] = 0.75;

  offsets[6][0] = 0.125;
  offsets[6][1] = 0.125;
  offsets[6][2] = 0.125;

  // A16 (alpha-S)
  cells[7].resize(128*3);

  cells[7][0] = 0.784402;
  cells[7][1] = 0.219899;
  cells[7][2] = 0.173704;
  cells[7][3] = 0.456903;
  cells[7][4] = 0.520498;
  cells[7][5] = 0.254011;
  cells[7][6] = 0.034397;
  cells[7][7] = 0.780101;
  cells[7][8] = 0.423694;
  cells[7][9] = 0.355399;
  cells[7][10] = 0.952596;
  cells[7][11] = 0.451598;
  cells[7][12] = 0.456903;
  cells[7][13] = 0.020498;
  cells[7][14] = 0.753991;
  cells[7][15] = 0.956903;
  cells[7][16] = 0.520498;
  cells[7][17] = 0.753991;
  cells[7][18] = 0.355399;
  cells[7][19] = 0.452596;
  cells[7][20] = 0.951619;
  cells[7][21] = 0.855399;
  cells[7][22] = 0.952596;
  cells[7][23] = 0.951619;
  cells[7][24] = 0.536198;
  cells[7][25] = 0.157298;
  cells[7][26] = 0.371004;
  cells[7][27] = 0.034397;
  cells[7][28] = 0.280101;
  cells[7][29] = 0.923714;
  cells[7][30] = 0.534397;
  cells[7][31] = 0.780101;
  cells[7][32] = 0.923714;
  cells[7][33] = 0.036198;
  cells[7][34] = 0.657298;
  cells[7][35] = 0.371004;
  cells[7][36] = 0.855399;
  cells[7][37] = 0.452596;
  cells[7][38] = 0.451598;
  cells[7][39] = 0.105404;
  cells[7][40] = 0.547403;
  cells[7][41] = 0.701588;
  cells[7][42] = 0.036198;
  cells[7][43] = 0.157298;
  cells[7][44] = 0.870983;
  cells[7][45] = 0.536198;
  cells[7][46] = 0.657298;
  cells[7][47] = 0.870983;
  cells[7][48] = 0.605404;
  cells[7][49] = 0.047403;
  cells[7][50] = 0.701588;
  cells[7][51] = 0.605404;
  cells[7][52] = 0.547403;
  cells[7][53] = 0.201608;
  cells[7][54] = 0.105404;
  cells[7][55] = 0.047403;
  cells[7][56] = 0.201608;
  cells[7][57] = 0.786203;
  cells[7][58] = 0.342701;
  cells[7][59] = 0.121014;
  cells[7][60] = 0.715598;
  cells[7][61] = 0.280101;
  cells[7][62] = 0.826295;
  cells[7][63] = 0.215598;
  cells[7][64] = 0.780101;
  cells[7][65] = 0.826295;
  cells[7][66] = 0.286203;
  cells[7][67] = 0.842701;
  cells[7][68] = 0.121014;
  cells[7][69] = 0.286203;
  cells[7][70] = 0.342701;
  cells[7][71] = 0.620994;
  cells[7][72] = 0.786203;
  cells[7][73] = 0.842701;
  cells[7][74] = 0.620994;
  cells[7][75] = 0.715598;
  cells[7][76] = 0.780101;
  cells[7][77] = 0.326316;
  cells[7][78] = 0.215598;
  cells[7][79] = 0.280101;
  cells[7][80] = 0.326316;
  cells[7][81] = 0.706898;
  cells[7][82] = 0.770502;
  cells[7][83] = 0.745989;
  cells[7][84] = 0.206898;
  cells[7][85] = 0.270502;
  cells[7][86] = 0.745989;
  cells[7][87] = 0.206898;
  cells[7][88] = 0.770502;
  cells[7][89] = 0.246009;
  cells[7][90] = 0.784402;
  cells[7][91] = 0.530097;
  cells[7][92] = 0.576306;
  cells[7][93] = 0.284402;
  cells[7][94] = 0.030097;
  cells[7][95] = 0.576306;
  cells[7][96] = 0.284402;
  cells[7][97] = 0.530097;
  cells[7][98] = 0.076285;
  cells[7][99] = 0.534397;
  cells[7][100] = 0.280101;
  cells[7][101] = 0.423694;
  cells[7][102] = 0.043097;
  cells[7][103] = 0.479502;
  cells[7][104] = 0.246009;
  cells[7][105] = 0.543097;
  cells[7][106] = 0.979502;
  cells[7][107] = 0.246009;
  cells[7][108] = 0.706898;
  cells[7][109] = 0.270502;
  cells[7][110] = 0.246009;
  cells[7][111] = 0.713797;
  cells[7][112] = 0.592697;
  cells[7][113] = 0.870983;
  cells[7][114] = 0.213797;
  cells[7][115] = 0.092697;
  cells[7][116] = 0.870983;
  cells[7][117] = 0.543097;
  cells[7][118] = 0.479502;
  cells[7][119] = 0.745989;
  cells[7][120] = 0.213797;
  cells[7][121] = 0.592697;
  cells[7][122] = 0.371004;
  cells[7][123] = 0.894596;
  cells[7][124] = 0.297400;
  cells[7][125] = 0.451598;
  cells[7][126] = 0.713797;
  cells[7][127] = 0.092697;
  cells[7][128] = 0.371004;
  cells[7][129] = 0.394596;
  cells[7][130] = 0.797400;
  cells[7][131] = 0.451598;
  cells[7][132] = 0.394596;
  cells[7][133] = 0.297400;
  cells[7][134] = 0.951619;
  cells[7][135] = 0.894596;
  cells[7][136] = 0.797400;
  cells[7][137] = 0.951619;
  cells[7][138] = 0.784402;
  cells[7][139] = 0.030097;
  cells[7][140] = 0.076285;
  cells[7][141] = 0.456903;
  cells[7][142] = 0.229498;
  cells[7][143] = 0.495999;
  cells[7][144] = 0.956903;
  cells[7][145] = 0.729498;
  cells[7][146] = 0.495999;
  cells[7][147] = 0.956903;
  cells[7][148] = 0.229498;
  cells[7][149] = 0.996019;
  cells[7][150] = 0.956903;
  cells[7][151] = 0.020498;
  cells[7][152] = 0.254011;
  cells[7][153] = 0.284402;
  cells[7][154] = 0.719899;
  cells[7][155] = 0.173704;
  cells[7][156] = 0.534397;
  cells[7][157] = 0.469902;
  cells[7][158] = 0.326316;
  cells[7][159] = 0.284402;
  cells[7][160] = 0.219899;
  cells[7][161] = 0.673684;
  cells[7][162] = 0.784402;
  cells[7][163] = 0.719899;
  cells[7][164] = 0.673684;
  cells[7][165] = 0.034397;
  cells[7][166] = 0.969902;
  cells[7][167] = 0.326316;
  cells[7][168] = 0.034397;
  cells[7][169] = 0.469902;
  cells[7][170] = 0.826295;
  cells[7][171] = 0.534397;
  cells[7][172] = 0.969902;
  cells[7][173] = 0.826295;
  cells[7][174] = 0.894596;
  cells[7][175] = 0.952596;
  cells[7][176] = 0.798391;
  cells[7][177] = 0.394596;
  cells[7][178] = 0.452596;
  cells[7][179] = 0.798391;
  cells[7][180] = 0.043097;
  cells[7][181] = 0.979502;
  cells[7][182] = 0.745989;
  cells[7][183] = 0.394596;
  cells[7][184] = 0.952596;
  cells[7][185] = 0.298412;
  cells[7][186] = 0.456903;
  cells[7][187] = 0.729498;
  cells[7][188] = 0.996019;
  cells[7][189] = 0.894596;
  cells[7][190] = 0.452596;
  cells[7][191] = 0.298412;
  cells[7][192] = 0.213797;
  cells[7][193] = 0.157298;
  cells[7][194] = 0.379006;
  cells[7][195] = 0.713797;
  cells[7][196] = 0.657298;
  cells[7][197] = 0.379006;
  cells[7][198] = 0.713797;
  cells[7][199] = 0.157298;
  cells[7][200] = 0.878985;
  cells[7][201] = 0.213797;
  cells[7][202] = 0.657298;
  cells[7][203] = 0.878985;
  cells[7][204] = 0.465603;
  cells[7][205] = 0.530097;
  cells[7][206] = 0.673684;
  cells[7][207] = 0.206898;
  cells[7][208] = 0.479502;
  cells[7][209] = 0.003980;
  cells[7][210] = 0.706898;
  cells[7][211] = 0.979502;
  cells[7][212] = 0.003980;
  cells[7][213] = 0.706898;
  cells[7][214] = 0.479502;
  cells[7][215] = 0.504001;
  cells[7][216] = 0.965603;
  cells[7][217] = 0.030097;
  cells[7][218] = 0.673684;
  cells[7][219] = 0.965603;
  cells[7][220] = 0.530097;
  cells[7][221] = 0.173704;
  cells[7][222] = 0.465603;
  cells[7][223] = 0.030097;
  cells[7][224] = 0.173704;
  cells[7][225] = 0.206898;
  cells[7][226] = 0.979502;
  cells[7][227] = 0.504001;
  cells[7][228] = 0.293101;
  cells[7][229] = 0.229498;
  cells[7][230] = 0.254011;
  cells[7][231] = 0.793101;
  cells[7][232] = 0.729498;
  cells[7][233] = 0.254011;
  cells[7][234] = 0.144601;
  cells[7][235] = 0.702600;
  cells[7][236] = 0.701588;
  cells[7][237] = 0.963802;
  cells[7][238] = 0.407302;
  cells[7][239] = 0.121014;
  cells[7][240] = 0.793101;
  cells[7][241] = 0.229498;
  cells[7][242] = 0.753991;
  cells[7][243] = 0.644601;
  cells[7][244] = 0.202600;
  cells[7][245] = 0.701588;
  cells[7][246] = 0.644601;
  cells[7][247] = 0.702600;
  cells[7][248] = 0.201608;
  cells[7][249] = 0.036198;
  cells[7][250] = 0.592697;
  cells[7][251] = 0.878985;
  cells[7][252] = 0.144601;
  cells[7][253] = 0.202600;
  cells[7][254] = 0.201608;
  cells[7][255] = 0.463802;
  cells[7][256] = 0.907302;
  cells[7][257] = 0.121014;
  cells[7][258] = 0.463802;
  cells[7][259] = 0.407302;
  cells[7][260] = 0.620994;
  cells[7][261] = 0.855399;
  cells[7][262] = 0.297400;
  cells[7][263] = 0.298412;
  cells[7][264] = 0.963802;
  cells[7][265] = 0.907302;
  cells[7][266] = 0.620994;
  cells[7][267] = 0.543097;
  cells[7][268] = 0.770502;
  cells[7][269] = 0.504001;
  cells[7][270] = 0.536198;
  cells[7][271] = 0.092697;
  cells[7][272] = 0.878985;
  cells[7][273] = 0.286203;
  cells[7][274] = 0.907302;
  cells[7][275] = 0.628996;
  cells[7][276] = 0.536198;
  cells[7][277] = 0.592697;
  cells[7][278] = 0.379006;
  cells[7][279] = 0.043097;
  cells[7][280] = 0.270502;
  cells[7][281] = 0.504001;
  cells[7][282] = 0.036198;
  cells[7][283] = 0.092697;
  cells[7][284] = 0.379006;
  cells[7][285] = 0.355399;
  cells[7][286] = 0.797400;
  cells[7][287] = 0.298412;
  cells[7][288] = 0.043097;
  cells[7][289] = 0.770502;
  cells[7][290] = 0.003980;
  cells[7][291] = 0.355399;
  cells[7][292] = 0.297400;
  cells[7][293] = 0.798391;
  cells[7][294] = 0.293101;
  cells[7][295] = 0.729498;
  cells[7][296] = 0.753991;
  cells[7][297] = 0.855399;
  cells[7][298] = 0.797400;
  cells[7][299] = 0.798391;
  cells[7][300] = 0.786203;
  cells[7][301] = 0.407302;
  cells[7][302] = 0.628996;
  cells[7][303] = 0.786203;
  cells[7][304] = 0.907302;
  cells[7][305] = 0.129016;
  cells[7][306] = 0.543097;
  cells[7][307] = 0.270502;
  cells[7][308] = 0.003980;
  cells[7][309] = 0.286203;
  cells[7][310] = 0.407302;
  cells[7][311] = 0.129016;
  cells[7][312] = 0.215598;
  cells[7][313] = 0.469902;
  cells[7][314] = 0.423694;
  cells[7][315] = 0.793101;
  cells[7][316] = 0.520498;
  cells[7][317] = 0.996019;
  cells[7][318] = 0.293101;
  cells[7][319] = 0.020498;
  cells[7][320] = 0.996019;
  cells[7][321] = 0.715598;
  cells[7][322] = 0.969902;
  cells[7][323] = 0.423694;
  cells[7][324] = 0.715598;
  cells[7][325] = 0.469902;
  cells[7][326] = 0.923714;
  cells[7][327] = 0.293101;
  cells[7][328] = 0.520498;
  cells[7][329] = 0.495999;
  cells[7][330] = 0.793101;
  cells[7][331] = 0.020498;
  cells[7][332] = 0.495999;
  cells[7][333] = 0.465603;
  cells[7][334] = 0.219899;
  cells[7][335] = 0.076285;
  cells[7][336] = 0.144601;
  cells[7][337] = 0.547403;
  cells[7][338] = 0.548401;
  cells[7][339] = 0.965603;
  cells[7][340] = 0.719899;
  cells[7][341] = 0.076285;
  cells[7][342] = 0.965603;
  cells[7][343] = 0.219899;
  cells[7][344] = 0.576306;
  cells[7][345] = 0.105404;
  cells[7][346] = 0.202600;
  cells[7][347] = 0.048381;
  cells[7][348] = 0.605404;
  cells[7][349] = 0.702600;
  cells[7][350] = 0.048381;
  cells[7][351] = 0.463802;
  cells[7][352] = 0.342701;
  cells[7][353] = 0.129016;
  cells[7][354] = 0.963802;
  cells[7][355] = 0.842701;
  cells[7][356] = 0.129016;
  cells[7][357] = 0.963802;
  cells[7][358] = 0.342701;
  cells[7][359] = 0.628996;
  cells[7][360] = 0.215598;
  cells[7][361] = 0.969902;
  cells[7][362] = 0.923714;
  cells[7][363] = 0.605404;
  cells[7][364] = 0.202600;
  cells[7][365] = 0.548401;
  cells[7][366] = 0.144601;
  cells[7][367] = 0.047403;
  cells[7][368] = 0.048381;
  cells[7][369] = 0.644601;
  cells[7][370] = 0.547403;
  cells[7][371] = 0.048381;
  cells[7][372] = 0.644601;
  cells[7][373] = 0.047403;
  cells[7][374] = 0.548401;
  cells[7][375] = 0.463802;
  cells[7][376] = 0.842701;
  cells[7][377] = 0.628996;
  cells[7][378] = 0.465603;
  cells[7][379] = 0.719899;
  cells[7][380] = 0.576306;
  cells[7][381] = 0.105404;
  cells[7][382] = 0.702600;
  cells[7][383] = 0.548401;

  offsets[7][0] = 0.0;
  offsets[7][1] = 0.0;
  offsets[7][2] = 0.0;


  // --- check
  if(mode != 0 && mode != 1) {
    cerr << "Atom_System::init_with_ucell(): Unknown mode " << mode << endl;
    exit(-1);
  }


  // --- set offset

  offset = new double[dim];
  if(offsetmode==0) {
    offset[0] = 0.0;
    offset[1] = 0.0;
    offset[2] = 0.0;
  }
  else if(offsetmode==1) {
    for(i=0;i<dim;i++) {
      offset[i] = offsets[crystal][i];
    }
  }
  else {
    cerr << "Atom_System::init_with_ucell(): Unknown offset mode " << offsetmode << endl;
    exit(-1);
  }


  // --- set some unit cell parameters
  nucell = cells[crystal].size()/dim;
  x0ucell = new double[nucell*dim];
  for(i=0;i<nucell*dim;i++) {
    x0ucell[i] = cells[crystal][i];
  }


  // --- check input
  if(atom_types.size() != nucell || atom_names.size() != nucell) {
    cout << "Atom_System::init_with_ucell(): Trying to init unit cell with wrong number of atoms." << endl;
    cout << "(This is a bug.) nucell atom_types.size() atom_names.size() " << nucell << " " << atom_types.size() << " " << atom_names.size() << endl;
    exit(-1);
  }
  if((int)size.size() != dim || (int)alat.size() != dim) {
    cerr << "Atom_System::init_with_ucell(): System size wrong: "
	 << (int)size.size() << ", " << dim << endl;
    exit(-1);
  }


  // --- set variables and reset system

  // number of atoms
  natoms = nucell;
  for(k=0;k<dim;k++) {
    natoms *= size[k];
  }

  // reset system
  reset(natoms);
  atypes.reset();

  // set box size
  for(k=0;k<dim;k++) {
    box[k] = alat[k]*size[k];
  }

  // add atom types
  for(i=0;i<atom_types.size();i++) {
    atypes.add(atom_names[i], atom_types[i]);
  }


  // --- create lattice
  ncreated = 0;

  // loop over unit cells
  for(ind[0]=0;ind[0]<size[0];ind[0]++) {
    for(ind[1]=0;ind[1]<size[1];ind[1]++) {
      for(ind[2]=0;ind[2]<size[2];ind[2]++) {

	// loop over atoms in unit cell
        for(n=0;n<nucell;n++) {

	  // check
	  if(ncreated >= natoms) {
	    cerr << "Atom_System::init(): Creating too many atoms!" << endl;
	    exit(-1);
	  }

	  // set coordinates
	  for(k=0;k<dim;k++) {
	    x0[3*ncreated + k] = (ind[k] + offset[k])*alat[k] + alat[k]*x0ucell[3*n + k];
	    if(mode==0) x0[3*ncreated + k] -= (0.5*size[k])*alat[k];
	  }

	  // check that atom doesn't exist already (badly done!)
	  for(k=0;k<ncreated;k++) {
	    if(x0[3*k] == x0[3*ncreated] && x0[3*k+1] == x0[3*ncreated+1]
	       && x0[3*k+2] == x0[3*ncreated+2]) {
	      cerr << "Atom_System::init(): Creating the same atom multiple times!" << endl;
	      exit(-1);
	    }
	  }

          // set atom type
	  atype[ncreated] = atom_types[n];
	  atypes.natype[atype[ncreated]]++;

	  ncreated++;
        }
      }
    }
  }

  if(ncreated != natoms) {
    cerr << "Atom_System::init(): --- Error with number of atoms." << endl;
    exit(-1);
  }


  delete [] x0ucell;
  delete [] offset;
}


void Atom_System::random_types(valarray<int> rtypes, vector<string> rnames, double pfirst) {

  /*
    Randomize atom types. At the moment, rtypes and rnames should be
    of length 2, i.e., only two types. pfirst gives the propability
    of each atom to be of the first type of the two.
  */


  // --- variables
  int i;  // loops
  int first[2];
  first[0] = 0;
  first[1] = 0;


  // --- check
  if(rtypes.size() != 2 || rnames.size() != 2 || pfirst < 0.0 || pfirst > 1.0) {
    cerr << "Atom_System::random_types(): rtypes.size, rnames.size, pfirst: " << rtypes.size() << " " << rnames.size() << " " << pfirst << endl;
    exit(-1);
  }


  // --- reset
  atypes.reset();


  // --- set types
  for(i=0;i<natoms;i++) {
    if(genrand() < pfirst) {
      if(first[0]==0) {
	first[0] = 1;
	atypes.add(rnames[0], rtypes[0]);
      }
      atype[i] = rtypes[0];
      atypes.natype[atype[i]]++;
    }
    else {
      if(first[1]==0) {
	first[1] = 1;
	atypes.add(rnames[1], rtypes[1]);
      }
      atype[i] = rtypes[1];
      atypes.natype[atype[i]]++;
    }
  }  // atom types set
}


void Atom_System::random_alloy(valarray<int> rtypes, vector<string> rnames, valarray<double> rconc) {

  /*
    Randomize atom types.

    All input arrays should be the same length.

    rtypes          typenumbers (Set to -1 to give one automatically, see Atom_Types::add().)
    rnames          atom type
    rconc           concentration, e.g., for Cu3Ag 3 and 1.
  */


  // --- variables
  int i;  // loops
  valarray<double> probs;  // probabilities for different components
  double totalprob;


  // --- check
  if(rtypes.size() != rnames.size() || rtypes.size() != rconc.size() || rtypes.size() <= 1) {
    cerr << "Atom_System::random_alloy(): rtypes.size, rnames.size, rconc.size: " << rtypes.size() << " " << rnames.size() << " " << rconc.size() << endl;
    exit(-1);
  }


  // --- compute probabilities
  totalprob = 0;
  for(i=0;i<rconc.size();i++) {
    totalprob += rconc[i];
  }

  probs.resize(rconc.size());
  for(i=0;i<probs.size();i++) {
    probs[i] = rconc[i]/totalprob;
  }

  totalprob = 0;
  for(i=0;i<probs.size();i++) {
    totalprob += probs[i];
  }


  // --- report
  cout << "Atom_System::random_alloy(): Creating random alloy with the following:" << endl;
  cout << "  type number concentration probability" << endl;
  for(i=0;i<probs.size();i++) {
    cout << "  " << rnames[i] << " " << rtypes[i] << " " << rconc[i] << " " << probs[i] << endl;
  }


  // --- reset atom types
  atypes.reset();
  for(i=0;i<rtypes.size();i++) {
    atypes.add(rnames[i], rtypes[i]);
  }
  

  // --- set types
  for(i=0;i<natoms;i++) {
    atype[i] = rtypes[pickrandom(probs)];
    atypes.natype[atype[i]]++;
  }
}


void Atom_System::makewulff(double cut100, double cut111) {


  /*
    Makes a Wulff polyhedron by cutting a piece of the system
    from the origin.

    cut100          cuts
    cut111          cuts

    Trims the box size to be 2*rmax, where rmax is the maximum
    distance of an atom from the centre.
  */


  // --- variables

  int i, j;  // loops
  int nwulff = 0;  // number of atoms in W polyhedron
  int take2w;  // 1 to take an atom to the polyhedron
  double rmax = 0;  // max distance of an atom from the origin along any axis

  // temp variables to store the polyhedron while making it
  double rp;
  double *w_at;
  int *w_type;


  // --- begin
  cout << "Atom_System::makewulff(): Making a Wulff polyhedron out of a system of " << natoms << " atoms." << endl;

  // check cuts
  if(cut100 <= 0 || cut111 <= 0) {
    cout << "Atom_System::makewulff(): Please give positive cuts for polyhedron." << endl;
    exit(-1);
  }

  // init temporary arrays
  w_at = new double[natoms*dim];
  w_type = new int[natoms];

  // reset number of atoms of different types
  atypes.zero();

  // --- loop over atoms
  nwulff = 0;
  for(i=0;i<natoms;i++) {

    // pick atoms to Wulff polyhedron
    take2w = 1;  // pick by default

    // remove unwanted atoms
    if (cut100 > 0) {
      if (x0[dim*i] > cut100 || x0[dim*i] < -cut100 || x0[dim*i+1] > cut100 || x0[dim*i+1] < -cut100 || x0[dim*i+2] > cut100 || x0[dim*i+2] < -cut100) {
	take2w=0;
      }
    }

    if (cut111 > 0) {
      rp=(x0[dim*i]+x0[dim*i+1]+x0[dim*i+2])/sqrt(3);
      if (rp > cut111 || rp < -cut111) {
	take2w=0;
      }
      rp=(-x0[dim*i]+x0[dim*i + 1]+x0[dim*i + 2])/sqrt(3);
      if (rp > cut111 || rp < -cut111) {
	take2w=0;
      }
      rp=(x0[dim*i]-x0[dim*i + 1]+x0[dim*i + 2])/sqrt(3);
      if (rp > cut111 || rp < -cut111) {
	take2w=0;
      }
      rp=(x0[dim*i]+x0[dim*i + 1]-x0[dim*i + 2])/sqrt(3);
      if (rp > cut111 || rp < -cut111) {
	take2w=0;
      }
    }

    // add wanted atom
    if(take2w==1) {
      for(j=0;j<dim;j++) {
	w_at[nwulff*dim + j] = x0[dim*i + j];
      }
      w_type[nwulff] = atype[i];
      nwulff++;

      // find atom with rmax
      for(j=0;j<dim;j++) {
	if(abs(x0[3*i + j]) > rmax) rmax = abs(x0[3*i + j]);
      }
    }
  }  // end of loop over atoms


  // --- move atoms in Wulff polyhedron to original arrays
  for(i=0;i<nwulff;i++) {
    for(j=0;j<dim;j++) {
      x0[dim*i + j] = w_at[dim*i + j];
    }
    atype[i] = w_type[i];
    atypes.natype[atype[i]]++;
  }

  // --- set box size
  box[0] = rmax*2;
  box[1] = rmax*2;
  box[2] = rmax*2;

  natoms = nwulff;
  delete [] w_at;
  delete [] w_type;
}


void Atom_System::createmti(double clayers, string crname, int crtype) {

  /*
    Creates a multiply twinned icosahedron (from Kristoffer).
    All atoms are created the same type.

    a                lattice constant
    clayers          FCC layers in one face
  */


  // --- variables
  int i, j, x, y, k;  // loops

  int n = 0;  // number of atoms
  double IcoposX[12], IcoposY[12], IcoposZ[12];
  int Polygon[20][3];
  double XDir[3], YDir[3], ZDir[3];
  int FoundA;  // found this atom already
  double newX, newY, newZ;

  double nnd = a/sqrt(2);  // nearest neighbour distance
  double dtheta = 2*M_PI/5;
  double theta = 0.0;
  double phi = atan2(2,1);
  int narray = 40000;
  double atoms_x[40000], atoms_y[40000], atoms_z[40000];

  double rmax = 0;  // max distance of an atom from the origin along any axis.

  // --- checks
  cerr << "Dt: " << dtheta << endl;
  fprintf(stderr, "Angle : %.10g\n",(phi/3.141592654)*180);


  // --- init Icopos*

  IcoposX[0] = 0;
  IcoposY[0] = 0;
  IcoposZ[0] = nnd;

  for(i=1;i<6;i++) {
    IcoposX[i] = cos(theta)*sin(phi)*nnd;
    IcoposY[i] = sin(theta)*sin(phi)*nnd;
    IcoposZ[i] = cos(phi)*nnd;
    theta += dtheta;
  }
  for(i=0;i<6;i++) {
    IcoposX[11-i] = -IcoposX[i];
    IcoposY[11-i] = -IcoposY[i];
    IcoposZ[11-i] = -IcoposZ[i];
  }

  Polygon[0][0] = 0;
  Polygon[0][1] = 1;
  Polygon[0][2] = 2;

  Polygon[1][0] = 0;
  Polygon[1][1] = 2;
  Polygon[1][2] = 3;

  Polygon[2][0] = 0;
  Polygon[2][1] = 3;
  Polygon[2][2] = 4;

  Polygon[3][0] = 0;
  Polygon[3][1] = 4;
  Polygon[3][2] = 5;

  Polygon[4][0] = 0;
  Polygon[4][1] = 5;
  Polygon[4][2] = 1;

  Polygon[5][0] = 1;
  Polygon[5][1] = 2;
  Polygon[5][2] = 7;

  Polygon[6][0] = 2;
  Polygon[6][1] = 3;
  Polygon[6][2] = 6;

  Polygon[7][0] = 3;
  Polygon[7][1] = 4;
  Polygon[7][2] = 10;

  Polygon[8][0] = 4;
  Polygon[8][1] = 5;
  Polygon[8][2] = 9;

  Polygon[9][0] = 5;
  Polygon[9][1] = 1;
  Polygon[9][2] = 8;

  Polygon[10][0] = 6;
  Polygon[10][1] = 7;
  Polygon[10][2] = 2;

  Polygon[11][0] = 7;
  Polygon[11][1] = 8;
  Polygon[11][2] = 1;

  Polygon[12][0] = 8;
  Polygon[12][1] = 9;
  Polygon[12][2] = 5;

  Polygon[13][0] = 9;
  Polygon[13][1] = 10;
  Polygon[13][2] = 4;

  Polygon[14][0] = 10;
  Polygon[14][1] = 6;
  Polygon[14][2] = 3;

  Polygon[15][0] = 11;
  Polygon[15][1] = 6;
  Polygon[15][2] = 7;

  Polygon[16][0] = 11;
  Polygon[16][1] = 7;
  Polygon[16][2] = 8;

  Polygon[17][0] = 11;
  Polygon[17][1] = 8;
  Polygon[17][2] = 9;

  Polygon[18][0] = 11;
  Polygon[18][1] = 9;
  Polygon[18][2] = 10;

  Polygon[19][0] = 11;
  Polygon[19][1] = 10;
  Polygon[19][2] = 6;


  // Consistency check for polygons, calculates the distances between vertices in polygons and
  // prints them
  // If polygonlist is OK, the distance should be the same for all.
  if(1==0) {
    for (i=0;i<20;i++) {
      printf("%.10g\n",sqrt((IcoposX[Polygon[i][0]]-IcoposX[Polygon[i][1]])*(IcoposX[Polygon[i][0]]-IcoposX[Polygon[i][1]])+(IcoposY[Polygon[i][0]]-IcoposY[Polygon[i][1]])*(IcoposY[Polygon[i][0]]-IcoposY[Polygon[i][1]])+(IcoposZ[Polygon[i][0]]-IcoposZ[Polygon[i][1]])*(IcoposZ[Polygon[i][0]]-IcoposZ[Polygon[i][1]])));
    }
    for (i=0;i<20;i++) {
      printf("%.10g\n",sqrt((IcoposX[Polygon[i][1]]-IcoposX[Polygon[i][2]])*(IcoposX[Polygon[i][1]]-IcoposX[Polygon[i][2]])+(IcoposY[Polygon[i][1]]-IcoposY[Polygon[i][2]])*(IcoposY[Polygon[i][1]]-IcoposY[Polygon[i][2]])+(IcoposZ[Polygon[i][1]]-IcoposZ[Polygon[i][2]])*(IcoposZ[Polygon[i][1]]-IcoposZ[Polygon[i][2]])));
    }
    for (i=0;i<20;i++) {
      printf("%.10g\n",sqrt((IcoposX[Polygon[i][0]]-IcoposX[Polygon[i][2]])*(IcoposX[Polygon[i][0]]-IcoposX[Polygon[i][2]])+(IcoposY[Polygon[i][0]]-IcoposY[Polygon[i][2]])*(IcoposY[Polygon[i][0]]-IcoposY[Polygon[i][2]])+(IcoposZ[Polygon[i][0]]-IcoposZ[Polygon[i][2]])*(IcoposZ[Polygon[i][0]]-IcoposZ[Polygon[i][2]])));
    }
  }


  // --- create icosahedron

  n = 0;  // number of atoms
  for (i=0;i<20;i++) {
    XDir[0] = IcoposX[Polygon[i][1]] - IcoposX[Polygon[i][0]];
    YDir[0] = IcoposY[Polygon[i][1]] - IcoposY[Polygon[i][0]];
    ZDir[0] = IcoposZ[Polygon[i][1]] - IcoposZ[Polygon[i][0]];
    XDir[1] = IcoposX[Polygon[i][2]] - IcoposX[Polygon[i][1]];
    YDir[1] = IcoposY[Polygon[i][2]] - IcoposY[Polygon[i][1]];
    ZDir[1] = IcoposZ[Polygon[i][2]] - IcoposZ[Polygon[i][1]];
    XDir[2] = IcoposX[Polygon[i][0]];
    YDir[2] = IcoposY[Polygon[i][0]];
    ZDir[2] = IcoposZ[Polygon[i][0]];

    for (j=0;j<clayers+1;j++) {

      for (x=0;x<=j;x++) {
	for (y=0;y<=x;y++) {
	  newX = x*XDir[0]+y*XDir[1]+j*XDir[2];
	  newY = x*YDir[0]+y*YDir[1]+j*YDir[2];
	  newZ = x*ZDir[0]+y*ZDir[1]+j*ZDir[2];

	  // check that atom is not there already
	  FoundA = 0;

	  for (k=0;k<n;k++) {
	    if(sqrt((atoms_x[k]-newX)*(atoms_x[k]-newX)+(atoms_y[k]-newY)*(atoms_y[k]-newY)+(atoms_z[k]-newZ)*(atoms_z[k]-newZ)) < 0.5*nnd) {
	      FoundA = 1;
	      break;
	    }
	  }

	  // if not, add
	  if (FoundA == 0) {
	    // check that the array is big enough
	    if(n > narray - 1) {
	      printf("createmti: Please give larger arrays.\n");
	      exit(-1);
	    }

	    // add
	    atoms_x[n] = newX;
	    atoms_y[n] = newY;
	    atoms_z[n] = newZ;
	    n++;

	  }
	}  // end of loop over y
      }  // end of loop over x
    }  // end of loop over layers
  }


  // --- put atoms where they belong
  reset(n);
  natoms = n;
  atypes.reset();
  atypes.add(crname, crtype);

  for(i=0;i<natoms;i++) {
    x0[3*i] = atoms_x[i];
    x0[3*i + 1] = atoms_y[i];
    x0[3*i + 2] = atoms_z[i];

    // find rmax
    for(j=0;j<dim;j++) {
      if(abs(x0[3*i + j]) > rmax) rmax = abs(x0[3*i + j]);
    }
    
    atype[i] = crtype;
    atypes.natype[crtype]++;
  }

  // box size
  for(i=0;i<dim;i++) {
    box[i] = 2*rmax;
  }
}


void Atom_System::createrandomcl(int nat, double ra, double rb, string crname, int crtype) {

  /*
    Create a random configuration with nat atoms with one atom at the
    origin and     others around it with nearest neighbour distances
    between ra and rb.
  */


  // --- variables
  int i, j, x, y, k;  // loops
  valarray<double> newpos(dim);  // position of new atom
  bool added;
  int nn;  // nearest neighbour to currently considered pos.
  double nr;  // distance of same


  // --- reset system
  reset(nat);
  atypes.reset();
  atypes.add(crname, crtype);
  natoms = 0;


  // --- add first atom
  for(i=0;i<dim;i++) {
    newpos[i] = 0.0;
  }
  add_atom(newpos, crname, 10);
  init_nlist(rb, rb + 0.01);


  // --- add other atoms
  while(natoms < nat) {
    added = false;

    while(!added) {
      // pick random position in box of size [ 2*natoms*rb ]^3
      for(i=0;i<dim;i++) {
	newpos[i] = 2*(genrand() - 1)*natoms*rb;
      }
      nn = find_closest_atom(newpos, nr);

      if(nn != -1 && nr > ra && nr < rb) {
	added = true;
      }
    }

    add_atom(newpos, crname, 10);
    add_to_linkedlist(natoms-1);
  }
}


void Atom_System::init_FCC_twin(string atname, const valarray<int> &size, int twinn) {

  // Create FCC twin system
  // a has to be set before
  // use twinn < 0 to get ordinary 111 FCC


  // --- variables

  valarray<int> ind(dim);  // loops
  int k, n;  // loops
  int ncreated;  // number of atoms created
  int nuc;  //  nuc = 6
  int temp;
  bool twinning;  // are we in the twin domain?
  double alat;
  int nucell;
  valarray<double> x0ucell;
  valarray<double> offset(dim);
  valarray<int> atypeucell;


  // --- check
  if((int)size.size() != dim) {
    cerr << "Atom_System::init(): System size wrong: "
	 << (int)size.size() << ", " << dim << endl;
    exit(-1);
  }
  if(twinn >= size[2] - 2 && twinn > 0) {
    cerr << "Atom_System::init_FCC_twin(): twinn, size[2] " << twinn << " " << size[2] << endl;
    exit(-1);
  }
  if(twinn > 0) cout << "Atom_System::init_FCC_twin(): twinn, size[2] " << twinn << " " << size[2] << endl;


  // --- set unit cell
  nuc = 6;
  nucell = nuc;
  x0ucell.resize(nucell*dim);
  atypeucell.resize(nucell);

  temp = 1;
  for(k=0;k<nuc;k++) {
    atypeucell[k] = temp;
  }

  offset[0]= 0.25;
  offset[1]= 0.0833333;
  offset[2]= 0.1666667;

  x0ucell[0] = 0.0;
  x0ucell[1] = 0.0;
  x0ucell[2] = 0.0;

  x0ucell[3] = 0.5;
  x0ucell[4] = 0.5;
  x0ucell[5] = 0.0;

  x0ucell[6] = 0.0;
  x0ucell[7] = 0.3333333;
  x0ucell[8] = 0.3333333;

  x0ucell[9] = 0.5;
  x0ucell[10] = 0.8333333;
  x0ucell[11] = 0.3333333;

  x0ucell[12] = 0.5;
  x0ucell[13] = 0.1666667;
  x0ucell[14] = 0.6666667;

  x0ucell[15] = 0.0;
  x0ucell[16] = 0.6666667;
  x0ucell[17] = 0.6666667;


  // --- set variables

  natoms = nucell;
  for(k=0;k<dim;k++) {
    natoms *= size[k];
  }

  reset(natoms);
  atypes.reset();
  atypes.add(atname, temp);

  for(k=0;k<dim;k++) {
    // take different a's into account
    if(k==0) alat = a/sqrt(2.0);
    else if(k==1) alat = a*sqrt(3.0/2);  // AARGH! Had to put the 3.0 here!
    else if(k==2) alat = a*sqrt(3.0);
    else {
      cerr << "Atom_System::init_FCC_twin(): k(dim) = " << k << endl;
      exit(-1);
    }

    box[k] = alat*size[k];
    cout << "Atom_System::init_FCC_twin(): Setting box[" << k << "] = " << box[k] << endl;
  }


  // --- create lattice
  ncreated = 0;
  twinning = false;

  // loop over unit cells
  for(ind[2]=0;ind[2]<size[2];ind[2]++) {
    for(ind[0]=0;ind[0]<size[0];ind[0]++) {
      for(ind[1]=0;ind[1]<size[1];ind[1]++) {

	// take twin into account
	if(ind[2] >= twinn && twinn >= 0) {
	  // alkup.
	  x0ucell[0] = 0.0;
	  x0ucell[1] = 0.0;
	  x0ucell[2] = 0.0;

	  x0ucell[3] = 0.5;
	  x0ucell[4] = 0.5;
	  x0ucell[5] = 0.0;

	  x0ucell[6] = 0.0;
	  x0ucell[7] = 0.3333333;
	  x0ucell[8] = 0.3333333;

	  x0ucell[9] = 0.5;
	  x0ucell[10] = 0.8333333;
	  x0ucell[11] = 0.3333333;

	  // ABC -> BAC
	  x0ucell[2] = 0.3333333;
	  x0ucell[5] = 0.3333333;

	  x0ucell[8] = 0.0;
	  x0ucell[11] = 0.0;

	  if(!twinning) {
	    twinning = true;

	    alat = (ind[2] + offset[2] - 0.5*size[2])*alat;
	    alat += (ind[2]-1 + offset[2] - 0.5*size[2])*alat + alat*0.6666667;
	    alat /= 2;

	    cout << "Twin limit: " << alat << " Å" << endl;
	  }
	}

	// loop over atoms in unit cell
        for(n=0;n<nucell;n++) {

	  // check
	  if(ncreated >= natoms) {
	    cerr << "Atom_System::init_FCC_twin(): Creating too many atoms!" << endl;
	    exit(-1);
	  }

	  // set coordinates
	  for(k=0;k<dim;k++) {
	    // take different a's into account
	    if(k==0) alat = a/sqrt(2.0);
	    else if(k==1) alat = a*sqrt(3.0/2);
	    else if(k==2) alat = a*sqrt(3.0);
	    else {
	      cerr << "Atom_System::init_FCC_twin(): k(dim) = " << k << endl;
	      exit(-1);
	    }

	    x0[3*ncreated + k] = (ind[k] + offset[k] - 0.5*size[k])*alat + alat*x0ucell[3*n + k];
	  }

	  // check that atom doesn't exist already (badly done!)
	  for(k=0;k<ncreated;k++) {
	    if(x0[3*k] == x0[3*ncreated] && x0[3*k+1] == x0[3*ncreated+1]
	       && x0[3*k+2] == x0[3*ncreated+2]) {
	      cerr << "Atom_System::init_FCC_twin(): Creating the same atom multiple times!" << endl;
	      exit(-1);
	    }
	  }

          // set atom type
	  atype[ncreated] = atypeucell[n];
	  atypes.natype[atype[ncreated]]++;

	  ncreated++;
        }
      }
    }
  }

  if(ncreated != natoms) {
    cerr << "Atom_System::init_FCC_twin(): --- Error with number of atoms." << endl;
    exit(-1);
  }

}


// --- getting stuff


double Atom_System::angle(int i, int j, int k) const {

  /*
    Returns the angle (ijk) between atoms i, j, and k.
    (In radians, in the range [0,pi].)
  */

  int l;
  valarray<double> v1, v2;  // vectors j->i and j->k
  double lv1, lv2;  // lengths of v1 and v2
  double alpha;  // angle in radians
  double dot;  // v1 \cdot v2

  // --- init
  v1.resize(dim);
  v2.resize(dim);

  // --- calculate
  dot = 0;
  lv1 = 0;
  lv2 = 0;
  for(l=0;l<dim;l++) {
    v1[l] = x0[3*i + l] - x0[3*j + l];
    v2[l] = x0[3*k + l] - x0[3*j + l];

    // pbc's, v1
    if(pbc[l]==1 && v1[l] > box[l]/2) {
      v1[l] -= box[l];
    }
    if(pbc[l]==1 && v1[l] < -box[l]/2) {
      v1[l] += box[l];
    }

    // pbc's, v2
    if(pbc[l]==1 && v2[l] > box[l]/2) {
      v2[l] -= box[l];
    }
    if(pbc[l]==1 && v2[l] < -box[l]/2) {
      v2[l] += box[l];
    }

    // v1 \cdot v2
    dot += v1[l]*v2[l];

    // lengths
    lv1 += v1[l]*v1[l];
    lv2 += v2[l]*v2[l];
  }

  // lenghts
  lv1 = sqrt(lv1);
  lv2 = sqrt(lv2);

  alpha = acos(dot/(lv1*lv2));
  if(alpha < 0 || alpha > M_PI) {
    cerr << "Atom_System::angle(): Error: It seems that the C acos() function is giving angles outside [0,pi]." << endl;
    exit(-1);
  }

  return alpha;
}


double Atom_System::dr(int i, int j) const {

  /*
    Returns the distance between atoms i and j.
  */

  int k;
  double dx;
  double dr = 0.0;

  for (k=0;k<dim;k++) {
    dx = abs(x0[3*i + k] - x0[3*j + k]);
    if(dx > box[k]/2.0 && pbc[k]==1) dx = box[k] - dx;
    dr += dx*dx;
  }
  dr = sqrt(dr);

  return dr;
}


double Atom_System::dr(int i, const Atom_System &system, int j) const {

  /*
    Returns the distance between atoms i and j, where j is from
    another system.

    No checks on i or j and if pbc=1 box is assumed to be the same
    size!
  */


  // --- variables
  int k;
  double dx;
  double dr = 0.0;

  for (k=0;k<dim;k++) {
    dx = abs(x0[3*i + k] - system.x0[3*j + k]);
    if(pbc[k]==1 && dx > box[k]/2.0) dx = box[k] - dx;
    dr += dx*dx;
  }
  dr = sqrt(dr);

  return dr;
}


double Atom_System::dr(int at, valarray<double> point) const {

  /*
    Returns the distance between atom at and a point in space.

    No checks on at.
  */


  // --- variables
  int k;  // loops
  double dx;
  double dr = 0.0;


  // --- check
  if(point.size() != dim) {
    cerr << "Atom_System::dr(): Point dimension " << point.size() << endl;
    exit(-1);
  }

  for (k=0;k<dim;k++) {
    dx = abs(x0[3*at + k] - point[k]);
    if(pbc[k]==1 && dx > box[k]/2.0) dx = box[k] - dx;
    dr += dx*dx;
  }
  dr = sqrt(dr);

  return dr;
}


valarray<double> Atom_System::dri(int i, int j) const {

  /*
    Return the vector x0[i] - x0[j] taking pbc's into account.
  */


  // --- variables
  int k;  // loops
  valarray<double> result(dim);  // result


  // --- calculate
  for(k=0;k<dim;k++) {
    result[k] = x0[3*i + k] - x0[3*j + k];

    // pbc's
    if(pbc[k]==1 && result[k] > box[k]/2) {
      result[k] -= box[k];
    }
    if(pbc[k]==1 && result[k] < -box[k]/2) {
      result[k] += box[k];
    }
  }

  return result;
}


valarray<double> Atom_System::dri(int i, int j, double &dr) const {

  /*
    Return the vector x0[i] - x0[j] and the i-j distance taking pbc's into account.
  */


  // --- variables
  int k;  // loops
  valarray<double> result(dim);  // result


  // --- calculate
  dr = 0.0;
  for(k=0;k<dim;k++) {
    result[k] = x0[3*i + k] - x0[3*j + k];

    // pbc's
    if(pbc[k]==1 && result[k] > box[k]/2) {
      result[k] -= box[k];
    }
    if(pbc[k]==1 && result[k] < -box[k]/2) {
      result[k] += box[k];
    }

    dr += result[k]*result[k];
  }
  dr = sqrt(dr);

  return result;
}


valarray<double> Atom_System::ri(int i) const {

  /*
    Return position of atom i
  */


  // --- variables
  int k;  // loops
  valarray<double> result(dim);  // result


  // --- calculate
  for(k=0;k<dim;k++) {
    result[k] = x0[3*i + k];
  }

  return result;
}


// moving / changing stuff


void Atom_System::moveatom(int i, const valarray<double> &dr, bool movefixed) {

  /*
    Moves atom i by the vector dr. Takes periodic boundary
    conditions into account and updates dx0.
  */


  // --- variables
  int j;

  // --- check
  if(dr.size() != dim) {
    cerr << "Atom_System::moveatom(): Wrong number of dimensions in argument." << endl;
    exit(-1);
  }
  if(i < 0 || i >= natoms) {
    cerr << "Atom_System::moveatom(): Invalid atom number." << endl;
    exit(-1);
  }

  // --- move atom
  for(j=0;j<dim;j++) {

    // check for fixed atoms
    if(fixx0[3*i + j] != 0 && dr[j] != 0) {
      if(!movefixed) {
        cerr << "Atom_System::moveatom(): Trying to move a fixed atom." << endl;
        exit(-1);}
    }

    x0[3*i + j] += dr[j];
    if(x0[3*i + j] > box[j]/2 && pbc[j]==1) x0[3*i + j] -= box[j];
    if(x0[3*i + j] < -box[j]/2 && pbc[j]==1) x0[3*i + j] += box[j];

    dx0[3*i + j] += dr[j];
  }
}


// --- input / output


void Atom_System::savexyz(string fname, int xyzmode, bool overwrite) {

  /*
    Saves the atom system to an xyz-file with name fname.

    xyzmode        columns
    0              plain xyz
    1              plain xyz + potential energy
    2              Cu x y z atype atindex vx vy vz
                     atype - atom type, negative for a fixed atom (fixed in any dof)
                     atindex - atom index, starting from 1
                     v - velocity (in Å/fs)
    3              Cu x y z atype atindex
    4              Cu x y z atype V block
    5              Cu x y z block
    6              Cu x y z T
    7              Cu x y z atype atindex V

    Remember to add new modes to xyzmodes().                     
  */


  // --- variables
  int i, j;
  ofstream afile;
  int tempatype;  // temp


  // --- report
  cout << "Atom_System::savexyz(): Saving xyz-file " << fname << "... (" << natoms << " atoms, box size: ";
  for(i=0;i<dim-1;i++) {
    cout << box[i] << " ";
  }
  cout << box[dim-1] << ")";


  // --- open output file
  if(overwrite) afile.open(fname.c_str(), ios::trunc);
  else afile.open(fname.c_str(), ios::app);
  assert(afile);

  afile.precision(14);
  //afile << setw(20);


  // --- write to file, coordinates
  afile << natoms << endl;
  afile << comment << " ";
  afile << endl;

  for(i=0;i<natoms;i++) {
    afile << atypes.atypename[atype[i]] << "  ";
    for(j=0;j<dim;j++) {
      afile << x0[3*i + j] << " ";
    }


    // --- xyz-modes

    // atom type for fixed atoms
    // (set negative if any d.o.f. fixed)
    tempatype = atype[i];
    for(j=0;j<dim;j++) {
      if(fixx0[3*i + j] == 1) {
        tempatype = -tempatype;
        break;
      }
    }

    if(xyzmode==0) {
      // ok.
    }
    else if(xyzmode==1) {
      afile << V[i];
    }
    else if(xyzmode==2) {
      afile << tempatype << " " << i+1 << " ";
      for(j=0;j<dim;j++) {
	afile << x1[3*i + j] << " ";
      }
    }
    else if(xyzmode==3) {
      afile << tempatype << " " << i+1;
    }
    else if(xyzmode==4) {
      afile << tempatype << " " << V[i] << " " << block[i];
    }
    else if(xyzmode==5) {
      afile << block[i];
    }
    else if(xyzmode==6) {
      afile << K[i]/(1.5*KBEV);
    }
    else if(xyzmode==7) {
      afile << tempatype << " " << i+1 << " " << V[i];
    }
    else {
      cerr << "Atom_System::savexyz(): Unknown xyzmode " << xyzmode << endl;
    }

    afile << endl;
  }


  cout << " ...done." << endl;
}


void Atom_System::savepov(string fname, int mode, double min, double max, bool overwrite) {

  /*
    Saves the atom system to an xyz-file with name fname.

    mode           columns
    0              only atom positions
    1              + colour [min,max] (from K[i])
    2              + potential energy (from V[i])
  */


  // --- variables
  int i, j;
  ofstream afile;
  double temp, r, g, b;


#if defined(DEBUG)
  cout << "Saving pov-file " << fname << "... (natoms: " << natoms << ") ";
#endif


  // --- open output file
  if(overwrite) afile.open(fname.c_str(), ios::trunc);
  else afile.open(fname.c_str(), ios::app);
  assert(afile);

  afile.precision(14);
  //afile << setw(20);


  // --- write to file

  // header
  {
    double time;
    comment_time(time);
    afile << "#declare time = " << time << " ;" << endl;
    afile << "#declare frame = union {" << endl;
  }

  // atoms
  for(i=0;i<natoms;i++) {
    afile << "  object { atom_" << atypes.atypename[atype[i]] << " translate < ";
    for(j=0;j<dim;j++) {
      afile << x0[3*i + j];
      if(j!=dim-1) afile << " , ";
    }
    afile << " >";

    // colour
    if(mode==1 || mode==2) {

      // colour, temperature (K[i]->T)
      if(mode==1) {
	temp = K[i]/(1.5*KBEV);
      }
      // colour, potential energy (V[i])
      else if(mode==2) {
	temp = V[i];
      }

      temp = (temp - min)/(max - min);

      if(temp < 0.5) {
        temp = 2*temp;
        if(temp>1.0) temp = 1.0;
        if(temp<0.0) temp = 0.0;
        r = 0.0;
        g = temp;
        b = 1.0 - temp;
      }
      else {
        temp = 2*(temp - 0.5);
        if(temp>1.0) temp = 1.0;
        if(temp<0.0) temp = 0.0;
        r = temp;
        g = 1.0 - temp;
        b = 0.0;
      }

      afile << " pigment { color rgb ( < " << r << " , " << g << " , " << b << " > ) }";
    }
    else if(mode != 0) {
      cerr << "Atom_System::savepov(): Unknown mode " << mode << endl;
      exit(-1);
    }

    afile << " }" << endl;
  }

  afile << "}" << endl;


#if defined(DEBUG)
  cout << " ...done." << endl;
#endif
}


void Atom_System::savexyz(string fname, int savetype, int xyzmode, bool overwrite) {

  /*
    Saves the atom system to an xyz-file with name fname.
    Saves only atoms of type savetype.

    xyzmode         additional columns
    1               potential energy
  */


  // --- variables
  int i, j;
  ofstream afile;


  // open output file
  if(overwrite) afile.open(fname.c_str(), ios::trunc);
  else afile.open(fname.c_str(), ios::app);
  assert(afile);


  // write to file
  afile << atypes.natype[savetype] << endl;
  afile << comment << " (Only atoms of type " << atypes.atypename[savetype] << ") ";
  afile << endl;

  for(i=0;i<natoms;i++) {
    if(atype[i] != savetype) continue;
    afile << atypes.atypename[atype[i]] << "  ";
    for(j=0;j<dim;j++) {
      afile << x0[3*i + j] << " ";
    }
    if(xyzmode==1) {
      afile << V[i];
    }

    afile << endl;
  }
}


void Atom_System::saveposcar(string fname, int mode, valarray<int> order) {

  /*
    Saves the atom system to a poscar-file, the format used by VASP. Coordinates
    will be scaled by the lattice constant.

    fname          file name
    mode           not used
    order          order in which atom types should be saved
                   (may contain types that aren't present in the system)

    mode           columns
    0              save x, y, z
    1              save x, y, z + "f1 f2 f3" where fi is F for fixed dof's and T for non-fixed
  */


  // --- variables
  int i, j, k;
  ofstream afile;
  bool overwrite = true;
  valarray<double> x0cell(dim*dim);


  // --- checks
  //if(order.size() != atypes.ntypes) {
  //  cerr << "Atom_System::saveposcar: Error: order.size() atypes.ntypes: " << order.size() << " " << atypes.ntypes << endl;
  //  exit(-1);
  //}
  if(a <= 0) {
    cerr << "Atom_System::saveposcar: Error: lattice constant: " << a << endl;
    exit(-1);
  }
  if(mode != 0 && mode != 1) {
    cerr << "Atom_System::saveposcar: Error: unknown mode " << mode << endl;
    exit(-1);
  }


  // --- unit cell (simple cell with box sizes assumed)
  if(dim != 3) {
    cout << "Atom_System::saveposcar: Error: dim = " << endl;
    exit(-1);
  }
  x0cell[0] = box[0];
  x0cell[1] = 0.0;
  x0cell[2] = 0.0;
  x0cell[3] = 0.0;
  x0cell[4] = box[1];
  x0cell[5] = 0.0;
  x0cell[6] = 0.0;
  x0cell[7] = 0.0;
  x0cell[8] = box[2];


  // --- open output file
  if(overwrite) afile.open(fname.c_str(), ios::trunc);
  else afile.open(fname.c_str(), ios::app);
  assert(afile);
  //afile.precision(14);
  //afile << setw(20);


  // --- write, comment
  //   afile << comment << endl;
  //   -> write atom types as comment -> read by VMD + makes sense
  afile << "   ";
  for(i=0;i<order.size();i++) {
    if(atypes.natype[order[i]] > 0) {
      afile << atypes.atypename[order[i]] << " ";
    }
  }
  afile << endl;

  // --- write, lattice constant
  afile << "   " << a << endl;

  // --- write, unit cell vectors
  for(i=0;i<dim;i++) {
    for(j=0;j<dim;j++) {
      afile << x0cell[dim*i + j] << " ";
    }
    afile << endl;
  }


  // --- write, numbers of atom types
  afile << "   ";
  for(i=0;i<order.size();i++) {
    if(atypes.natype[order[i]] > 0) {
      afile << atypes.natype[order[i]] << " ";
    }
  }
  afile << endl;


  // --- mark if fixed atoms are used
  if(mode==1) {
    afile << "Selective dynamics" << endl;
  }


  // --- write, coordinate convention for atom types
  afile << "Cartesian" << endl;


  // --- write, atoms
  for(k=0;k<order.size();k++) {

    if(atypes.natype[order[k]] == 0) {
      continue;
    }
    else if(atypes.natype[order[k]] < 0) {
      cerr << "Atom_System::saveposcar(): Error: Number of atoms of type " << order[k] << " = " << atypes.natype[order[k]] << endl;
      exit(-1);
    }

    for(i=0;i<natoms;i++) {

      // output types in correct order
      if(atype[i] == order[k]) {
        afile << " ";

	// coordinates
        for(j=0;j<dim;j++) {
          afile << x0[dim*i + j]/a << " ";
        }

	// fixed degreed of freedom for mode=1
	if(mode==1) {
	  for(j=0;j<dim;j++) {
	    if(fixx0[dim*i + j] == 0) {
	      afile << "T ";
	    }
	    else {
	      afile << "F ";
	    }
	  }
	}

        afile << endl;
      }
    }
  }
}


void Atom_System::save_siesta_fdf(string fname) {

  /*
    Saves the atom system to the fdf-format used by the DFT code
    Siesta.

    fname          file name

    Cell output but will be wrong if non-rectangular!
  */


  // --- variables
  ofstream afile;
  bool overwrite = true;


  // --- open output file
  if(overwrite) afile.open(fname.c_str(), ios::trunc);
  else afile.open(fname.c_str(), ios::app);
  assert(afile);
  //afile.precision(14);
  //afile << setw(20);


  // --- write, lattice vectors

  afile << "%block LatticeVectors" << endl;
  for(int i=0; i<dim; i++) {
    for(int j=0; j<dim; j++) {
      if(i==j) afile << box[i];
      else afile << "0.0";
      if(j<dim-1) afile << " ";
    }
    afile << endl;
  }
  afile << "%endblock LatticeVectors" << endl;


  // --- write, atomic species info

  afile << "%block Chemical_Species_label" << endl;
  for(int i=0; i<atypes.ntypes ; i++) {
    afile << atypes.types[i] << " " << atypes.Z[atypes.types[i]] << " " << atypes.atypename[atypes.types[i]] << endl;
  }
  afile << "%endblock Chemical_Species_label" << endl;
  afile << "Number_of_species " << atypes.ntypes << endl;


  // --- write, atoms

  afile << "NumberOfAtoms " << natoms << endl;
  afile << "AtomicCoordinatesFormat Ang" << endl;
  afile << "%block AtomicCoordinatesAndAtomicSpecies" << endl;

  // --- write, atoms
  for(int i=0;i<natoms;i++) {

    // coordinates
    for(int j=0;j<dim;j++) {
      afile << x0[dim*i + j] << " ";
    }

    // type
    afile << atype[i] << endl;
  }

  afile << "%endblock AtomicCoordinatesAndAtomicSpecies" << endl;
}


void Atom_System::readlammps_data(string fname, int mode) {

  /*
    Reads the atom system to a file in the format used for input by LAMMPS.

    fname          file name

    mode           lammps data file style
    -------------------------------------
    1              atomic
    2              atomic with velocities

    not implemented yet:
    3              charge
    4              charge with velocities
    5              molecular
    6              molecular with velocities

    sets:
    ntypes, atype, comment, atypename, natype, x0, natoms, x1, mass


    Assuming things in following order. There should be comment lines
    only before the actual data. The data should only contain empty
    lines apart from the actual data.


    5782 atoms
    6 atom types

    -12.66 12.66 xlo xhi
    -17.4295 17.4295 ylo yhi
    -100 100 zlo zhi

    Masses

    1 12.0107
    2 1.00794
    3 183.84
    4 12.0107
    5 183.84
    6 8.5722e+08

    Atoms

    1 1 -0.0900893 0.126966 -14.9911
    .
    .
    .


    Velocities

    1 0.715185 -0.626119 2.76185
    .
    .
    .
  */


  // --- variables
  ifstream afile;
  int expect;  // number of atoms expected
  string line;
  string dummy;
  bool ok;  // this part read
  int ntypes;  // number of atom types in file


  // --- checks
  if(mode < 1 || mode > 2) {
    cerr << "Atom_System::readlammps_data: Error: Unknown mode " << mode << endl;
    exit(-1);
  }


  // --- open file
  afile.open(fname.c_str());
  assert(afile);
  cout << "Atom_System::readlammps_data(): Reading file " << fname << " with mode " << mode << endl;


  // --- start reading frame

  // read number of atoms
  ok = false;
  while(!ok) {
    getline(afile, line);
    if(line != "") {
      istringstream temp(line);
      temp >> dummy;
      temp >> dummy;
      if(dummy == "atoms") {
	istringstream t2(line);
	t2 >> expect;
	ok = true;
      }
    }
  }
  if(!ok) {
    cerr << "Atom_System::readlammps_data(): Did not find number of atoms." << endl;
    exit(-1);
  }
  if(expect <= 0) {
    cerr << "Atom_System::readlammps_data(): expecting " << expect << " atoms?!" << endl;
    exit(-1);
  }

  // reset arrays
  moreatoms(expect);
  atypes.zero();

  // read number of atom types
  ok = false;
  while(!ok) {
    getline(afile, line);
    if(line != "") {
      istringstream temp(line);
      temp >> dummy;
      temp >> dummy;
      if(dummy == "atom") {
	temp >> dummy;
	if(dummy == "types") {
	  istringstream t2(line);
	  t2 >> ntypes;
	  ok = true;
	}
      }
    }
  }
  if(!ok) {
    cerr << "Atom_System::readlammps_data(): Did not find number of atom types." << endl;
    exit(-1);
  }
  if(ntypes <= 0) {
    cerr << "Atom_System::readlammps_data(): expecting " << ntypes << " atom types?!" << endl;
    exit(-1);
  }

  // read box, x
  ok = false;
  while(!ok) {
    double a, b;

    getline(afile, line);
    if(line != "") {
      istringstream temp(line);
      temp >> dummy;
      temp >> dummy;
      temp >> dummy;
      if(dummy == "xlo") {
	temp >> dummy;
	if(dummy == "xhi") {
	  istringstream t2(line);
	  t2 >> a;
	  t2 >> b;
	  ok = true;

	  box[0] = b - a;
	}
      }
    }
  }
  if(!ok) {
    cerr << "Atom_System::readlammps_data(): Did not find xlo, xhi." << endl;
    exit(-1);
  }

  // read box, y
  ok = false;
  while(!ok) {
    double a, b;

    getline(afile, line);
    if(line != "") {
      istringstream temp(line);
      temp >> dummy;
      temp >> dummy;
      temp >> dummy;
      if(dummy == "ylo") {
	temp >> dummy;
	if(dummy == "yhi") {
	  istringstream t2(line);
	  t2 >> a;
	  t2 >> b;
	  ok = true;

	  box[1] = b - a;
	}
      }
    }
  }
  if(!ok) {
    cerr << "Atom_System::readlammps_data(): Did not find ylo, yhi." << endl;
    exit(-1);
  }

  // read box, z
  ok = false;
  while(!ok) {
    double a, b;

    getline(afile, line);
    if(line != "") {
      istringstream temp(line);
      temp >> dummy;
      temp >> dummy;
      temp >> dummy;
      if(dummy == "zlo") {
	temp >> dummy;
	if(dummy == "zhi") {
	  istringstream t2(line);
	  t2 >> a;
	  t2 >> b;
	  ok = true;

	  box[2] = b - a;
	}
      }
    }
  }
  if(!ok) {
    cerr << "Atom_System::readlammps_data(): Did not find zlo, zhi." << endl;
    exit(-1);
  }

  // report box
  cout << "Atom_System::readlammps_data(): Read box size: " << box[0] << " " << box[1] << " " << box[2] << endl;

  // find line indicating atom masses
  ok = false;
  while(!ok) {
    getline(afile, line);
    if(line != "") {
      istringstream temp(line);
      temp >> dummy;
      if(dummy == "masses" || dummy == "Masses") {
	ok = true;
      }
    }
  }
  if(!ok) {
    cerr << "Atom_System::readlammps_data(): Did not find line \"Masses\"." << endl;
    exit(-1);
  }

  // read in atom masses, set atom types
  while(ntypes > 0) {
    ok = false;
    while(!ok) {
      int t;
      double m;
      string name;

      getline(afile, line);
      if(line != "") {
	istringstream temp(line);
	temp >> t;
	temp >> m;
	ok = true;

	name = stringify(t);

	atypes.add(name, t);
	atypes.setmass(t, m);
      }
    }
    ntypes--;
  }
  if(!ok) {
    cerr << "Atom_System::readlammps_data(): Did not find all atom masses." << endl;
    exit(-1);
  }
  //atypes.printtypes();

  // find line indicating atom coordinates
  ok = false;
  while(!ok) {
    getline(afile, line);
    if(line != "") {
      istringstream temp(line);
      temp >> dummy;
      if(dummy == "atoms" || dummy == "Atoms") {
	ok = true;
      }
    }
  }
  if(!ok) {
    cerr << "Atom_System::readlammps_data(): Did not find line \"Atoms\"." << endl;
    exit(-1);
  }

  // read in atom coordinates

  natoms = 0;
  while(natoms < expect) {
    getline(afile, line);
    if(line != "") {
      istringstream temp(line);
      int index;
      int type;

      if(mode==1 || mode==2) {
	temp >> index;
	index--;  // [1,N] -> [0,N-1]
	temp >> type;

	temp >> x0[index*dim + 0];
	temp >> x0[index*dim + 1];
	temp >> x0[index*dim + 2];

	atype[index] = type;
	atypes.natype[type]++;
	natoms++;
      }
    }
  }
  atypes.printtypes();

  // read in velocities

  if(mode==2) {
    // find line indicating atom velocities
    ok = false;
    while(!ok) {
      getline(afile, line);
      if(line != "") {
	istringstream temp(line);
	temp >> dummy;
	if(dummy == "velocities" || dummy == "Velocities") {
	  ok = true;
	}
      }
    }
    if(!ok) {
      cerr << "Atom_System::readlammps_data(): Did not find line \"Velocities\"." << endl;
      exit(-1);
    }

    // read in atom velocities

    int nr = 0;
    while(nr < expect) {
      getline(afile, line);
      if(line != "") {
	istringstream temp(line);
	int index;

	temp >> index;
	index--;  // [1,N] -> [0,N-1]

	temp >> x1[index*dim + 0];
	temp >> x1[index*dim + 1];
	temp >> x1[index*dim + 2];

	nr++;
      }
    }
  }
}


void Atom_System::savelammps(string fname, int mode, bool overwrite, string extra_headers) {

  /*
    Saves the atom system to a file in the format used by LAMMPS.

    fname          file name
    mode           mode, see below
    overwrite      overwrite file?
    extra_headers  for including extra headers (think "1 bond types\n123 bonds")
                     This way, extra headers can be added and angle,
		     bond etc. data appended afterwards as done by
		     euler.

    mode           lammps data file style
    -------------------------------------
    See check below
  */


  // --- variables
  int i;
  ofstream afile;


  // --- checks and notes

  if(dim != 3) {
    cerr << "Atom_System::savelammps: Error: dim = " << dim << endl;
    exit(-1);
  }
  if(mode < 1 || mode > 8) {
    cerr << "Atom_System::savelammps: Error: Unknown mode " << mode << endl;
    cerr << "  Available modes:" << endl;
    cerr << "  1              atomic" << endl;
    cerr << "  2              atomic with velocities" << endl;
    cerr << "  3              charge (charge will always be output as zero as not supported yet)" << endl;
    cerr << "  4              charge with velocities" << endl;
    cerr << "  5              full" << endl;
    cerr << "  6              full with velocities" << endl;
    cerr << "  7              molecular" << endl;
    cerr << "  8              molecular with velocities" << endl;
    cerr << endl;
    exit(-1);
  }
  if(mode==3 || mode==4 || mode==5 || mode==6) {
    cout << "Atom_System::savelammps: Warning: Setting charges to 0.0 as charge not really directly supported yet." << endl;
  }
  cout << "  Note: Outputting atom and molecule id's in [1,N] instead of [0,N-1]." << endl;


  // --- open output file
  if(overwrite) afile.open(fname.c_str(), ios::trunc);
  else afile.open(fname.c_str(), ios::app);
  assert(afile);
  //afile.precision(14);
  //afile << setw(20);


  // --- write, comment
  afile << "# " << comment << endl;
  for(i=0;i<atypes.ntypes;i++) {
    afile << "# " << atypes.types[i] << " = " << atypes.atypename[atypes.types[i]] << endl;
  }
  afile << "" << endl;
  afile << endl;

  // --- write, number of atoms, number of types
  afile << natoms << " atoms" << endl;
  afile << atypes.ntypes << " atom types" << endl;
  afile << endl;

  // --- write, box
  afile << -box[0]/2 << " " << box[0]/2 << " xlo xhi" << endl;
  afile << -box[1]/2 << " " << box[1]/2 << " ylo yhi" << endl;
  afile << -box[2]/2 << " " << box[2]/2 << " zlo zhi" << endl;
  afile << "" << endl;

  // --- write, extra headers
  if(extra_headers != "") {
    afile << extra_headers << endl;
  }

  // --- write, masses
  afile << "Masses" << endl;
  afile << "" << endl;
  atypes.setmasses(true);
  for(i=0;i<atypes.ntypes;i++) {
    afile << atypes.types[i] << " " << atypes.mass[atypes.types[i]] << endl;
  }
  afile << endl;

  // --- write, atoms
  afile << "Atoms" << endl;
  afile << endl;

  if(mode==1 || mode==2) for(i=0;i<natoms;i++) afile << i+1 << " " << atype[i] << " " << x0[dim*i + 0] << " " << x0[dim*i + 1] << " " << x0[dim*i + 2] << endl;
  if(mode==3 || mode==4) for(i=0;i<natoms;i++) afile << i+1 << " " << atype[i] << "  0.0  " << x0[dim*i + 0] << " " << x0[dim*i + 1] << " " << x0[dim*i + 2] << endl;
  if(mode==5 || mode==6) for(i=0;i<natoms;i++) afile << i+1 << " " << block[i]+1 << " " << atype[i] << "  0.0  " << x0[dim*i + 0] << " " << x0[dim*i + 1] << " " << x0[dim*i + 2] << endl;
  if(mode==7 || mode==8) for(i=0;i<natoms;i++) afile << i+1 << " " << block[i]+1 << " " << atype[i] << " " << x0[dim*i + 0] << " " << x0[dim*i + 1] << " " << x0[dim*i + 2] << endl;


  // --- write velocities
  if(mode==2 || mode==4 || mode==6) {
    afile << endl;
    afile << "Velocities" << endl;
    afile << endl;

    for(i=0;i<natoms;i++) {
      afile << i+1 << " " << x1[dim*i + 0] << " " << x1[dim*i + 1] << " " << x1[dim*i + 2] << endl;
    }
  }

  afile.close();
}


void Atom_System::savebgf(string fname, string descr, int mode, valarray<double> boxsize, valarray<double> angles, bool overwrite) {

  /*
    Saves a biogrf file.

    mode
    ----
    1         BIOGRF (these as defined in ReaxFF)
    2         XTLGRF
  */


  // --- variables
  int i;  // loops
  ofstream afile;  // file


  // --- check
  if(boxsize.size() != dim || angles.size() != dim || dim != 3) {
    cerr << "Atom_System::savebgf(): Error: box size, angles, dimension (=3): " << boxsize.size() << " " << angles.size() << " " << dim << endl;
    exit(-1);
  }


  // --- open output file
  if(overwrite) afile.open(fname.c_str(), ios::trunc);
  else afile.open(fname.c_str(), ios::app);
  assert(afile);


  // --- write to file

  // BIOGRF / XTLGRF
  if(mode==1) {
    afile << "BIOGRF";
  }
  else if(mode==2) {
    afile << "XTLGRF";
  }
  else {
    cerr << "Atom_System::savebgf(): Error: Unknown mode: " << mode << endl;
    exit(-1);
  }
  afile << " 200" << endl;

  // DESCRP
  afile << "DESCRP " << descr << endl;

  // REMARK
  afile << "REMARK " << comment << endl;

  // RUTYPE
  afile << "RUTYPE NORMAL RUN" << endl;

  // CRYSTX
  // format?
  if(mode==2) {
    afile.precision(5);
    afile.setf(ios::fixed);
    afile << "CRYSTX  ";
    for(i=0;i<dim;i++) {
      afile.width(10);
      afile << boxsize[i] << " ";
    }
    for(i=0;i<dim;i++) {
      afile.width(10);
      //afile << "90.00000";
      afile << angles[i];
      if(i<dim-1) {
        afile << " ";
      }
    }
    afile << endl;
  }

  // atoms
  // FORMAT ATOM   (a6,1x,i5,1x,a5,1x,a3,1x,a1,1x,a5,3f10.5,1x,a5,i3,i2,1x,f8.5)
  for(i=0;i<natoms;i++) {
    afile.precision(5);
    afile.setf(ios::fixed);

    afile << "HETATM" << " ";  // a6, 1x
    afile.width(5);
    afile << i+1 << " ";  // i5, 1x
    afile.width(2);  // ->
    afile << atypes.atypename[atype[i]] << " ";  // -> a5, 1x, a3, 1x, a1, 1x, a5 -> 17
    afile.width(14);  // ->
    afile << " ";  // ->
    afile.width(10);
    afile << x0[dim*i];  // f10.5
    afile.width(10);
    afile << x0[dim*i+1];  // f10.5
    afile.width(10);
    afile << x0[dim*i+2] << "    ";  // f10.5
    afile.width(2);  // loput ->
    afile << atypes.atypename[atype[i]] << "  1 1  0.00000";
    afile << endl;
  }

  // END
  afile << "END" << endl;
}


void Atom_System::savebgf(string fname, string descr, int mode, bool overwrite) {

  /*
    Saves a biogrf file with default box size and shape.

    mode
    ----
    1         BIOGRF (these as defined in ReaxFF)
    2         XTLGRF
  */


  // --- variables
  int i;  // loops
  valarray<double> boxsize(dim);
  valarray<double> angles(dim);


  // --- set defaults
  for(i=0;i<dim;i++) {
    boxsize[i] = box[i];
    angles[i] = 90.0;
  }

  savebgf(fname, descr, mode, boxsize, angles, overwrite);
}


void Atom_System::savevels(string fname) {

  /*
    Saves a vels-file as used by the ReaxFF code.
  */


  // --- variables
  int i, j;  // loops
  ofstream afile;  // file
  bool overwrite = true;  // overwrite file?
  valarray<double> dr(dim);  // translation to get to positive coordinate values


  // --- open output file
  if(overwrite) afile.open(fname.c_str(), ios::trunc);
  else afile.open(fname.c_str(), ios::app);
  assert(afile);


  // --- write to file

  // lattice parameters
  afile << " Lattice parameters:" << endl;
  afile.precision(8);
  afile.setf(ios::fixed);
  afile.setf(ios::uppercase);
  for(i=0;i<dim;i++) {
    afile << " ";
    afile.width(14);
    afile << box[i];
  }
  afile << endl;
  for(i=0;i<dim;i++) {
    afile << " ";
    afile.width(14);
    afile << 90.0;
  }
  afile << endl;

  // coordinates and atom types
  afile.width(4);
  afile << natoms;
  afile << " Atom coordinates (Angstrom):" << endl;

  // move to positive coordinate values
  /*
  cout << "Atom_System::savevels(): Warning: Moving all atoms to positive coordinate values by translating with box/2." << endl;
  dr[0] = box[0]/2;
  dr[1] = box[1]/2;
  dr[2] = box[2]/2;
  translate(dr, true);
  */

  afile.precision(15);
  afile.setf(ios::scientific, ios::floatfield);
  for(i=0;i<natoms;i++) {
    for(j=0;j<dim;j++) {
      afile << " ";
      afile.width(23);
      afile << x0[dim*i + j];
    }
    afile << " ";
    afile.width(2);
    afile << atypes.name(atype[i]);
    afile << endl;
  }

  // velocities
  //   -0.103271368325707E+10   0.111163551024464E+10   0.176918657817305E+10
  afile << " Atom velocities (Angstrom/s):" << endl;
  afile.precision(15);
  //  afile.setf(ios::fixed);
  for(i=0;i<natoms;i++) {
    for(j=0;j<dim;j++) {
      afile << " ";
      afile.width(23);
      afile << x1[dim*i + j]*1e15;  // in Å/s
    }
    afile << endl;
  }

  // accelerations (0 for now)
  afile << " Atom accelerations (Angstrom/s**2):" << endl;
  afile.precision(15);
  //  afile.setf(ios::fixed);
  for(i=0;i<natoms;i++) {
    for(j=0;j<dim;j++) {
      afile << " ";
      afile.width(23);
      afile << 0.0;  // should be in Å/s^2
    }
    afile << endl;
  }

  // previous accelerations (0 for now)
  afile << " Previous atom accelerations:" << endl;
  afile.precision(15);
  //  afile.setf(ios::fixed);
  for(i=0;i<natoms;i++) {
    for(j=0;j<dim;j++) {
      afile << " ";
      afile.width(23);
      afile << 0.0;  // should be in Å/s^2
    }
    afile << endl;
  }

  // other stuff:
  afile << " MD-temperature (K):" << endl;
  afile.precision(15);
  //  afile.setf(ios::fixed);
  afile.width(23);
  afile << 0.0;
  afile << endl;
  afile << " Connections, bond orders and lone pairs:" << endl;

  return;
}


void Atom_System::xyzmodes() {

  /*
    Print xyz modes to stdout
  */

  cout << "xyz-file modes used in Atom_System" << endl;
  cout << "----------------------------------" << endl;
  cout << endl;
  cout << "Input (additional columns are ignored)" << endl;
  cout << "-----" << endl;
  cout << endl;
  cout << "xyzmode        columns" << endl;
  cout << "----------------------" << endl;
  cout << "0              Cu x y z" << endl;
  cout << "1              Cu x y z atype" << endl;
  cout << "                 atype - atom type, negative for a fixed atom" << endl;
  cout << "2              Cu x y z atype atindex vx vy vz" << endl;
  cout << "                 atindex - atom index, starting from 1" << endl;
  cout << "                 v - velocity (in Å/fs)" << endl;
  cout << "3              Cu x y z V (see output mode 1)" << endl;
  cout << "4              Cu x y z atype vx vy vz" << endl;
  cout << "                 v - velocity (in Å/fs)" << endl;
  cout << "5              Cu x y z block (see output mode 5)" << endl;
  cout << endl;
  cout << "Output" << endl;
  cout << "------" << endl;
  cout << endl;
  cout << "xyzmode        columns" << endl;
  cout << "----------------------" << endl;
  cout << "0              plain xyz" << endl;
  cout << "1              plain xyz + potential energy" << endl;
  cout << "2              Cu x y z atype atindex vx vy vz" << endl;
  cout << "                 atype - atom type, negative for a fixed atom" << endl;
  cout << "                 atindex - atom index, starting from 1" << endl;
  cout << "                 v - velocity (in Å/fs)" << endl;
  cout << "3              Cu x y z atype atindex" << endl;
  cout << "4              Cu x y z atype V block" << endl;
  cout << "5              Cu x y z block" << endl;
  cout << "6              Cu x y z T" << endl;
  cout << "                 T of fixed atoms will be wrong!" << endl;
  cout << "7              Cu x y z atype atindex V" << endl;
  cout << endl;
  cout << "  Note that atom types will be negative if any degree of freedom is fixed." << endl;
  cout << endl;
  cout << endl;
}


void Atom_System::savemdcore(string fname, int mode) {

  /*
    Saves the atom system to a file in the format used by MDCORE.

    fname          file name
    mode           1 - no clusters
                   2 - with clusters (if clustering not done, arbitrary results!)
  */


  // --- variables
  int i, j, k;
  ofstream afile;
  bool overwrite = true;
  ofstream atoms;

  vector<int> next(natoms);  // if next(i)!=0, it points to the index of the next atom in the cluster
  vector<bool> clusterdone(natoms);  // this cluster already done?


  // --- checks
  if(dim != 3) {
    cerr << "Atom_System::savemdcore: Error: dim = " << endl;
    exit(-1);
  }
  if(mode < 1 || mode > 2) {
    cerr << "Atom_System::savemdcore: Unknown mode: " << mode << endl;
    exit(-1);
  }


  // --- construct next-array for mode 2
  if(mode==2) {
    int iat;  // current atom
    int icl;  // current cluster

    for(i=0;i<natoms;i++) {
      next[i] = -1;
      clusterdone[i] = false;
    }

    // loop over clusters
    for(i=0;i<natoms;i++) {
      icl = block[i];  // current cluster

      if(clusterdone[icl]==false) {
	clusterdone[icl] = true;

	if(blocknat.size() < icl+1) {
	  cerr << "Atom_System::savemdcore: Error: Saving with cluster information but clustering not done / read in!" << endl;
	  exit(-1);
	}

	if(blocknat[icl]==1) {
	  next[i] = 0;
	}
	else {
	  // loop over other atoms
	  iat = i;
	  for(j=0;j<natoms;j++) {
	    if(block[j]==block[i]) {
	      next[iat] = j+1;
	      iat = j;
	    }
	  }  // end of loop over other atoms
	  next[iat] = 0;
	}
      }
    }
  }


  // --- open output file

  if(overwrite) afile.open(fname.c_str(), ios::trunc);
  else afile.open(fname.c_str(), ios::app);
  assert(afile);
  //afile.precision(14);
  //afile << setw(20);
  atoms.open("euler_mdcorehelp", ios::trunc);
  assert(atoms);


  // --- write coordinates

  afile << " <--- Total number of atoms" << endl;
  afile << natoms << endl;

  afile << " <--- Number of occupied orbitals" << endl;
  afile << 0.0 << endl;

  afile << " <--- Element, atomic mass, coordinates, group (=1), two zeroes, possibly molecule index" << endl;
  for(i=0;i<natoms;i++) {
    afile << atypes.atypename[atype[i]] << " " << atypes.mass[atype[i]] << " " << x0[3*i+0] << " " << x0[3*i+1] << " " << x0[3*i+2] << "  1    0.0000000000E+00    0.0000000000E+00  ";
    if(mode==2) {
      afile << next[i] << endl;
    }
    else {
      afile << endl;
    }
    atoms << atypes.atypename[atype[i]] << " " << i << endl;
  }


  // --- write velocities

  afile << " <--- Velocities" << endl;
  for(i=0;i<natoms;i++) {
    afile << x1[3*i+0] << " " << x1[3*i+1] << " " << x1[3*i+2] << endl;
  }


  // --- write cell

  afile << " <--- cell" << endl;
  afile << box[0] << " " << 0.0    << " " << 0.0    << endl;
  afile << 0.0    << " " << box[1] << " " << 0.0    << endl;
  afile << 0.0    << " " << 0.0    << " " << box[2] << endl;
}


void Atom_System::readxyz(string fname, int xyzmode, int frame) {

  /*
    Reads an xyz-file.

    Remember to add new modes to xyzmodes().                     

    sets:
    ntypes, atype, comment, atypename, natype, x0, natoms, fixx0, V, K, x1, F, mass, block=0
  */


  // --- variables
  int i;  // loops
  ifstream afile;  // xyz-file
  int thisframe;  // current frame

  int expect;  // number of atoms expected
  string dummy;
  bool ok;  // finished reading atoms?

  // fixed atoms
  int nfixed;
  valarray<int> fixd(dim);

  int temp;

  // additional columns
  int c5;
  bool indexwarned = false;  // warn of index mismatch only once

  valarray<double> xmax(1.0, dim);


  // --- start reading file

  afile.open(fname.c_str());
  assert(afile);
  cout << "Atom_System::readxyz(): Reading frame " << frame << " of file " << fname << endl;


  // --- skip frames if necessary

  // check
  if(frame <= 0) {
    cerr << "Atom_System::readxyz(): Cannot read frame " << frame << endl;
    exit(-1);
  }

  // skip frames
  thisframe = 1;
  while(thisframe < frame) {

    afile >> expect;
    getline(afile, dummy, '\n');  // dummy, next line
    //cout << "  Expecting " << expect << " atoms in frame " << thisframe << endl;
    getline(afile, dummy, '\n');  // comment line
    //cout << "  Comment " << thisframe << ": " << dummy << endl;
    for(i=0;i<expect;i++) {
      //getline(afile, dummy, '\n');  // atoms
      afile.ignore(99999999, '\n');
    }
    thisframe++;
  }


  readxyzframe(afile, xyzmode);
}


void Atom_System::readxyzframe(ifstream &afile, int xyzmode) {

  /*
    Reads the next frame from an xyz-file.

    See Atom_System::xyzmodes() for xyz-modes.

    sets:
    ntypes, atype, comment, atypename, natype, x0, natoms, fixx0, V, K, x1, F, mass, block=0
  */


  // --- variables
  int i;                                       // loops
  int expect;                                  // number of atoms expected
  string dummy;
  bool ok;                                     // finished reading atoms?

  // fixed atoms
  int nfixed;
  valarray<int> fixd(dim);

  int temp;

  // additional columns
  int c5;
  bool indexwarned = false;                    // warn of index mismatch only once
  int ncl;                                     // number of blocks for mode 5

  valarray<double> xmax(1.0, dim);


  cout << "Atom_System::readxyzframe(): Reading frame with mode " << xyzmode << endl;


  // --- start reading frame

  // --- number of atoms
  afile >> expect;
  if(expect <= 0) {
    cerr << "Atom_System::readxyzframe(): expecting " << expect << " atoms?!" << endl;
    exit(-1);
  }
  cout << "  Expecting " << expect << " atoms." << endl;

  // --- dummy, next line
  getline(afile, comment, '\n');

  // --- comment line
  getline(afile, comment, '\n');

  // init arrays and reset stuff
  cout << "Atom_System::readxyzframe(): Resizing arrays with size " << expect << "..." << endl;
  moreatoms(expect);
  zero_arrays();
  atypes.zero();

  // read atoms
  natoms = 0;
  nfixed = 0;
  fixd = 0;

  ok = false;
  while(!ok) {
    string line;
    getline(afile, line, '\n');
    stringstream ll(line);

    // --- atom type name
    //     (
    //      don't know why I used getline earlier: this fails when
    //      first character on line is a whitespace:
    //      getline(afile, dummy, ' ');
    //     )
    ll >> dummy;

    // --- coordinates
    for(i=0;i<dim;i++) {
      ll >> x0[natoms*dim + i];
      if(abs(x0[natoms*dim + i]) > xmax[i]) xmax[i] = abs(x0[natoms*3 + i]);
      // default to non-fixed atoms
      fixx0[natoms*dim + i] = 0;
    }

    // --- different modes = other stuff

    // mode 0
    if(xyzmode == 0) {
      // set atom type
      temp = -1;
      atypes.add(dummy, temp);
      atype[natoms] = temp;
    }

    // mode 1
    else if(xyzmode == 1) {
      // get atom type number, negative for a fixed atom
      ll >> c5;

      // take fixed atom into account
      if(c5 < 0) {
	c5 = -c5;
	nfixed++;
	for(i=0;i<dim;i++) {
	  fixx0[3*natoms + i] = 1;
	  fixd[i]++;
	}
      }

      // set atom type
      atypes.add(dummy, c5);
      atype[natoms] = c5;
    }

    // mode 2
    else if(xyzmode == 2) {
      // get atom type number, negative for a fixed atom
      ll >> c5;
      // take fixed atom into account
      if(c5 < 0) {
	c5 = -c5;
	nfixed++;
	for(i=0;i<dim;i++) {
	  fixx0[3*natoms + i] = 1;
	  fixd[i]++;
	}
      }
      // set atom type
      atypes.add(dummy, c5);
      atype[natoms] = c5;

      // atom index
      ll >> c5;
      if(!indexwarned && c5 != natoms+1) {
        indexwarned = true;
	cerr << "Atom_System::readxyzframe(): Warning: atom index, i+1: " << c5 << " " << natoms+1 << " (supressing further warnings)" << endl;
      }

      // velocities
      for(i=0;i<dim;i++) {
	ll >> x1[natoms*dim + i];
      }
    }

    // mode 3
    else if(xyzmode == 3) {
      ll >> V[natoms];

      // set atom type
      temp = -1;
      atypes.add(dummy, temp);
      atype[natoms] = temp;
    }

    // mode 4
    else if(xyzmode == 4) {
      // get atom type number, negative for a fixed atom
      ll >> c5;
      // take fixed atom into account
      if(c5 < 0) {
	c5 = -c5;
	nfixed++;
	for(i=0;i<dim;i++) {
	  fixx0[3*natoms + i] = 1;
	  fixd[i]++;
	}
      }
      // set atom type
      atypes.add(dummy, c5);
      atype[natoms] = c5;

      // velocities
      for(i=0;i<dim;i++) {
	ll >> x1[natoms*dim + i];
      }
    }

    // mode 5
    else if(xyzmode == 5) {
      // get atom type number, negative for a fixed atom
      ll >> block[natoms];
      if(block[natoms] > ncl) ncl = block[natoms];

      // set atom type
      temp = -1;
      atypes.add(dummy, temp);
      atype[natoms] = temp;
    }

    // error: unknown mode
    else {
      cerr << "Atom_System::readxyzframe(): Error: Unknown xyz-mode " << xyzmode << endl;
      cerr << "                             Printing allowed modes to standard output." << endl;
      xyzmodes();
      exit(-1);
    }

    // --- check if line was too short
    if(ll.fail()) {
      cerr << "Atom_System::readxyzframe(): Error: Line has too few entries." << endl;
      exit(-1);
    }

    // --- calc. number of atoms
    atypes.natype[atype[natoms]]++;
    natoms++;

    // --- check if we're done
    if(natoms == expect) {
      ok = true;
      break;
    }
  }  // atoms read in


  // check
  if(expect != natoms) {
    cerr << "Expected " << expect << " atoms but found " << natoms << "." << endl;
    exit(-1);
  }


  // for mode 5, calculate the number of atoms in each block
  if(xyzmode==5) {
    blocknat.resize(ncl);
    blocknat = 0;
    for(i=0;i<natoms;i++) {
      // check
      if(block[i] < 0) {
	cerr << "Atom_System::cluster(): Error: block[" << i << "] = " << block[i] << ". Has to be >= 0." << endl;
	exit(-1);
      }
      blocknat[block[i]]++;
    }
  }


  // report
  cout << "  Read in " << natoms << " atoms of which " << nfixed << " fixed. (";
  for(i=0;i<dim;i++) {
    cout << fixd[i];
    if(i<dim-1) cout << ", ";
  }
  cout << ")" << endl;

  atypes.printtypes();
}


double Atom_System::readlammps_custom_dump_frame(ifstream &afile, bool skip) {

  /*
    Read the next frame from a LAMMPS output file of "dump custom" style.

    - For skip = true, not really read, just skip frame.
    - The routine returns the time read from the frame.
    - Atoms not translated to within periodic boundaries, but possibility exists.

    sets:
    ntypes, atype, atypename, natype, x0, natoms, x1, F, mass?, block=0, fixx0=0
      or block=mol if molecule id read in

    The file format looks like this:
      ITEM: TIMESTEP
      0
      ITEM: NUMBER OF ATOMS
      8914834
      ITEM: BOX BOUNDS
      -516.354 516.354
      -103.659 103.659
      -413.383 413.383
      ITEM: ATOMS id x y z vx vy vz c_pe c_stress[1] c_stress[2] c_stress[3] c_stress[4] c_stress[5] c_stress[6]
      1 -513.936 -101.594 -411.21 0 0 0 -3.40682 -1.51225e+06 -988022 -629579 -667546 1.43086e+06 244147
      ...
  */


  // --- variables
  int i, at;      // loops
  int expect;    // number of atoms expected
  string dummy, dummy2;
  double t1, t2;
  valarray<double> translation(3);  // negative origin of system: used as translation to get origin to -box/2
  vector<int> col_x(3), col_v(3);   // columns where x, v, found
  int col_id, col_type;             // columns where index, type found
  int col_mol;                      // column: mol
  double time;                      // time read in


  // --- checks

  if(dim != 3) {
    cerr << "Atom_System::readlammps_custom_dump_frame(): Error: Only supporting three-dimensional systems (now dim = " << dim << ")." << endl;
    exit(-1);
  }


  // --- start reading frame

  if(!skip) cout << "Atom_System::readlammps_custom_dump_frame(): Reading frame." << endl;


  // --- time
  //     ITEM: TIMESTEP
  //     0

  // header
  getline(afile, dummy);
  {
    istringstream line(dummy);
    line >> dummy;
    line >> dummy;
    if(dummy != "TIMESTEP") {
      cerr << "Atom_System::readlammps_custom_dump_frame(): Error: Expected \"ITEM: TIMESTEP\"" << endl;
      exit(-1);
    }
  }
  // time
  getline(afile, dummy);
  {
    istringstream line(dummy);
    line >> time;
  }

  // --- number of atoms
  //     ITEM: NUMBER OF ATOMS
  //     8914834

  // header
  getline(afile, dummy);
  {
    istringstream line(dummy);
    line >> dummy;
    line >> dummy;
    if(dummy != "NUMBER") {
      cerr << "Atom_System::readlammps_custom_dump_frame(): Error: Expected \"ITEM: NUMBER OF ATOMS\", got \"" << dummy << "\"." << endl;
      exit(-1);
    }
  }
  // natoms
  getline(afile, dummy);
  {
    istringstream line(dummy);
    line >> expect;
    if(expect <= 0) {
      cerr << "Atom_System::readlammps_custom_dump(): expecting " << expect << " atoms?!" << endl;
      exit(-1);
    }
  }
  if(!skip) cout << "  Expecting " << expect << " atoms." << endl;


  // --- box size and origin
  //     ITEM: BOX BOUNDS
  //     -516.354 516.354
  //     -103.659 103.659
  //     -413.383 413.383

  // header
  getline(afile, dummy);
  {
    istringstream line(dummy);
    line >> dummy;
    line >> dummy;
    if(dummy != "BOX") {
      cerr << "Atom_System::readlammps_custom_dump_frame(): Error: Expected \"ITEM: BOX BOUNDS\"" << endl;
      exit(-1);
    }
  }
  // box
  for(i=0;i<dim;i++) {
    getline(afile, dummy);
    {
      istringstream line(dummy);
      line >> t1;
      line >> t2;
      if(t2 < t1) {
	cerr << "Atom_System::readlammps_custom_dump_frame(): Error: Second box bound < first." << endl;
	exit(-1);
      }
    }
    box[i] = t2-t1;
    translation[i] = -box[i]/2.0 - t1;
  }

  // --- atoms
  //     ITEM: ATOMS id x y z vx vy vz c_pe c_stress[1] c_stress[2] c_stress[3] c_stress[4] c_stress[5] c_stress[6]
  //     1 -513.936 -101.594 -411.21 0 0 0 -3.40682 -1.51225e+06 -988022 -629579 -667546 1.43086e+06 244147

  // skip, if not actually reading
  if(skip) {
    // header
    getline(afile, dummy);
    // atoms
    for(i=0;i<expect;i++) {
      getline(afile, dummy);
    }
    return 0.0;
  }

  // header, determine columns
  for(i=0;i<dim;i++) {
    col_x[i] = 0;
    col_v[i] = 0;
  }
  col_id = 0;
  col_type = 0;
  col_mol = 0;

  getline(afile, dummy);
  {
    istringstream line(dummy);
    int col = 0;

    line >> dummy;
    line >> dummy;
    if(dummy != "ATOMS") {
      cerr << "Atom_System::readlammps_custom_dump_frame(): Error: Expected \"ITEM: ATOMS ...\"" << endl;
      exit(-1);
    }
    while(line >> dummy) {
      col++;
      if(dummy=="id") col_id = col;
      if(dummy=="mol") col_mol = col;
      if(dummy=="type") col_type = col;
      if(dummy=="x") col_x[0] = col;
      if(dummy=="y") col_x[1] = col;
      if(dummy=="z") col_x[2] = col;
      if(dummy=="vx") col_v[0] = col;
      if(dummy=="vy") col_v[1] = col;
      if(dummy=="vz") col_v[2] = col;
    }
  }

  // check
  if(col_id==0) {
    cout << "  Warning: Did not locate atom data column for \"id\", taking the order in which atoms read in." << endl;
  }
  else {
    cout << "  Note: Atom id will be converted from [1,N] to [0,N-1]." << endl;
  }
  if(col_mol != 0) {
    cout << "  Note: Molecule id will be converted from [1,N] to [0,N-1]." << endl;
  }
  if(col_type==0) {
    cout << "  Warning: Did not locate atom data column for \"type\", assuming all type \"1\"." << endl;
  }
  if(col_x[0]==0) {
    cerr << "Atom_System::readlammps_custom_dump_frame(): Error: Did not locate atom data column for \"x\"." << endl;
    exit(-1);
  }
  if(col_x[1]==0) {
    cerr << "Atom_System::readlammps_custom_dump_frame(): Error: Did not locate atom data column for \"y\"." << endl;
    exit(-1);
  }
  if(col_x[2]==0) {
    cerr << "Atom_System::readlammps_custom_dump_frame(): Error: Did not locate atom data column for \"z\"." << endl;
    exit(-1);
  }

  // report what is being read in
  cout << "  Reading in ";
  if(col_id != 0) cout << "id ";
  if(col_mol != 0) cout << "mol ";
  if(col_type != 0) cout << "type ";
  if(col_x[0] != 0) cout << "x ";
  if(col_x[1] != 0) cout << "y ";
  if(col_x[2] != 0) cout << "z ";
  if(col_v[0] != 0) cout << "vx ";
  if(col_v[1] != 0) cout << "vy ";
  if(col_v[2] != 0) cout << "vz ";
  cout << "from columns ";
  if(col_id != 0) cout << col_id << " ";
  if(col_mol != 0) cout << col_mol << " ";
  if(col_type != 0) cout << col_type << " ";
  if(col_x[0] != 0) cout << col_x[0] << " ";
  if(col_x[1] != 0) cout << col_x[1] << " ";
  if(col_x[2] != 0) cout << col_x[2] << " ";
  if(col_v[0] != 0) cout << col_v[0] << " ";
  if(col_v[1] != 0) cout << col_v[1] << " ";
  if(col_v[2] != 0) cout << col_v[2] << " ";
  cout <<  endl;

  // read atom data

  cout << "  Resizing arrays with size " << expect << "." << endl;
  reset(expect);
  atypes.zero();
  natoms = 0;

  for(at=0;at<expect;at++) {

    // get line
    getline(afile, dummy);

    // check
    {
      istringstream line(dummy);
      line >> dummy2;
      if(dummy2 == "ITEM:") {
	line >> dummy2;
	cerr << "Atom_System::readlammps_custom_dump_frame(): Error: Expected atom data, not \"" << line << "\"" << endl;
	exit(-1);
      }
    }

    // get data
    {
      istringstream line(dummy);
      int col = 0;

      vector<double> temp_x(3), temp_v(3);
      int temp_id, temp_type, temp_mol;

      // read
      if(col_id==0) temp_id = natoms;
      else temp_id = 0;
      temp_type = 1;
      temp_mol = 0;
      for(i=0;i<dim;i++) {
	temp_x[i] = 0.0;
	temp_v[i] = 0.0;
      }
      while(line >> t1) {
	col++;
	if(col==col_id) temp_id = (int)t1 - 1;
	if(col==col_mol) temp_mol = (int)t1 - 1;
	if(col==col_type) temp_type = (int)t1;
	if(col==col_x[0]) temp_x[0] = t1;
	if(col==col_x[1]) temp_x[1] = t1;
	if(col==col_x[2]) temp_x[2] = t1;
	if(col==col_v[0]) temp_v[0] = t1;
	if(col==col_v[1]) temp_v[1] = t1;
	if(col==col_v[2]) temp_v[2] = t1;
      }

      // check
      if(temp_id > expect-1) {
	cerr << "Atom_System::readlammps_custom_dump_frame(): Error: Atom id > expected number of atoms (minus one): " << temp_id << " > " << expect-1 << endl;
	exit(-1);
      }

      // save
      for(i=0;i<dim;i++) {
	x0[dim*temp_id + i] = temp_x[i] + translation[i];
	x1[dim*temp_id + i] = temp_v[i];
	fixx0[dim*temp_id + i] = 0;
      }

      // block
      block[temp_id] = temp_mol;

      // set atom type
      dummy2 = stringify(temp_type);
      atypes.add(dummy2, temp_type);
      atype[temp_id] = temp_type;

      natoms++;
      atypes.natype[temp_type]++;
    }
  }

  // check
  if(expect != natoms) {
    cerr << "Atom_System::readlammps_custom_dump_frame: Expected " << expect << " atoms but found " << natoms << "." << endl;
    exit(-1);
  }

  // report
  cout << "  Read in " << natoms << " atoms" << endl;
  atypes.printtypes();

  return time;
}


void Atom_System::readbgf(string fname) {

  /*
    Reads a bgf-file.

    sets: (maybe?)
    ntypes, atype, comment, atypename, natype, x0, natoms, fixx0, V, K, x1, F, mass, block=0
  */


  cout << "Atom_System::readbgf(): I don't probably work yet!" << endl;
  cerr << "Atom_System::readbgf(): I don't probably work yet!" << endl;
  exit(-1);


  // --- variables
  int i;  // loops
  ifstream afile;  // xyz-file
  //int thisframe;  // current frame

  int expect;  // number of atoms expected
  string dummy;
  //bool ok;  // finished reading atoms?

  // fixed atoms
  //int nfixed;
  //valarray<int> fixd(dim);

  int i1;
  double t1;
  string s1;
  int extrasize = 100;  // add room for this many atoms at once

  // additional columns
  //int c5;
  //bool indexwarned = false;  // warn of index mismatch only once

  valarray<double> pos(dim);  // coordinates of new atom
  //valarray<double> xmax(1.0, dim);


  // --- init
  expect = 100;

  // init arrays and reset stuff
  moreatoms(expect);
  atypes.zero();

  // read atoms
  natoms = 0;
  //nfixed = 0;
  //fixd = 0;


  // --- start reading file

  afile.open(fname.c_str());
  assert(afile);
  //cout << "Atom_System::readbgf(): Reading frame " << frame << " of file " << fname << endl;
  cout << "Atom_System::readbgf(): Reading file " << fname << endl;


  /*
  // --- skip frames if necessary

  // check
  if(frame <= 0) {
    cerr << "Atom_System::readxyz(): Cannot read frame " << frame << endl;
    exit(-1);
  }

  // skip frames
  thisframe = 1;
  while(thisframe < frame) {

    afile >> expect;
    getline(afile, dummy, '\n');  // dummy, next line
    //cout << "  Expecting " << expect << " atoms in frame " << thisframe << endl;
    getline(afile, dummy, '\n');  // comment line
    //cout << "  Comment " << thisframe << ": " << dummy << endl;
    for(i=0;i<expect;i++) {
      getline(afile, dummy, '\n');  // atoms
    }
    thisframe++;
  }
  */


  // --- start reading frame
  // cout << "Atom_System::readbgf():"

  // loop over keywords / lines
  afile >> dummy;
  while(dummy != "END") {

    // bgf-version
    if(dummy=="XTLGRF") {
      afile >> i1;
      if(i1 != 200) {
        cerr << "Atom_System::readbgf(): Only bgf-files of version 200 supported. (Whatever that means!)" << endl;
        exit(-1);
      }
    }

    // description
    else if(dummy=="DESCRP") {
      getline(afile, comment, '\n');
      cout << "Atom_System::readbgf(): Description: " << comment << endl;
    }

    else if(dummy=="CRYSTX") {
      for(i=0;i<dim;i++) {
        afile >> box[i];
      }
      for(i=0;i<dim;i++) {
        afile >> t1;
        if(t1 != 90.0) {
          cerr << "Atom_System::readbgf(): Only orthogonal boxes supported!" << endl;
          exit(-1);
        }
      }
    }

    else if(dummy=="HETATM") {
      afile >> i1;
      afile >> s1;
      for(i=0;i<dim;i++) {
        afile >> pos[i];
      }
      add_atom(pos, s1, extrasize);

      // ignore FF type, two numbers, partial charge
      getline(afile, s1, '\n');
    }

    else {
      // ignore
    }


    // read next keyword
    afile >> dummy;
            
  }  // end of loop over keywords / lines


  // --- report
  /*
  cout << "  Read in " << natoms << " atoms of which " << nfixed << " fixed. (";
  for(i=0;i<dim;i++) {
    cout << fixd[i];
    if(i<dim-1) cout << ", ";
  }
  cout << ")" << endl;
  */

  atypes.printtypes();
}


void Atom_System::report() {

  int i;  // loops

  cout << "The system contains " << natoms << " atom(s) of type(s)" << endl;
  cout << "  ";

  for(i=0;i<atypes.ntypes;i++) {
    cout << "type " << atypes.types[i] << " (" << atypes.atypename[atypes.types[i]]
	 << ", mass " << atypes.mass[atypes.types[i]] <<  "): " << atypes.natype[atypes.types[i]] << endl;
  }
}


void Atom_System::readvels(string fname) {

  /*
    Reads a vels-file as produced by the ReaxFF code.

    // Does not:
    //Translates atoms inside +-box/2 in each direction.

    sets: (maybe?)
    ntypes, atype, comment, atypename, natype, x0, natoms, fixx0, V, K, x1, F, mass, block=0
  */


  // --- variables
  int i, j;  // loops
  ifstream afile;  // xyz-file
  int expect;  // number of atoms expected
  string dummy;
  valarray<double> pos(dim);  // coordinates of new atom
  int type;  // current atom type number;


  // --- set some stuff
  comment = "Read in vels-file";


  // --- open file and read
  afile.open(fname.c_str());
  assert(afile);
  cout << "Atom_System::readvels(): Reading file " << fname << endl;

  // lattice parameters
  getline(afile, dummy, '\n');  // should read: " Lattice parameters:"
  for(i=0;i<dim;i++) {
    afile >> box[i];
  }
  getline(afile, dummy, '\n');  // ingore rest of line
  getline(afile, dummy, '\n');  // cell vector angles, ingoring

  // number of atoms
  afile >> expect;  // number of atoms
  getline(afile, dummy, '\n');  // ignore rest of the line

  // report
  cout << "Atom_System::readvels(): Expecting " << expect << " atoms. Box size: ";
  for(i=0;i<dim;i++) {
    cout << box[i] << " ";
  }
  cout << endl;

  // reset system
  reset(expect);
  atypes.reset();

  // set defaults on fixed atoms etc.
  for(i=0;i<expect;i++) {
    fixx0[i] = 0;
  }

  // read atoms, coordinates, atom types
  natoms = 0;
  for(i=0;i<expect;i++) {
    // coordinates
    for(j=0;j<dim;j++) {
      afile >> x0[dim*natoms + j];
      //x0[dim*natoms + j] -= box[j]/2;
    }
    // atom type
    afile >> dummy;  // atom type (name, like C etc.)
    type = -1;
    atypes.add(dummy, type);
    atype[i] = type;
    atypes.natype[type]++;
    natoms++;
  }
  getline(afile, dummy, '\n');  // ignore rest of the line

  // velocities begin
  getline(afile, dummy, '\n');  // should read: " Atom velocities (Angstrom/s):"

  // read in velocities
  natoms = 0;
  for(i=0;i<expect;i++) {
    for(j=0;j<dim;j++) {
      afile >> x1[dim*natoms + j];
      x1[dim*natoms + j] *= 1e-15;  // units Å/s -> Å/fs
    }
    natoms++;
  }

  // ignore the rest of the file

  // check
  if(natoms != expect) {
    cerr << "Atom_System::readvels(): Error / Bug: Expected " << expect << " atoms but only read in " << natoms << "." << endl;
    exit(-1);
  }


  // --- report
  cout << "Atom_System::readvels(): Read in " << natoms << " atoms." << endl;
  atypes.printtypes();

  return;
}


void Atom_System::interpolate(const Atom_System &second, double weight) {

  int i, j;
  int ind;
  valarray<double> dx(dim);

  if(weight < 0 || weight > 1) {
    cout << "Atom_System::interpolate(): Weight should be in [0,1]." << endl;
    exit(-1);
  }

  cout << "Atom_System::interpolate(): Interpolating with weight "
       << weight << endl;

  for(i=0;i<natoms;i++) {
    for(j=0;j<dim;j++) {
      ind = 3*i + j;

      // fixed degrees of freedom
      if(fixx0[ind] != 0) {
	dx[j] = 0;
	continue;
      }

      dx[j] = second.x0[ind] - x0[ind];

      // periodic b.c.s
      if(pbc[j] == 1 && dx[j] > box[j]/2) {
	dx[j] -= box[j];
      }
      if(pbc[j] == 1 && dx[j] < -box[j]/2) {
	dx[j] += box[j];
      }
    }

    moveatom(i, weight*dx);
  }
}


double Atom_System::rgyr() {

  /*
    Returns the radius of gyration of the system. This is
    apparently defined as

    R^2_g = (1/N) \sum_{i=1}^{N} (R_i - R_{CM}).

    See Wang et al., Phys. Lett. A 310, 197 (2003)

    There has to be a type above. I will use

    R^2_g = \frac{ \sum_{i=1}^{N} m_i (R_i - R_{CM})^2 }{ \sum_{i=1}^{N} m_i }
  */


  // --- variables
  int i, j;  // loops
  valarray<double> cm(dim);  // centre of mass
  double rg;  // radius of gyration
  double m;  // mass of the whole system


  // --- find mass and centre of mass
  m = cmcoord(cm);


  // --- calculate radius of gyration
  rg = 0;
  m = 0;
  for(i=0;i<natoms;i++) {
    m += atypes.mass[atype[i]];
    for(j=0;j<dim;j++) {
      rg += atypes.mass[atype[i]]*x0[3*i + j]*x0[3*i + j];
    }
  }
  rg = rg/m;
  rg = sqrt(rg);

  return rg;
}


void Atom_System::angmom(double &angmom, double &f, double &J, double &TcorrtoL) {

  /*
    Calculate the angular momentum and rotation frequency
    of the system.

    (This is probably wrong! One should have a defined axis
     of rotation! A warning is printed!)

    These are calculated or returned                unit                                             ok?

    angular momentum  L = \sum r \times p
  * angular momentum  L = angmom = |L|              au*Å^2/fs      |L| = J \omega = J 2\pi f
  * moment of inertia J = J = \sum mr^2             au*Å^2
  * frequency of rot. f = f = |L|/(2*pi*J)          1/fs
  * T corr. to L      T = TcorrtoL = J*f^2*1.2e6    K              TcorrtoL = 2*0.5*J*f*f*1.2e6
  */


  cerr << "Atom_System::angmom(): Error / Warning: Axis of rotation not defined properly in current version!" << endl;


  // --- variables
  int i, j;
  double pi = 3.141592653589793;

  valarray<double> L(dim);
  valarray<double> p(dim);
  double Ltot;

  // --- calculate
  J = 0;
  for(i=0;i<natoms;i++) {

    for(j=0;j<dim;j++) {
      p[j] = atypes.mass[atype[i]]*x1[3*i + j];

      J += atypes.mass[atype[i]]*x0[3*i + j]*x0[3*i + j];
    }

    L[0] += x0[3*i + 1]*p[2] - x0[3*i + 2]*p[1];
    L[1] += x0[3*i]*p[2] - x0[3*i + 2]*p[0];
    L[2] += x0[3*i]*p[1] - x0[3*i + 1]*p[0];
  }

  Ltot = 0;
  for(j=0;j<dim;j++) {
    Ltot += L[j]*L[j];
  }
  Ltot = sqrt(Ltot);

  angmom = Ltot;  // au*Å^2/fs
  f = Ltot/(2.0*pi*J);  // 1/fs
  TcorrtoL = 2*0.5*J*f*f*1.2e6;  // K
}


void Atom_System::updateT() {

  /*
    Subroutine to get the temperature of the atoms in the system.
    Fixed (in any direction) atoms are ignored.
  */


  // --- variables
  int  i, j;  // loops
  double  Ekinsum;
  double temp = 0.5*AU*VUNIT*VUNIT/EC;  // factor
  bool fixedat;
  int nfixedat = 0;


  // --- Get sum of kinetic energies
  Ekinsum=0;
  for (i=0;i<natoms;i++) {
    fixedat = false;
    K[i] = 0;
    for(j=0;j<dim;j++) {
      if(fixx0[3*i + j] != 0) {
	fixedat = true;
	nfixedat++;
	K[i] = 0;
	break;
      }
      K[i] += temp*atypes.mass[atype[i]]*x1[3*i + j]*x1[3*i + j];
    }
    Ekinsum += K[i];
  }

  // Ekinsum now in eV
  // Do necessary unit transformations to get E in J.
  // Ekinsum=0.5*AU*Ekinsum*VUNIT*VUNIT;
  // and get temperature using E=3/2 N k T
  T = Ekinsum/(1.5*(natoms-nfixedat)*KBEV);
}


void Atom_System::init_nlist(double cut_pot, double cut_list) {

  // init and update neighbour list and linked list
  // this only allocates linkedlist_head


  // --- variables

  int i;
  int maxneib = 15;


  // --- check
  if(natoms <= 0) {
    cerr << "Atom_System::init_nlist(): natoms = " << natoms << endl;
    exit(-1);
  }


  // --- init variables

  cout << "Initialising Verlet neighbour list and linked list:" << endl;

  if(cells != 0) delete [] linkedlist_head;
  cells = 0;

  nlist_cut = cut_list;
  nlist_potcut = cut_pot;

  if(nlistsize != 0) delete [] nlist;
  nlistsize = maxneib*natoms;
  nlist = new int[nlistsize];

  // set length and number of linked list cells
  for(i=0;i<dim;i++) {

    // choose number of cells
    cells = (int)floor(box[i]/nlist_cut);
    if(cells < 4) {
      cells = 1;
    }

    nsubcell[i] = cells;
    lsubcell[i] = box[i]/cells;
  }

  cells = 1;
  for(i=0;i<dim;i++) {
    cells *= nsubcell[i];
  }

  // resize head-array
  linkedlist_head = new int[cells];

  cout << "  pbc:     " << pbc[0] << " " << pbc[1] << " " << pbc[2] << endl;
  cout << "  box:     " << box[0] << " " << box[1] << " " << box[2] << endl;
  cout << "  cells:   " << nsubcell[0] << " " << nsubcell[1] << " " << nsubcell[2] << endl;
  cout << "  lengths: " << lsubcell[0] << " " << lsubcell[1] << " " << lsubcell[2] << endl;

  maxneighbours = maxneib;
  update_linkedlist();
  update_nlist();
}


void Atom_System::update_linkedlist() {

  /*
    Update the linked list. If boundary conditions are periodic,
    the simulation cell has to be centered at the origin.

    (Works only in 3d for now.)
  */


  // --- variables
  int i, j;
  double dx;  // position with x > 0
  valarray<int> ind(dim);
  int icell;  // subcell of current atom
  int at;
  int check;


  // --- init arrays
  for(i=0;i<atypesize;i++) {
    linkedlist_list[i] = -1;  // -1 indicates that atoms in this cell are over.
  }
  for(i=0;i<cells;i++) {
    linkedlist_head[i] = -1;  // -1 indicates that atoms in this cell are over.
  }


  // --- fill lists

  // loop over atoms
  for(i=0;i<natoms;i++) {

    // loop over coordinates
    for(j=0;j<dim;j++) {
      // check for errors on atom positions with pbc's
      if(pbc[j]==1 && abs(x0[3*i + j]) > box[j]/2.0) {
	cerr << "Atom_System::update_linkedlist(): Error: Atom " << i << " outside periodic boundary conditions at ";
	cerr << x0[3*i] << " " << x0[3*i + 1] << " " << x0[3*i + 2] << endl;
	cerr << "                                  To wrap atoms within periodics, use \"translate 0.0 0.0 0.0\", multiple" << endl;
	cerr << "                                  times if necessary." << endl;
	exit(-1);
      }

      // go on
      dx = x0[3*i + j] + box[j]/2.0;

      // periodic boundary conditions
      if(pbc[j] == 1 && dx < 0.0) dx += box[j];
      if(pbc[j] == 1 && dx > box[j]) dx -= box[j];

      // index of subcell
      ind[j] = (int)floor((dx/box[j])*nsubcell[j]);

      // take into account non-periodic b.c.s or the case dx=box[j] => ind[j]=nsubcell[j].
      if(ind[j] < 0) ind[j] = 0;
      if(ind[j] >= nsubcell[j]) ind[j] = nsubcell[j] - 1;
    }

    // set cell
    icell = (ind[2]*nsubcell[1] + ind[1])*nsubcell[0] + ind[0];
    linkedlist_list[i] = linkedlist_head[icell];
    linkedlist_head[icell] = i;
  }
}


void Atom_System::add_to_linkedlist(int i) {

  /*
    Adds atom at to linked list. Assumes that the linked
    list has been initialized.
  */


  // --- variables
  int j;  // loops
  double dx;
  valarray<int> ind(dim);  // linked list cell indices
  int icell;  // number of linked list cell


  // --- check
  if(i > natoms-1 || i < 0) {
    cerr << "Atom_System::add_to_linkedlist(): Trying to add atom " << i << " but the atoms are [0," << natoms - 1 << "]." << endl;
    exit(-1);
  }


  // --- add atom

  // loop over coordinates
  for(j=0;j<dim;j++) {
    dx = x0[3*i + j] + box[j]/2.0;

    // periodic boundary conditions
    if(pbc[j] == 1 && dx < 0.0) dx += box[j];
    if(pbc[j] == 1 && dx > box[j]) dx -= box[j];

    // index of subcell
    ind[j] = (int)floor((dx/box[j])*nsubcell[j]);

    // take into account non-periodic b.c.s or the case dx=box[j] => ind[j]=nsubcell[j].
    if(ind[j] < 0) ind[j] = 0;
    if(ind[j] >= nsubcell[j]) ind[j] = nsubcell[j] - 1;
  }

  // set cell
  icell = (ind[2]*nsubcell[1] + ind[1])*nsubcell[0] + ind[0];
  linkedlist_list[i] = linkedlist_head[icell];
  linkedlist_head[icell] = i;
}


void Atom_System::update_nlist() {

  // Update neighbour list based on linked list. (Only 3d for now.)

  int i, j, k, l, m;  // loops
  valarray<int> ind(dim);  // indices of subcell of current atom
  valarray<int> indnei(dim);  // indices of current subcell
  double dx, drsq;  // multiple uses
  int at;  // current atom in neighbouring cell
  int icell;  // current neighbouring cell
  double nlistcutsq = nlist_cut*nlist_cut;

  int startnei;  // starting point of neighbours of current atom
  int nnei;  // number of neighbours of current atom
  double nneiave;  // average number of neighbours
  double rminsq;  // for finding minimum distance of two atoms


  // --- reset nlist
  startnei = 0;

  cout << "Updating neighbour list... ";

  // --- go trough atoms in neighbouring subcells using linked list

  // loop over atoms
  nneiave = 0.0;
  rminsq = -1.0;
  for(i=0;i<natoms;i++) {

    nnei = 0;

    // find cell indices
    for(j=0;j<dim;j++) {
      dx = x0[3*i + j] + box[j]/2.0;

      // periodic boundary conditions
      if(pbc[j] == 1 && dx < 0.0) dx += box[j];
      if(pbc[j] == 1 && dx > box[j]) dx -= box[j];

      // index of subcell
      ind[j] = (int)floor((dx/box[j])*nsubcell[j]);

      // take into account non-periodic b.c.s or the case dx=box[j] => ind[j]=nsubcell[j].
      if(ind[j] < 0) ind[j] = 0;
      if(ind[j] >= nsubcell[j]) ind[j] = nsubcell[j] - 1;
    }

    // loop over neighbouring cells
    for(j=ind[0]-1;j<=ind[0]+1;j++) {
      for(k=ind[1]-1;k<=ind[1]+1;k++) {
	for(l=ind[2]-1;l<=ind[2]+1;l++) {

	  // x
	  if(nsubcell[0] == 1 && j != 0) {
	    continue;
	  }
	  else if(nsubcell[0] != 1) {
	    indnei[0] = j;

	    // periodic b.c.s
	    if(pbc[0] == 1 && j < 0) indnei[0] = nsubcell[0]-1;
	    if(pbc[0] == 1 && j > nsubcell[0]-1) indnei[0] = 0;

	    // non-periodic b.c.s
	    if(indnei[0] < 0 || indnei[0] > nsubcell[0] - 1) continue;
	  }
	  // y
	  if(nsubcell[1] == 1 && k != 0) {
	    continue;
	  }
	  else if(nsubcell[1] != 1) {
	    indnei[1] = k;

	    // periodic b.c.s
	    if(pbc[1] == 1 && k < 0) indnei[1] = nsubcell[1]-1;
	    if(pbc[1] == 1 && k > nsubcell[1]-1) indnei[1] = 0;

	    // non-periodic b.c.s
	    if(indnei[1] < 0 || indnei[1] > nsubcell[1] - 1) continue;
	  }
	  // z
	  if(nsubcell[2] == 1 && l != 0) {
	    continue;
	  }
	  else if(nsubcell[2] != 1) {
	    indnei[2] = l;

	    // periodic b.c.s
	    if(pbc[2] == 1 && l < 0) indnei[2] = nsubcell[2]-1;
	    if(pbc[2] == 1 && l > nsubcell[2]-1) indnei[2] = 0;

	    // non-periodic b.c.s
	    if(indnei[2] < 0 || indnei[2] > nsubcell[2] - 1) continue;
	  }

	  // subcell
	  icell = (indnei[2]*nsubcell[1] + indnei[1])*nsubcell[0] + indnei[0];

	  // loop over atoms in cell
	  at = linkedlist_head[icell];
	  while(true) {
	    if(at==-1) break;  // atoms gone through

	    // don't pair up atoms with themselves
	    if(i==at) {
	      at = linkedlist_list[at];
	      continue;
	    }

	    // take each pair into account only once
	    if(i < at) {
	      at = linkedlist_list[at];
	      continue;
	    }

	    drsq = 0.0;
	    for(m=0;m<dim;m++) {
	      dx = abs(x0[3*i + m] - x0[3*at + m]);

	      // periodic b.c.s
	      if(pbc[m] == 1 && dx > box[m]/2.0) {
		dx = box[m] - dx;
	      }

	      drsq += dx*dx;
	    }

	    // add atom to neighbour list
	    if(drsq < nlistcutsq) {
	      nnei++;
	      nneiave += 1.0;

	      // find mimimum distance between two atoms
	      if(rminsq < 0.0 || drsq < rminsq) {
		rminsq = drsq;
	      }

	      // check that there is enough room
	      if(startnei + nnei >= nlistsize) {
		maxneighbours += 10;
		cout << "  resizing (maxneib=" << maxneighbours << ")...";
		nlist = grow(nlist, nlistsize, natoms*maxneighbours);
		nlistsize = natoms*maxneighbours;
	      }

	      nlist[startnei + nnei] = at;
	    }  // atom added

	    at = linkedlist_list[at];
	  }  // end of loop over atoms in cell

	}  // end of...
      }  // loops over...
    }  // neighbouring subcells

    // write number of neighbours to start of list for this atom
    nlist[startnei] = nnei;
    startnei += nnei + 1;

  }  // end of loop over atoms

  // zero dx0
  for(i=0;i<x0size;i++) {
    dx0[i] = 0.0;
  }

  nneiave /= natoms/2.0;  // half because of half loop
  cout << " done. Average neighbours, mimimum distance: "
       << nneiave << " " << sqrt(rminsq) << endl;

}


bool Atom_System::update_already() {

  int i, j;
  double drmax, drmax2;
  double dr;

  drmax = -1;
  drmax2 = -2;

  for(i=0;i<natoms;i++) {

    dr = 0;
    for(j=0;j<dim;j++) {
      dr += dx0[3*i + j]*dx0[3*i + j];
    }
    dr = sqrt(dr);

    if(dr > drmax) {
      drmax2 = drmax;
      drmax = dr;
    }
    else if(dr > drmax2) {
      drmax2 = dr;
    }
  }

  if(drmax + drmax2 > nlist_cut - nlist_potcut) return true;
  else return false;
}


double Atom_System::Vtot() {

  int i;
  double vtot = 0;

  for(i=0;i<natoms;i++) {
    vtot += V[i];
  }

  return vtot;
}


void Atom_System::check_nlist() {


  int i, j, startnei, at;

  cout << "Atom_System::check_nlist(): Checking..." << endl;

  // loop over atoms
  startnei = 0;
  for(i=0;i<natoms;i++) {

    cout << "  atom " << i << " (" << nlist[startnei] << "): ";
    // loop over neighbour list
    for(j=startnei+1;j<=startnei + nlist[startnei];j++) {
      at = nlist[j];
      cout << at << " ";
    }  // end of loop over neighbours
    cout << endl;
    startnei += nlist[startnei] + 1;
  }  // end of loop over atoms
}


void Atom_System::set_temperature(double temp, double tdebye) {

  /*
    Give initial velocities to non-fixed atoms corresponding to
    a temperature of "temp" K. The total center-of-mass velocity
    is zeroed. Give also initial displacements according to
    "temp" and "tdebye".

    -The displacement distribution is a gaussian
    distribution of the form

    \rho(x) = (2 \pi \sigma)^{-0.5} exp{ -x^2 / (2 \sigma^2) },

    with dsigma = 20.89/T_Debye[K] * sqrt{ T[K] / (3A[au]) }

    -The velocity is distributed according to the Maxwell-
     Boltzmann distribution

     \rho(v_{i}^{a}) = ( m_a / (2 \pi k_B T ) )^(-0.5)
                       exp{ -0.5 m_a (v^a_i)^2 / (k_B T) },

     where a signifies the atom and i the dimension. This can be
     done with

     x1[3*i + j] = vrms*gaussian,

     where vrms = sqrt(k_B * 2 * T / (m * au))/VUNIT. The 2 comes
     along because of the equipartition theorem.
  */


  int i, j;
  double vrms;
  double sigmaf;
  valarray<double> dr(dim);

  if(temp <= 0) return;

  cout << "Atom_System()::set_temperature(): Giving a temperature of "
       << temp << " K." << endl;

  // loop over atoms
  for(i=0;i<natoms;i++) {

    vrms = sqrt(KB*2*temp/(atypes.mass[atype[i]]*AU))/VUNIT;
    sigmaf = (20.89/(tdebye/KBEV))*sqrt( temp / (3.0*atypes.mass[atype[i]]) );

    for(j=0;j<dim;j++) {
      if(fixx0[3*i + j] == 1) continue;
      dr[j] = gaussian(sigmaf);
      x1[3*i + j] = vrms*gaussian();
    }
    moveatom(i, dr);

  }  // end of loop over atoms


  /*
  if (initemp > 0.0) then
            tdebye_K=tdebye*invkBeV
	    ! 20.89 is sqrt(9 hbar^2/k_B u) 
	    ! Expression comes from second order approximation of a nasty
	    ! elliptic integral
            sd=20.89*sqrt(initemp*invkBeV/                                &
		 mass(abs(atype)))/(sqrt(3.0)*tdebye_K)
	    
            x=x+sd*gasdev(0)/box(1)
            y=y+sd*gasdev(0)/box(2)
            z=z+sd*gasdev(0)/box(3)
         endif
  */
}


int Atom_System::find_closest_atom(const valarray<double> &pos, double &r) {

  /*
    Find closest atom to position pos(3). Below in the comments
    "current atom" translates to "pos". Return index of atom or
    -1 if nothing found within the adjacent linked list cells.

    Uses linked list to go through the neighbouring linked list
    cells.
  */


  // --- variables

  int j, k, l, m;  // loops
  valarray<int> ind(dim);  // indices of subcell of current atom
  valarray<int> indnei(dim);  // indices of current subcell
  double dx, drsq;  // multiple uses
  int at;  // current atom in neighbouring cell
  int icell;  // current neighbouring cell

  double rclosestsq = 0;  // squared distance of closest atom
  int closest = -1;  // index of closest atom


  // --- check
  if(pos.size() != dim) {
    cerr << "Atom_System::find_closest_atom(): Position vector size not dim." << endl;
    exit(-1);
  }


  // --- check for the case of no atoms around
  if(natoms==0) {
    return -1;
  }


  // --- go trough atoms in neighbouring subcells using linked list

  // find cell indices
  for(j=0;j<dim;j++) {
    dx = pos[j] + box[j]/2;

    // periodic boundary conditions
    if(pbc[j] == 1 && dx < 0.0) dx += box[j];
    if(pbc[j] == 1 && dx > box[j]) dx -= box[j];

    // index of subcell
    ind[j] = (int)floor((dx/box[j])*nsubcell[j]);

    // take into account non-periodic b.c.s or the case dx=box[j] => ind[j]=nsubcell[j].
    if(ind[j] < 0) ind[j] = 0;
    if(ind[j] >= nsubcell[j]) ind[j] = nsubcell[j] - 1;
  }  // cell indices found

  // loop over neighbouring cells
  for(j=ind[0]-1;j<=ind[0]+1;j++) {
    for(k=ind[1]-1;k<=ind[1]+1;k++) {
      for(l=ind[2]-1;l<=ind[2]+1;l++) {

	// x
	if(nsubcell[0] == 1 && j != 0) {
	  continue;
	}
	else if(nsubcell[0] != 1) {
	  indnei[0] = j;

	  // periodic b.c.s
	  if(pbc[0] == 1 && j < 0) indnei[0] = nsubcell[0]-1;
	  if(pbc[0] == 1 && j > nsubcell[0]-1) indnei[0] = 0;

	  // non-periodic b.c.s
	  if(indnei[0] < 0 || indnei[0] > nsubcell[0] - 1) continue;
	}
	// y
	if(nsubcell[1] == 1 && k != 0) {
	  continue;
	}
	else if(nsubcell[1] != 1) {
	  indnei[1] = k;

	  // periodic b.c.s
	  if(pbc[1] == 1 && k < 0) indnei[1] = nsubcell[1]-1;
	  if(pbc[1] == 1 && k > nsubcell[1]-1) indnei[1] = 0;

	  // non-periodic b.c.s
	  if(indnei[1] < 0 || indnei[1] > nsubcell[1] - 1) continue;
	}
	// z
	if(nsubcell[2] == 1 && l != 0) {
	  continue;
	}
	else if(nsubcell[2] != 1) {
	  indnei[2] = l;

	  // periodic b.c.s
	  if(pbc[2] == 1 && l < 0) indnei[2] = nsubcell[2]-1;
	  if(pbc[2] == 1 && l > nsubcell[2]-1) indnei[2] = 0;

	  // non-periodic b.c.s
	  if(indnei[2] < 0 || indnei[2] > nsubcell[2] - 1) continue;
	}

	// --- number of current subcell
	icell = (indnei[2]*nsubcell[1] + indnei[1])*nsubcell[0] + indnei[0];

	// --- loop over atoms in cell
	at = linkedlist_head[icell];
	while(true) {

	  if(at==-1) break;  // atoms gone through

	  // calculate drsq
	  drsq = 0;
	  for(m=0;m<dim;m++) {
	    dx = abs(pos[m] - x0[3*at + m]);

	    // periodic b.c.s
	    if(pbc[m] == 1 && dx > box[m]/2) {
	      dx = box[m] - dx;
	    }

	    drsq += dx*dx;
	  }

	  // take atom if it's closer
	  if(drsq < rclosestsq || closest == -1) {
	    closest = at;
	    rclosestsq = drsq;
	  }

	  // move to next atom in cell
	  at = linkedlist_list[at];
	}  // end of loop over atoms in cell

      }  // end of...
    }  // loops over...
  }  // neighbouring subcells

  r = sqrt(rclosestsq);

#if defined(DEBUG)
  // report
  cout << "Found closest atom: i r " << closest << " " << r << endl;
#endif

  return closest;  // this is -1 if nothing found
}


//int Atom_System::find_closest_atom(const valarray<double> &pos, double &r) {
int Atom_System::find_closest_atom(const int i , double &r) {

  /*
    Find closest atom to atom i. Return index of atom or
    -1 if nothing found within the adjacent linked list cells.

    Uses linked list to go through the neighbouring linked list
    cells.
  */


  // --- variables

  int j, k, l, m;  // loops
  valarray<int> ind(dim);  // indices of subcell of current atom
  valarray<int> indnei(dim);  // indices of current subcell
  valarray<double> pos(dim);  // position of atom i
  double dx, drsq;  // multiple uses
  int at;  // current atom in neighbouring cell
  int icell;  // current neighbouring cell

  double rclosestsq = 0;  // squared distance of closest atom
  int closest = -1;  // index of closest atom


  // --- check
  if(i < 0 || i >= natoms) {
    cerr << "Atom_System::find_closest_atom(): Asking for negative atom index or higher than number of atoms." << endl;
    exit(-1);
  }


  // --- check for the case of no atoms around
  if(natoms==0) {
    return -1;
  }


  // --- go trough atoms in neighbouring subcells using linked list

  pos[0] = x0[3*i + 0];
  pos[1] = x0[3*i + 1];
  pos[2] = x0[3*i + 2];

  // find cell indices
  for(j=0;j<dim;j++) {
    dx = pos[j] + box[j]/2;

    // periodic boundary conditions
    if(pbc[j] == 1 && dx < 0.0) dx += box[j];
    if(pbc[j] == 1 && dx > box[j]) dx -= box[j];

    // index of subcell
    ind[j] = (int)floor((dx/box[j])*nsubcell[j]);

    // take into account non-periodic b.c.s or the case dx=box[j] => ind[j]=nsubcell[j].
    if(ind[j] < 0) ind[j] = 0;
    if(ind[j] >= nsubcell[j]) ind[j] = nsubcell[j] - 1;
  }  // cell indices found

  // loop over neighbouring cells
  for(j=ind[0]-1;j<=ind[0]+1;j++) {
    for(k=ind[1]-1;k<=ind[1]+1;k++) {
      for(l=ind[2]-1;l<=ind[2]+1;l++) {

	// x
	if(nsubcell[0] == 1 && j != 0) {
	  continue;
	}
	else if(nsubcell[0] != 1) {
	  indnei[0] = j;

	  // periodic b.c.s
	  if(pbc[0] == 1 && j < 0) indnei[0] = nsubcell[0]-1;
	  if(pbc[0] == 1 && j > nsubcell[0]-1) indnei[0] = 0;

	  // non-periodic b.c.s
	  if(indnei[0] < 0 || indnei[0] > nsubcell[0] - 1) continue;
	}
	// y
	if(nsubcell[1] == 1 && k != 0) {
	  continue;
	}
	else if(nsubcell[1] != 1) {
	  indnei[1] = k;

	  // periodic b.c.s
	  if(pbc[1] == 1 && k < 0) indnei[1] = nsubcell[1]-1;
	  if(pbc[1] == 1 && k > nsubcell[1]-1) indnei[1] = 0;

	  // non-periodic b.c.s
	  if(indnei[1] < 0 || indnei[1] > nsubcell[1] - 1) continue;
	}
	// z
	if(nsubcell[2] == 1 && l != 0) {
	  continue;
	}
	else if(nsubcell[2] != 1) {
	  indnei[2] = l;

	  // periodic b.c.s
	  if(pbc[2] == 1 && l < 0) indnei[2] = nsubcell[2]-1;
	  if(pbc[2] == 1 && l > nsubcell[2]-1) indnei[2] = 0;

	  // non-periodic b.c.s
	  if(indnei[2] < 0 || indnei[2] > nsubcell[2] - 1) continue;
	}

	// --- number of current subcell
	icell = (indnei[2]*nsubcell[1] + indnei[1])*nsubcell[0] + indnei[0];

	// --- loop over atoms in cell
	at = linkedlist_head[icell];
	while(true) {

	  if(at==-1) break;  // atoms gone through

	  if(at != i) {
	    // calculate drsq
	    drsq = 0;
	    for(m=0;m<dim;m++) {
	      dx = abs(pos[m] - x0[3*at + m]);

	      // periodic b.c.s
	      if(pbc[m] == 1 && dx > box[m]/2) {
		dx = box[m] - dx;
	      }

	      drsq += dx*dx;
	    }

	    // take atom if it's closer
	    if(drsq < rclosestsq || closest == -1) {
	      closest = at;
	      rclosestsq = drsq;
	    }
	  }

	  // move to next atom in cell
	  at = linkedlist_list[at];
	}  // end of loop over atoms in cell

      }  // end of...
    }  // loops over...
  }  // neighbouring subcells

  r = sqrt(rclosestsq);

#if defined(DEBUG)
  // report
  cout << "Found closest atom: i r " << closest << " " << r << endl;
#endif

  return closest;  // this is -1 if nothing found
}


void Atom_System::moreatoms(int n) {

  /*
    Make room for n more atoms, keep all data
    these arrays should always be resized here
  */


  // --- variables
  int i;  // loops


  // --- report
  cout << "Atom_System::moreatoms(): Making room for additional atoms: n dim " << n << " " << dim << endl;


  // resize
  x0 = grow(x0, x0size, x0size+n*dim);
  dx0 = grow(dx0, x0size, x0size+n*dim);
  x1 = grow(x1, x0size, x0size+n*dim);
  F = grow(F, x0size, x0size+n*dim);
  fixx0 = grow(fixx0, x0size, x0size+n*dim);
  x0size += n*dim;

  atype = grow(atype, atypesize, atypesize+n);
  V = grow(V, atypesize, atypesize+n);
  K = grow(K, atypesize, atypesize+n);
  linkedlist_list = grow(linkedlist_list, atypesize, atypesize+n);
  block = grow(block, atypesize, atypesize+n);
  atypesize += n;
}


void Atom_System::reset(int n) {

  /*
    Reset system size to n atoms.
    Reset also atom types with reset().
  */


  cout << "Atom_System::reset(): Resetting system size to: n dim " << n << " " << dim << endl;

  x0 = grow(x0, x0size, n*dim);
  dx0 = grow(dx0, x0size, n*dim);
  x1 = grow(x1, x0size, n*dim);
  F = grow(F, x0size, n*dim);
  fixx0 = grow(fixx0, x0size, n*dim);
  x0size = n*dim;

  atype = grow(atype, atypesize, n);
  V = grow(V, atypesize, n);
  K = grow(K, atypesize, n);
  block = grow(block, atypesize, n);
  linkedlist_list = grow(linkedlist_list, atypesize, n);
  atypesize = n;

  atypes.reset();
}


void Atom_System::add_atom(valarray<double> pos, string name, int extrasize) {

  /*
    Adds an atom of type "name" at "pos". Assume zero force,
    velocity, fixing, etc.
    (resize arrays and set variables: x0, x1, F, fixx0, natoms, atype-stuff)

    Whenever array sizes need to be increased, add room for "extrasize" atoms.
  */


  // --- variables
  int i;  // loops
  valarray<double> tempd;
  valarray<int> tempi;
  int temp;


  // --- check
  if(pos.size() != dim) {
    cerr << "Atom_System::add_atom(): Position vector size not dim." << endl;
    exit(-1);
  }


  // --- resize arrays if needed
  if(atypesize<natoms+1) {
    moreatoms(extrasize);
  }

  // --- add atom
  for(i=0;i<dim;i++) {
    x0[dim*natoms + i] = pos[i];
    x1[dim*natoms + i] = 0;
    F[dim*natoms + i] = 0;
    fixx0[dim*natoms + i] = 0;
  }

  temp = -1;
  atypes.add(name, temp);
  atype[natoms] = temp;

  // --- natoms
  natoms++;
  atypes.natype[temp]++;
}


void Atom_System::update_symmnlist() {

  /*
    Updates the symmetric neighbour list assuming the linked list is up to date.
    Uses the same cutoff as the Verlet list.

    Initialises the list if it has not been used before.

    symmnlist[i][0]: number of neighbours of atom i
    symmnlist[i][j]: index of neighbour j of atom i
  */

  // --- variables

  int i, j, k, l, m;  // loops
  valarray<int> ind(dim);  // indices of subcell of current atom
  valarray<int> indnei(dim);  // indices of current subcell
  double dx, drsq;  // multiple uses
  int at;  // current atom in neighbouring cell
  int icell;  // current neighbouring cell
  double nlistcutsq = nlist_cut*nlist_cut;

  int nnei;  // number of neighbours of current atom
  int maxneib = 15;
  int addsize = 10;
  double nneiave;  // check average number of neighbours
  valarray<int> temp;


  // --- output and check array sizes
  cout << "Updating symmetric neighbour list... (cutoff " << nlist_cut << ", ";

  if(symmnlist.size() < natoms) {
    symmnlist.resize(natoms);
    for(i=0;i<natoms;i++) {
      symmnlist[i].resize(maxneib + 1);
    }
  }


  // --- go through atoms in neighbouring subcells using linked list

  // loop over atoms
  for(i=0;i<natoms;i++) {

    nnei = 0;

    // find cell indices
    for(j=0;j<dim;j++) {
      dx = x0[3*i + j] + box[j]/2.0;

      // periodic boundary conditions
      if(pbc[j] == 1 && dx < 0.0) dx += box[j];
      if(pbc[j] == 1 && dx > box[j]) dx -= box[j];

      // index of subcell
      ind[j] = (int)floor((dx/box[j])*nsubcell[j]);

      // take into account non-periodic b.c.s or the case dx=box[j] => ind[j]=nsubcell[j].
      if(ind[j] < 0) ind[j] = 0;
      if(ind[j] >= nsubcell[j]) ind[j] = nsubcell[j] - 1;
    }

    // loop over neighbouring cells
    for(j=ind[0]-1;j<=ind[0]+1;j++) {
      for(k=ind[1]-1;k<=ind[1]+1;k++) {
	for(l=ind[2]-1;l<=ind[2]+1;l++) {

	  // x
	  if(nsubcell[0] == 1 && j != 0) {
	    continue;
	  }
	  else if(nsubcell[0] != 1) {
	    indnei[0] = j;

	    // periodic b.c.s
	    if(pbc[0] == 1 && j < 0) indnei[0] = nsubcell[0]-1;
	    if(pbc[0] == 1 && j > nsubcell[0]-1) indnei[0] = 0;

	    // non-periodic b.c.s
	    if(indnei[0] < 0 || indnei[0] > nsubcell[0] - 1) continue;
	  }
	  // y
	  if(nsubcell[1] == 1 && k != 0) {
	    continue;
	  }
	  else if(nsubcell[1] != 1) {
	    indnei[1] = k;

	    // periodic b.c.s
	    if(pbc[1] == 1 && k < 0) indnei[1] = nsubcell[1]-1;
	    if(pbc[1] == 1 && k > nsubcell[1]-1) indnei[1] = 0;

	    // non-periodic b.c.s
	    if(indnei[1] < 0 || indnei[1] > nsubcell[1] - 1) continue;
	  }
	  // z
	  if(nsubcell[2] == 1 && l != 0) {
	    continue;
	  }
	  else if(nsubcell[2] != 1) {
	    indnei[2] = l;

	    // periodic b.c.s
	    if(pbc[2] == 1 && l < 0) indnei[2] = nsubcell[2]-1;
	    if(pbc[2] == 1 && l > nsubcell[2]-1) indnei[2] = 0;

	    // non-periodic b.c.s
	    if(indnei[2] < 0 || indnei[2] > nsubcell[2] - 1) continue;
	  }

	  // subcell
	  icell = (indnei[2]*nsubcell[1] + indnei[1])*nsubcell[0] + indnei[0];

	  // loop over atoms in cell
	  at = linkedlist_head[icell];
	  while(true) {
	    if(at==-1) break;  // atoms gone through

	    // don't pair up atoms with themselves
	    if(i==at) {
	      at = linkedlist_list[at];
	      continue;
	    }

	    drsq = 0;
	    for(m=0;m<dim;m++) {
	      dx = abs(x0[3*i + m] - x0[3*at + m]);

	      // periodic b.c.s
	      if(pbc[m] == 1 && dx > box[m]/2.0) {
		dx = box[m] - dx;
	      }

	      drsq += dx*dx;
	    }

	    // add atom to neighbour list
	    if(drsq < nlistcutsq) {
	      nnei++;

	      // check that there is enough room
	      if(symmnlist[i].size() < nnei+1) {
		temp.resize(symmnlist[i].size());
		temp = symmnlist[i];
		symmnlist[i].resize(symmnlist[i].size()+addsize);
		for(m=0;m<temp.size();m++) {
		  symmnlist[i][m] = temp[m];
		}
	      }  // resized

	      // add neighbour
	      symmnlist[i][nnei] = at;
#if defined(DEBUG)
              cout << "Atom_System::update_symmnlist(): Added atom neighbour: " << i << " " << at << endl;
#endif
              // check again that an atom is not marked its own neighbour
              if(i==at) {
                cerr << "Atom_System::update_symmnlist(): Error: Atom neighbour: " << i << " " << at << endl;
                exit(-1);
              }
	    }  // atom added

	    at = linkedlist_list[at];
	  }  // end of loop over atoms in cell

	}  // end of...
      }  // loops over...
    }  // neighbouring subcells

    // write number of neighbours to start of list for this atom
    symmnlist[i][0] = nnei;
  }  // end of loop over atoms

  // zero dx0
  for(i=0;i<x0size;i++) {
    dx0[i] = 0;
  }

  // check average number of neighbours
  nneiave = 0;
  for(i=0;i<natoms;i++) {
    nneiave += symmnlist[i][0];
  }
  nneiave /= natoms;
  cout << "average neighbours " << nneiave << ") ";

  cout << "done." << endl;


#if defined(DEBUG)
  for(i=0;i<natoms;i++) {
    for(j=1;j<=symmnlist[i][0];j++) {
      at = symmnlist[i][j];
      if(i==at) {
        cerr << "Atom_System::update_symmnlist(): Error:" << endl;
        cerr << i << " " << symmnlist[i][0] << ": ";
        for(k=1;k<=symmnlist[i][0];k++) {
          cerr << symmnlist[i][k] << " ";
        }
        cerr << endl;
      }
    }
  }
#endif
}


void Atom_System::translate(const valarray<double> &dr, bool movefixed) {

  /*
    Moves all atoms by dr using moveatom().
  */


  // --- variables
  int i;

  // --- move atoms
  //     (This is not the most efficient way though!)
  for(i=0;i<natoms;i++) {
    moveatom(i, dr, movefixed);
  }

  return;
}


void Atom_System::center() {

  /*
    Move the geometric center of the system to
    the origin. (Quit if there are periodic
    boundary conditions.)

    Move also fixed atoms.
  */


  // --- variables
  int i, j;  // loops
  valarray<double> centre(dim);
  bool check;  // check


  // --- check
  check = false;
  for(i=0;i<dim;i++) {
    if(pbc[i] != 0) check = true;
  }
  if(check) {
    cerr << "Atom_System::center(): Error: Trying to center system with periodic boundary conditions." << endl;
    exit(-1);
  }


  // --- find centre
  for(j=0;j<dim;j++) {
    centre[j] = 0;
  }
  for(i=0;i<natoms;i++) {
    for(j=0;j<dim;j++) {
      centre[j] += x0[3*i + j];
    }
  }
  for(j=0;j<dim;j++) {
    centre[j] /= natoms;
  }


  // --- translate
  cout << "Atom_System::center(): Moving geometric center to origin from " << centre[0] << " " << centre[1] << " " << centre[2] << " (Å)." << endl;
  translate(-centre, true);
}


valarray <double> Atom_System::cmvel() {

  /*
    Get center of mass velocity
  */


  // --- variables
  int i, j;  // loops
  valarray<double> centre(dim);
  double w;


  // --- find centre
  w = 0;
  for(j=0;j<dim;j++) {
    centre[j] = 0;
  }
  for(i=0;i<natoms;i++) {
    for(j=0;j<dim;j++) {
      centre[j] += x1[3*i + j]*atypes.mass[atype[i]];
    }
    w += atypes.mass[atype[i]];
  }
  for(j=0;j<dim;j++) {
    centre[j] /= w;
  }

  return centre;
}


void Atom_System::centerv() {

  /*
    Set center of mass velocity to zero
  */


  // --- variables
  int i, j;  // loops
  valarray<double> centre(dim);
  double w;


  // --- find centre
  w = 0;
  for(j=0;j<dim;j++) {
    centre[j] = 0;
  }
  for(i=0;i<natoms;i++) {
    for(j=0;j<dim;j++) {
      centre[j] += x1[3*i + j]*atypes.mass[atype[i]];
    }
    w += atypes.mass[atype[i]];
  }
  for(j=0;j<dim;j++) {
    centre[j] /= w;
  }


  // --- set v to zero
  cout << "Atom_System::centerv(): Setting cm-velocity to zero from " << centre[0] << " " << centre[1] << " " << centre[2] << " Å/fs." << endl;
  for(i=0;i<natoms;i++) {
    for(j=0;j<dim;j++) {
      x1[3*i + j] -= centre[j];
    }
  }
}


double Atom_System::cmekin() {

  /*
    Gives the center of mass kinetic energy of the atoms.
  */


  // --- variables
  int i, j;  // loops
  valarray<double> centre(dim);  // centre of mass velocity
  double w;  // total weight of atoms
  double E;  // result


  // --- find centre of mass velocity

  // reset
  w = 0;
  for(j=0;j<dim;j++) {
    centre[j] = 0;
  }

  // calculate
  for(i=0;i<natoms;i++) {
    for(j=0;j<dim;j++) {
      centre[j] += x1[3*i + j]*atypes.mass[atype[i]];
    }
    w += atypes.mass[atype[i]];
  }
  for(j=0;j<dim;j++) {
    centre[j] /= w;
  }


  // kinetic energy
  E = 0;
  for(i=0;i<dim;i++) {
    E += centre[i]*centre[i];
  }
  E *= 0.5*w;  // E = 0.5 m v^2 [au*Å/fs]

  E *= AU*VUNIT*VUNIT/EC;  // E in eV

  return E;
}


double Atom_System::cmcoord(valarray<double> &centre) {

  /*
    Gives the center of mass coordinates of the system.
    Returns the total mass.
  */


  // --- variables
  int i, j;  // loops
  double w;  // total weight of atoms


  // --- check
  if(centre.size() != dim) {
    cerr << "Atom_System::cmcoord(): centre.size() dim " << centre.size() << " " << dim << endl;
    exit(-1);
  }


  // --- find centre of mass

  // reset
  w = 0;
  for(j=0;j<dim;j++) {
    centre[j] = 0;
  }

  // calculate
  for(i=0;i<natoms;i++) {
    for(j=0;j<dim;j++) {
      centre[j] += x0[3*i + j]*atypes.mass[atype[i]];
    }
    w += atypes.mass[atype[i]];
  }
  for(j=0;j<dim;j++) {
    centre[j] /= w;
  }


  return w;
}


void Atom_System::unit_l(double factor) {

  /*
    Unit of lenght: 1.0 -> 1.0*factor
  */

  cout << "Atom_System::unit_l(): Scaling coordinates, velocities, forces and box size to new unit of length. l: 1 -> " << factor << "." << endl;

  for(int i=0;i<natoms;i++) {
    for(int j=0;j<dim;j++) {
      x0[dim*i + j] *= factor;
      x1[dim*i + j] *= factor;
      F[dim*i + j] *= factor;
    }
  }
  for(int i=0;i<dim;i++) {
    box[i] *= factor;
  }
}


void Atom_System::unit_t(double factor) {

  /*
    Unit of time: 1.0 -> 1.0*factor
  */

  cout << "Atom_System::unit_l(): Scaling velocities and forces to new unit of time. t: 1 -> " << factor << "." << endl;

  for(int i=0;i<natoms;i++) {
    for(int j=0;j<dim;j++) {
      x1[dim*i + j] /= factor;
      F[dim*i + j] /= factor*factor;
    }
  }
}


void Atom_System::scalec(double factor) {

  /*
    Scale x0 and dx0 by a factor. Don't touch anything else.
  */


  // --- variables
  int i;  // loops


  // --- report
  cout << "Atom_System::scalec(): Scaling atom coordinates (not velocities, box size, etc.) by " << factor << "." << endl;


  // --- scale
  for(i=0;i<x0size;i++) {
    x0[i] *= factor;
    dx0[i] *= factor;
  }
}


void Atom_System::rotate_axis(double phi, double theta, double delta) {

  /*
    Rotates the system through angle delta around an axis with angles (phi, theta).
    Equations are from Korn & Korn: Mathematical Handbook for Scientists and Engineers.

    (The rotated variables are x0, x1, and F.)

    All angles in radians.
  */


  // --- variables

  int i;  // loops
  double xrot, yrot, zrot;  // rotated coordinates

  // helps
  double a = 1 - cos(delta);
  double b = sin(delta);
  double c = 1 - a;

  // axis of rotation
  double axis_x = sin(theta)*cos(phi);
  double axis_y = sin(theta)*sin(phi);
  double axis_z = cos(theta);


  // --- rotate

  for(i=0;i<natoms;i++) {

    xrot = x0[3*i + 0]*(c + a*axis_x*axis_x) + x0[3*i + 1]*(a*axis_x*axis_y - b*axis_z) + x0[3*i + 2]*(a*axis_x*axis_z + b*axis_y);
    yrot = x0[3*i + 0]*(a*axis_y*axis_x + b*axis_z) + x0[3*i + 1]*(c + a*axis_y*axis_y) + x0[3*i + 2]*(a*axis_y*axis_z - b*axis_x);
    zrot = x0[3*i + 0]*(a*axis_z*axis_x - b*axis_y) + x0[3*i + 1]*(a*axis_z*axis_y + b*axis_x) + x0[3*i + 2]*(c + a*axis_z*axis_z);
    x0[3*i + 0] = xrot;
    x0[3*i + 1] = yrot;
    x0[3*i + 2] = zrot;

    xrot = x1[3*i + 0]*(c + a*axis_x*axis_x) + x1[3*i + 1]*(a*axis_x*axis_y - b*axis_z) + x1[3*i + 2]*(a*axis_x*axis_z + b*axis_y);
    yrot = x1[3*i + 0]*(a*axis_y*axis_x + b*axis_z) + x1[3*i + 1]*(c + a*axis_y*axis_y) + x1[3*i + 2]*(a*axis_y*axis_z - b*axis_x);
    zrot = x1[3*i + 0]*(a*axis_z*axis_x - b*axis_y) + x1[3*i + 1]*(a*axis_z*axis_y + b*axis_x) + x1[3*i + 2]*(c + a*axis_z*axis_z);
    x1[3*i + 0] = xrot;
    x1[3*i + 1] = yrot;
    x1[3*i + 2] = zrot;

    xrot = F[3*i + 0]*(c + a*axis_x*axis_x) + F[3*i + 1]*(a*axis_x*axis_y - b*axis_z) + F[3*i + 2]*(a*axis_x*axis_z + b*axis_y);
    yrot = F[3*i + 0]*(a*axis_y*axis_x + b*axis_z) + F[3*i + 1]*(c + a*axis_y*axis_y) + F[3*i + 2]*(a*axis_y*axis_z - b*axis_x);
    zrot = F[3*i + 0]*(a*axis_z*axis_x - b*axis_y) + F[3*i + 1]*(a*axis_z*axis_y + b*axis_x) + F[3*i + 2]*(c + a*axis_z*axis_z);
    F[3*i + 0] = xrot;
    F[3*i + 1] = yrot;
    F[3*i + 2] = zrot;
  }
}


void Atom_System::rotate_euler(double phi, double theta, double psi) {

  /*
    Rotates the system about the origin by the Euler-angles
    phi, theta, psi according to the y-convention as explained
    in mathworld.wolfram.com.

    The ranges are:
    phi   [0,2pi]
    theta [0,pi]
    psi   [0,2pi]

    If theta is chosen such that theta = acos(u[-1,1]), where
    u[-1,1] is a uniform random number in the range [-1,1], then
    at least a unit vector will point to a uniformly random point
    on the unit sphere after the rotation. The third degree of
    freedom has not been tested.

    (The rotated variables are x0, x1, and F.)
  */


  // --- variables
  int i, j, k;

  double pi = 3.14159265359;

  double newx[3];

  double a[3][3];
  a[0][0] = cos(psi)*cos(theta)*cos(phi)-sin(psi)*sin(phi);
  a[0][1] = cos(psi)*cos(theta)*sin(phi)+sin(psi)*cos(phi);
  a[0][2] = -cos(psi)*sin(theta);
  a[1][0] = -sin(psi)*cos(theta)*cos(phi)-cos(psi)*sin(phi);
  a[1][1] = -sin(psi)*cos(theta)*sin(phi)+cos(psi)*cos(phi);
  a[1][2] = sin(psi)*sin(theta);
  a[2][0] = sin(theta)*cos(phi);
  a[2][1] = sin(theta)*sin(phi);
  a[2][2] = cos(theta);


  // --- rotate
  for(i=0;i<natoms;i++) {

    // x0
    for(j=0;j<dim;j++) {
      newx[j] = 0;
      for(k=0;k<dim;k++) {
	newx[j] += a[j][k]*x0[3*i + k];
      }
    }

    for(j=0;j<dim;j++) {
      x0[3*i + j] = newx[j];
    }

    // x1
    for(j=0;j<dim;j++) {
      newx[j] = 0;
      for(k=0;k<dim;k++) {
	newx[j] += a[j][k]*x1[3*i + k];
      }
    }

    for(j=0;j<dim;j++) {
      x1[3*i + j] = newx[j];
    }

    // F
    for(j=0;j<dim;j++) {
      newx[j] = 0;
      for(k=0;k<dim;k++) {
	newx[j] += a[j][k]*F[3*i + k];
      }
    }

    for(j=0;j<dim;j++) {
      F[3*i + j] = newx[j];
    }
  }
  // --- done rotating
}


double Atom_System::smartsro(double cut, int typea, int typeb) {

  /*
    Calculate the short range order parameter

    SRO = 1 - (n_BA / n_B)/c_A,

    where:

    n_BA         average number of neighbours of type A of atoms of type B
    n_B          average number of neighbours of atoms of type B
    c_A          concentration of atoms of type A

    Note that n_BA/n_B = sro(*, B, A, *, *).

    NOTE THE STUPID SOLUTION TO REMOVE ZMIN, ZMAX!

    Example, Cu3Au: smartsro(*, Au, Au)

    SRO -> 1     ordered phase, Au atoms far from each other
    SRO ~= 0     random alloy
    SRO -> -3    segregation, Au atoms gather together
  */


  // --- variables
  double eta;
  double ca;


  // --- calculate
  eta = sro(cut, typeb, typea, -99999999.0, 99999999.0);
  ca = (double)atypes.natype[typea]/(double)natoms;
  cout << "Atom_System::smartsro(): eta n_A natoms ca " << eta << " " << atypes.natype[typea] << " " << natoms << " " << ca << endl;


  return 1 - eta/ca;
}


double Atom_System::sro(double cut, int atomtype, int neitype, double zmin, double zmax) {

  /*
    Calculate the short range order parameter

    \eta = n_1 / n_2,

    where n_1 is the average number of neighbours of type neitype
    and n_2 the average total number of neighbours of atoms of
    type atomtype.

    Uses the symmetric neighbour list:
      symmnlist[i][0]: number of neighbours of atom i
      symmnlist[i][j]: index of neighbour j of atom i

    Only takes into account atoms between zmin, zmax.
  */


  // --- variables
  int i, j;  // loops
  double eta;  // sro-parameter
  int n1, n2;  // see above
  int at;  // current neighbour
  double dx;  // distance of the atoms


  cout << "Atom_System::sro(): Calculating SRO for atoms-neighbour pair " << atypes.atypename[atomtype]
       << "-" << atypes.atypename[neitype] << "." << endl;

  // --- calculate
  n1 = 0;
  n2 = 0;

  // loop over atoms
  for(i=0;i<natoms;i++) {

    // only take atoms between zmin, zmax
    if(x0[3*i + 2] < zmin || x0[3*i + 2] > zmax) continue;

    // only take atoms of type atomtype
    if(atype[i] != atomtype) continue;

    // loop over neighbours
    for(j=1;j<=symmnlist[i][0];j++) {
      at = symmnlist[i][j];

      // check
      if(i==at) {
        cerr << "Atom_System::sro(): Error: This should not happen: atom type neighbour type dx cut " << i << " " << atype[i] << " " << at << " " << atype[at] << " " << dx << " " << cut << endl;
        exit(-1);
      }

      dx = dr(i, at);  // get distance of the atoms

      if(dx > cut) continue;

      // count
      n2++;
      if(atype[at]==neitype) {
        n1++;
        //cout << i << " " << atype[i] << " " << at << " " << atype[at] << " " << dx << " " << cut << endl;
      }
    }  // end of loop over neighbours
  }  // end of loop over atoms


  // calculate sro-parameter
  if(n2==0) {
    cout << "Atom_System::sro(): Average number of neighbours of atoms of type " << atypes.atypename[atomtype]
         << " (" << atomtype << ") is zero." << endl;
    eta = 0;
  }
  else {
    eta = (double)n1/(double)n2;
  }

  return eta;
}


double Atom_System::localsro(int atomtype, int neitype, double cuta, double cutb) {

  /*
    Calculates for each atom the local sro-parameter

    1 - f_BA/c_A,

    where

    f_BA         fraction of A-neighbours of atoms of type B
    c_A          concentration of atoms of type A


    Example, Cu3Au: localsro(Au, Au, *, *)

    SRO -> 1     ordered phase, Au atoms far from each other
    SRO ~= 0     random alloy
    SRO -> -3    segregation, Au atoms gather together

    Gives 0 for atoms with no neighbours.
  */


  // --- variables
  int i;  // loops
  int nt;  // check


  // --- check
  if(cutb - cuta <= 0) {
    cerr << "Atom_System::localsro(): Error: Neighbours are trying to be found at distance [" << cuta << ", " << cutb << endl;
    exit(-1);
  }


  // --- report
  cout << "Atom_System::localsro(): Calculating local SRO-parameter for atoms of type " << atypes.atypename[atomtype] << endl;
  cout << "Atom_System::localsro(): Neighbour type, distance [min,max] " << atypes.atypename[neitype] << " " << cuta << " " << cutb << endl;


  // --- loop over atoms
  nt = 0;
  for(i=0;i<=natoms;i++) {
    if(atype[i]==atomtype) {
      if(sronei(i, neitype, cuta, cutb) < 0) {
	V[i] = 0;
      }
      else {
	V[i] = 1 - sronei(i, neitype, cuta, cutb)/((double)atypes.natype[neitype]/(double)natoms);
      }
      nt++;
    }
  }


  // --- check
  if(nt != atypes.natype[atomtype]) {
    cerr << "Atom_System::localsro(): Error: Gone through " << nt << " atoms of type " << atypes.atypename[atomtype] << " but there should be " << atypes.natype[atomtype] << endl;
    exit(-1);
  }
}


double Atom_System::sronei(int i, int neitype, double cuta, double cutb) {

  /*
    Calculate the short range order parameter that gives the
    fraction of neitype-neighbours of atom i.

    Uses the symmetric neighbour list:
      symmnlist[i][0]: number of neighbours of atom i
      symmnlist[i][j]: index of neighbour j of atom i

    Only takes into account atoms that are at a distance [cuta,cutb]
    from atom i.

    Returns a negative value for an atom with no neighbours.
  */


  // --- variables
  int j;  // loops
  double eta;  // sro-parameter
  int at;  // current neighbour
  double dx;  // distance of the atoms

  //cout << "check" << endl;

  // --- check
  if(cutb - cuta <= 0) {
    cerr << "Atom_System::sro(): Error: Neighbours are trying to be found at distance [" << cuta << ", " << cutb << "] from atom " << i << endl;
    exit(-1);
  }


  // --- loop over neighbours
  eta = 0;
  for(j=1;j<=symmnlist[i][0];j++) {
      at = symmnlist[i][j];

      // check
      if(i==at) {
        cerr << "Atom_System::sro(): Error: This should not happen: atom type neighbour type dx cuta cutb " << i << " " << atype[i] << " " << at << " " << atype[at] << " " << dx << " " << cuta << " " << cutb << endl;
        exit(-1);
      }

      // count
      if(atype[at]==neitype) {
	dx = dr(i, at);  // get distance of the atoms
	if(dx > cutb || dx < cuta) continue;
        eta = eta + 1.0;
      }
  }  // end of loop over neighbours


  // --- calculate sro-parameter
  if(symmnlist[i][0]==0) {
    eta = -1;
  }
  else {
    eta = eta/(double)symmnlist[i][0];
  }

  //cout << "ok" << endl;

  return eta;
}


void Atom_System::cluster(double cutoff) {

  /*
    A cluster analysis algorithm.

    Groups the atoms to clusters with atoms closer than cutoff from
    each other belonging to the same group.

    Uses the symmetric neighbour list.
    sets: int* block[atypesize];                     // [0,atoms-1], which cluster the atom belongs to
          valarray<int> blocknat(number of blocks);  // number of atoms in different blocks

    The cluster an atom belongs to is contained in the array
    block:

    value           meaning
    ------------------------------
    -1              not assigned
    [0,inf[         cluster number
  */
  

  // --- variables
  int i, j, k;  // loops
  int nchanges;  // changes in the clustering on this round (stop when zero)
  int thiscl;  // number of the current cluster
  int removecl;  // cluster to be assimilated
  int ncl;  // number of clusters


  // --- checks
  if(cutoff > nlist_cut || cutoff <= 0) {
    cerr << "Atom_System::cluster(): Error: cutoff, nlist_cut " << cutoff << " " << nlist_cut << endl;
    exit(-1);
  }

  if(natoms==0) {
    cerr << "Atom_System::cluster(): Error: natoms =  " << natoms << endl;
    exit(-1);
  }


  // --- report
  cout << "Atom_System::cluster(): Clustering " << natoms << " atoms with cutoff " << cutoff << "..." << endl;


  // --- init
  for(i=0;i<atypesize;i++) {
    block[i] = -1;
  }
  nchanges = 1;
  ncl = 0;


  // --- cluster
  // --- the cluster numbers will be [0,inf[

  while(nchanges > 0) {
    nchanges = 0;

    // loop over atoms
    for(i=0;i<natoms;i++) {

      // find out if there are marked atoms in this
      // neighbourhood
      thiscl = -1;

      if(block[i] != -1) {
        thiscl = block[i];
      }
      else {
        // loop over neighbours of current atom
        for(j=1;j<=symmnlist[i][0];j++) {
          if(block[symmnlist[i][j]] != -1) {
            thiscl = block[symmnlist[i][j]];
            break;
          }
        }  // end of loop over neighbours of current atom
      }

      // --- mark all atoms to the same cluster if one of them is
      // --- already assigned to one
      if(thiscl != -1) {

        // mark unmarked atoms
        if(block[i] == -1) {  // this one
          block[i] = thiscl;
          nchanges++;
        }
        for(j=1;j<=symmnlist[i][0];j++) {  // its neighbours
          if(block[symmnlist[i][j]] == -1) {
            block[symmnlist[i][j]] = thiscl;
            nchanges++;
          }
        }  // unmarked atoms marked

        /*
        cout << "--- " << thiscl << " " << block[i] << " ";
        for(k=1;k<=symmnlist[i][0];k++) {  // its neighbours
          cout << block[symmnlist[i][k]] << " ";
        }
        cout << endl;
        */

        // join other adjacent clusters to this one
        for(j=0;j<=symmnlist[i][0];j++) {
          if(j==0) removecl = block[i];
          else removecl = block[symmnlist[i][j]];

          // check
          if(removecl == -1) {
            cerr << "Atom_System::cluster(): Error: This should not happen!" << endl;
            exit(-1);
          }

          // check whether there is a cluster to be removed, if so, remove
          if(removecl != thiscl) {

            /*
            cout << "Removing cluster " << removecl << " " << thiscl << endl;
            cout << block[i] << " ";
            for(k=1;k<=symmnlist[i][0];k++) {  // its neighbours
              cout << block[symmnlist[i][k]] << " ";
            }
            cout << endl;
            */

            // mark atoms in the removed cluster to the this cluster
            for(k=0;k<natoms;k++) {
              if(block[k]==removecl) {
                block[k] = thiscl;
                nchanges++;
              }
            }

            // if the number of the removed cluster was not
            // the last one, use it for the current last cluster
            if(removecl != ncl-1) {
              // replace cluster number ncl-1 by removecl
              for(k=0;k<natoms;k++) {
                if(block[k]==ncl-1) block[k] = removecl;
              }

              if(thiscl==ncl-1) thiscl = removecl;
            }

            ncl--;
          }  // cluster removed
        }  // adjacent clusters joined
      }  // all atoms in the neighbourhood marked to same cluster


      // --- no atoms already in cluster => new cluster
      else {
        nchanges++;

        /*
        cout << "Adding cluster " << ncl << endl;
        cout << block[i] << " ";
        for(k=1;k<=symmnlist[i][0];k++) {  // its neighbours
          cout << block[symmnlist[i][k]] << " ";
        }
        cout << endl;
        */

        block[i] = ncl;  // this atom

        // neighbours
        for(j=1;j<=symmnlist[i][0];j++) {

          // check
          if(block[symmnlist[i][j]] != -1) {
            cerr << "Atom_System::cluster(): Error: This should not happen!" << endl;
            exit(-1);
          }

          block[symmnlist[i][j]] = ncl;
        }

        ncl++;
      }  // new cluster made

    }  // end of loop over atoms

  }  // end of rounds


  // --- report
  cout << "...done. Found " << ncl << " clusters." << endl;


  // --- There are at this point ncl clusters with
  // --- numbers [0,ncl-1]

  // find number of atoms in each cluster
  blocknat.resize(ncl);
  blocknat = 0;
  for(i=0;i<natoms;i++) {
    // check
    if(block[i]==-1) {
      cerr << "Atom_System::cluster(): Error / Bug: block[" << i << "] = " << block[i] << " but all atoms should be clustered." << endl;
      exit(-1);
    }
    blocknat[block[i]]++;
  }

  cout << "Atom_System::cluster(): Found " << ncl << " clusters with a total of " << blocknat.sum() << " atoms." << endl;
}


double Atom_System::sfactor_znorm(double zlimit, valarray<int> n) {

  /*
    Gives the structure factor of the atoms with z > zlimit normalised
    to the structure factor of those with z < zlimit.
  */


  // --- variables
  int i;  // loops
  valarray<int> incl(natoms);  // included atoms (see sfactor())
  double Sa, Sb;  // structure factors, above, below


  // --- atoms above zlimit
  for(i=0;i<natoms;i++) {
    if(x0[3*i + 2] >= zlimit) incl[i] = 1;
    else incl[i] = 0;
  }
  Sa = sfactor(n, incl);


  // --- atoms below zlimit
  for(i=0;i<natoms;i++) {
    if(x0[3*i + 2] < zlimit) incl[i] = 1;
    else incl[i] = 0;
  }
  Sb = sfactor(n, incl);


  // --- finished
  cout << "Atom_System::sfactor_znorm(): Calculated structure factor with zlimit, Sa, Sb, Sa/Sb: " << zlimit << " " << Sa << " " << Sb << " " << Sa/Sb << endl;

  return Sa/Sb;
}


double Atom_System::sfactor(valarray<int> n) {

  /*
    Calculates the structure factor for the entire system.
    See sfactor(valarray<int> n, valarray<int> incl).
  */


  // --- check
  if(natoms <= 1) {
    cerr << "Atom_System::sfactor(): Error: Trying to calculate structure factor for " << natoms << " atom(s)." << endl;
    exit(-1);
  }


  // --- calculate
  valarray<int> temp(1, natoms);
  return sfactor(n, temp);
}


double Atom_System::sfactor(valarray<int> n, valarray<int> incl) {

  /*
    Calculates the structure factor for the atoms i with
    incl[i]==1. n gives the Miller indices.

    Returns |S|^2, where

    S = (1/N) \sum_{i=0}^{natoms-1} exp(i k \cdot x_i)


    Need to be set: a
  */


  // --- variables
  int i, j;  // loops
  int nat;  // number of included atoms
  valarray<double> k(dim);  // wave vector
  gsl_complex S;
  double Ssq;
  gsl_complex temp;
  double kdx;


  /*
  // Miller indices for different lattices
  // ----------------------------------------
  int n_cubic = 13;
  int nx_cubic[13] = {1, 0, 0, 0, 1, 1, 1, 2, 0, 0, 3, 1, 2};
  int ny_cubic[13] = {0, 1, 0, 1, 0, 1, 1, 0, 2, 0, 1, 1, 2};
  int nz_cubic[13] = {0, 0, 1, 1, 1, 0, 1, 0, 0, 2, 1, 2, 2};
  int int_fcc[13]  = {0, 0, 0, 0, 0, 0, 4, 4, 4, 4, 4, 0, 4};
  */


  // --- check
  if(n.size() != dim) {
    cerr << "Atom_System::sfactor(): dim != n_Miller.size " << dim << " " << n.size() << endl;
    exit(-1);
  }
  if(incl.size() != natoms) {
    cerr << "Atom_System::sfactor(): natoms != include.size " << natoms << " " << incl.size() << endl;
    exit(-1);
  }


  // --- init

  // wave vectors
  cout << "Atom_System::sfactor(): Calculating structure factor using lattice constant a = " << a << endl;
  for(i=0;i<dim;i++) {
    k[i] = 2*M_PI*n[i]/a;
  }

  // reset structure factor
  S = gsl_complex_rect(0,0);


  // --- calculate

  // loop over atoms
  nat = 0;
  for(i=0;i<natoms;i++) {

    // take only atoms with incl[i]==1
    if(incl[i] != 1) continue;

    nat++;  // take this atom

    kdx = 0;
    for(j=0;j<dim;j++) {
      kdx += k[j]*x0[3*i + j];
    }
    temp = gsl_complex_rect(0, kdx);
    temp = gsl_complex_exp(temp);
    S = gsl_complex_add(S, temp);

  }  // end of loop over atoms


  // --- output
  cout << "  Atom_System::sfactor(): Processed " << nat << " atoms." << endl;


  S = gsl_complex_div_real(S, (double)nat);
  Ssq = gsl_complex_abs2(S);

  cout << "  Atom_System::sfactor(): |S|^2 = " << Ssq << endl;

  return Ssq;
}


double Atom_System::paircorr(double r, double shell) {

  /*
    Value of the pair correlation function at r with
    shell thickness shell.

    The calculation does not use neighbour lists. Also,
    it is specific to dimension due to the volume of the
    spherical shell.

    According to www.physics.emory.edu/~weeks/idl/gofr2.html,
    the result should be divided by the average number density
    of the atoms to get g(r)=1 for a system with no structure.
    I didn't implement that since the density calculation depends
    on the boundary conditions.
  */


  // --- variables
  int i, j;  // loops
  int n;  // number of particles in shell
  double d;  // distance of two particles
  double result;  // result
  double rmin, rmax;  // shell limits


  // --- check
  if(dim != 3) {
    cerr << "Atom_System::paircorr(): Error: You should implement me for d != 3." << endl;
    exit(-1);
  }


  // --- calculate

  n = 0;
  if(r - shell < 0) {
    rmin = 0;
  }
  else {
    rmin = r - shell;
  }
  rmax = r + shell;


  // loop over atoms
  for(i=0;i<natoms;i++) {

    // loop over neighbours, every pair once
    for(j=0;j<i;j++) {

      d = dr(i,j);
      if(d < rmax && d > rmin) {
        n += 2;  // 2 <-> divide by natoms in the end, not with number of pairs
      }
    }  // end of loop over neighbours
  }  // end of loop over atoms

  // normalise
  result = (double)n/(double)natoms;
  result /= (4.0/3.0)*M_PI*( rmax*rmax*rmax - rmin*rmin*rmin );  // volume of the shell in 3d

  return result;
}


gsl_complex Atom_System::y_lm(valarray<double> dx, int l, int m) {

  /*
    Gives the spherical harmonic Y_{lm}(dx).

    Tested for dx = (1,1,1), (1,1,0), (0,0,1), (1,0,0), (0,1,0).
    Gives the same results as the equations at
    http://mathworld.wolfram.com/SphericalHarmonic.html.
  */


  // --- variables
  int i, j;  // loops
  gsl_complex ylm;  // result
  double costheta, cosphi, sinphi, rho, r;  // coordinates


  // --- checks
  if(abs(m) > l || l < 0) {
    cerr << "Atom_System::y_lm(): l m: " << l << " " << m << endl;
    exit(-1);
  }
  if(dx.size() != 3) {
    cerr << "Atom_System::y_lm(): dim should be 3, now " << dim << endl;
    exit(-1);
  }


  // --- calculate

  rho = sqrt(dx[0]*dx[0] + dx[1]*dx[1]);

  if(rho==0) {
    cosphi = 1;
    sinphi = 0;
  }
  else {
    cosphi = dx[0]/rho;
    sinphi = dx[1]/rho;
  }

  r = sqrt(dx[0]*dx[0] + dx[1]*dx[1] + dx[2]*dx[2]);
  costheta = dx[2]/r;

  ylm = gsl_complex_rect(pow(-1.0, (double)m), 0.0);
  if(m >= 0) {
    ylm = gsl_complex_mul_real(ylm, gsl_sf_legendre_sphPlm(l, m, costheta));
    ylm = gsl_complex_mul(ylm, gsl_complex_pow_real(gsl_complex_rect(cosphi, sinphi), (double)m));
  }
  else {
    ylm = gsl_complex_mul_real(ylm, gsl_sf_legendre_sphPlm(l, -m, costheta)*pow(-1.0, -m));
    //ylm = gsl_complex_mul_real(ylm, gsl_sf_legendre_sphPlm(l, -m, costheta)*pow(-1.0, -m)*gsl_sf_fact(l+m)/gsl_sf_fact(l-m));
    ylm = gsl_complex_mul(ylm, gsl_complex_pow_real(gsl_complex_rect(cosphi, sinphi), (double)-m));
  }


  return ylm;
}


gsl_complex Atom_System::q_lm(int at, int l, int m) {

  /*
    Calculate q_lm for atom at.

    q_lm is defined as in Chu01a:
    Yaroslav G. Chushak and Lawrence S. Bartell,
    Melting and Freezing of Gold Nanoclusters,
    J. Phys. Chem. B 105, 11605 (2001)

    q_{lm}(i) = (1/N_neighbours) \Sigma_{j=1}^{j<N_neighbours} Y_{lm}(r_{ij})
  */


  // --- variables
  int i;  // loops
  gsl_complex q;  // result
  double w;  // weight of each neighbour
  double wsum;  // sum of weights


  // --- checks
  if(abs(m) > l || l < 0) {
    cerr << "Atom_System::q_lm(): l m: " << l << " " << m << endl;
    exit(-1);
  }


  // --- calculate

  // loop over neighbours
  q = gsl_complex_rect(0.0, 0.0);
  wsum = 0;
  for(i=1;i<=symmnlist[at][0];i++) {
    w = 1.0;
    wsum += w;
    q = gsl_complex_add(q, gsl_complex_mul_real(y_lm(dri(at, symmnlist[at][i]), l, m), w));
  }

  // normalise
  if(symmnlist[at][0] > 0) {
    q = gsl_complex_div_real(q, wsum);
  }
  else {
    q = gsl_complex_rect(0.0, 0.0);
  }


  return q;
}


double Atom_System::q_l(int at, int l) {

  /*
    Calculate q_l for atom at.

    q_l is defined as in Chu01a:
    Yaroslav G. Chushak and Lawrence S. Bartell,
    Melting and Freezing of Gold Nanoclusters,
    J. Phys. Chem. B 105, 11605 (2001)

    q_l(i) = sqrt( (4pi)/(2l+1) \Sigma_{m=-l}^{l} | q_{lm}(i) |^2 )
  */


  // --- variables
  int m;  // loop
  double q;  // result


  // --- calculate

  // loop over m
  q = 0.0;
  for(m=-l;m<=l;m++) {
  //    for(m=0;m<=l;m++) {
  //for(m=0;m<=0;m++) {
    q += gsl_complex_abs2(q_lm(at, l, m));
  }  // end of loop over m

  q *= 4.0*M_PI/(2.0*(double)l + 1.0);
  q = sqrt(q);

  return q;
}


gsl_complex Atom_System::Q_lm(int l, int m) {

  /*
    Get the global order parameter \bar{Q}_${l,m}$ as
    described in Chu01a:
    Yaroslav G. Chushak and Lawrence S. Bartell,
    Melting and Freezing of Gold Nanoclusters,
    J. Phys. Chem. B 105, 11605 (2001)
  */


  // --- variables
  int i;  // loops
  gsl_complex q;  // result
  int totnei;  // sum of number of neighbours of all atoms


  // --- calculate
  q = gsl_complex_rect(0, 0);
  totnei = 0;
  for(i=0;i<natoms;i++) {
    q = gsl_complex_add(q, gsl_complex_mul_real(q_lm(i, l, m), (double)symmnlist[i][0]));
    totnei += symmnlist[i][0];
  }
  q = gsl_complex_div_real(q, (double)totnei);


  return q;
}


double Atom_System::Q_l(int l) {

  /*
    Get the global order parameter \bar{Q}_$l$ as
    described in Chu01a:
    Yaroslav G. Chushak and Lawrence S. Bartell,
    Melting and Freezing of Gold Nanoclusters,
    J. Phys. Chem. B 105, 11605 (2001)
  */


  // --- variables
  int i;  // loops
  double q;  // result


  // --- calculate
  q = 0;
  for(i=-l;i<=l;i++) {
    q += gsl_complex_abs2(Q_lm(l, i));
  }
  q *= 4*M_PI/(2.0*l + 1.0);
  q = sqrt(q);


  return q;
}


double Atom_System::q_correlation(int i, int j, int l) {

  /*
    Calculate the q-correlation factor c_ij for atoms i and j.

    c_{ij} = \Sigma_{m=-l}{l} q_{lm}(i) q_{lm}^{*}(j),

    where * means complex conjugation.

    See Men07a for details:
    Eduardo Mendez-Villuendas and Richard K. Bowles
    Surface Nucleation in the Freezing of Gold Nanoparticles
    Phys. Rev. Lett. 98, 185503 (2007)
  */


  // --- variables
  int m;  // loop
  gsl_complex c;  // result


  // --- calculate

  // loop over m
  c = gsl_complex_rect(0, 0);
  for(m=-l;m<=l;m++) {
    c = gsl_complex_add(c, gsl_complex_mul(q_lm(i, l, m), gsl_complex_conjugate(q_lm(j, l, m))));
  }  // end of loop over m


  return gsl_complex_abs(c);
}


void Atom_System::getmoved(double rcut, Atom_System other) {

  /*
    Compares atoms to the ones in another system and removes
    the ones that have moved by less than rcut.

    DOES NOT UPDATE NEIGHBOUR LISTS!

    This is not very memory-efficient since it uses a local
    copy of the other system in order not to lose the
    correspondence between atom indices between the two
    systems.
  */


  // --- variables
  int i;  // loops
  int nrem;  // number of removed atoms


  // --- report
  cout << "Atom_System::getmoved(): Removing atoms (now " << natoms << ") moved by less than " << rcut << " compared to another system..." << endl;


  // --- remove atoms

  // loop over atoms
  nrem = 0;
  for(i=0;i<natoms;i++) {

    if(dr(i, other, i) < rcut) {
      removeat(i);
      other.removeat(i);
      i--;
      nrem++;
    }
  }

  cout << "Atom_System::getmoved(): ...done. Removed " << nrem << " atoms." << endl;
}


void Atom_System::removetype(string name) {

  /*
    Removes atoms of a certain type.
  */


  removetype(atypes.typenumber(name));
}


void Atom_System::removetype(int type) {

  /*
    Removes atoms of a certain type.
  */


  // --- variables
  int i;  // loops
  int nrem;  // number of removed atoms


  // --- report
  cout << "Atom_System::removetype(): Removing atoms (now " << natoms << ") of type " << type << " (" << atypes.name(type) << ")..." << endl;


  // --- remove atoms

  // loop over atoms
  nrem = 0;
  for(i=0;i<natoms;i++) {

    if(atype[i]==type) {
      removeat(i);
      i--;
      nrem++;
    }
  }

  cout << "Atom_System::removetype(): ...done. Removed " << nrem << " atoms." << endl;
}


void Atom_System::removeK(double ka, double kb) {

  /*
    Remove atoms with kinetic energy (K[i]) in ]a,b[.
  */


  // --- variables
  int i;  // loops
  int nrem;  // number of removed atoms


  // --- report
  cout << "Atom_System::removeK(): Removing atoms (now " << natoms << ") with E_kin in [";
  cout << ka << ", " << kb << "]." << endl;


  // --- check
  if(kb <= ka) {
    cerr << "Atom_System::removeK(): Error: Invalid E_kin interval: " << ka << " " << kb << endl;
    exit(-1);
  }


  // --- remove atoms

  // loop over atoms
  nrem = 0;
  for(i=0;i<natoms;i++) {

    if(K[i] > ka && K[i] < kb) {
      removeat(i);
      i--;
      nrem++;
    }
  }


  cout << "Atom_System::removeK(): ...done. Removed " << nrem << " atoms." << endl;
}


void Atom_System::removerand(int n) {

  /*
    Removes atoms randomly.
  */


  // --- variables
  int i;  // loops
  int at;  // atom to remove


  // --- report
  cout << "Atom_System::removerand(): Removing " << n << " atoms randomly." << endl;


  // --- remove atoms

  // loop over removed atoms
  for(i=0;i<n;i++) {
    at = (int)floor(genrand()*(double)natoms);
    removeat(at);
  }
}


void Atom_System::removeV(double a, double b) {

  /*
    Removes atoms with V[i] in ]a,b[
  */


  // --- variables
  int i;  // loops
  int nrem;  // number of removed atoms


  // --- check
  if(b <= a) {
    cerr << "Atom_System::removeV(): Error: range cannot be ]" << a << ", " << b << "[." << endl;
    exit(-1);
  }


  // --- report
  cout << "Atom_System::removeV(): Removing atoms (now " << natoms << ") with V in ]" << a << ", " << b << "[." << endl;


  // --- remove atoms

  // loop over atoms
  nrem = 0;
  for(i=0;i<natoms;i++) {

    if(V[i]>a && V[i]<b) {
      removeat(i);
      i--;
      nrem++;
    }
  }

  cout << "Atom_System::removeV(): ...done. Removed " << nrem << " atoms." << endl;
}


void Atom_System::remove(int dir, bool above, double cut) {

  /*
    Removes atoms below / above a value in x/y/z.
  */


  // --- variables
  int i;  // loops
  int nrem;  // number of removed atoms


  // --- check
  if(dir >= dim) {
    cerr << "Atom_System::remove(): dir dim " << dir << " " << dim << endl;
    exit(-1);
  }


  // --- report
  if(above) cout << "Atom_System::remove(): Removing atoms (now " << natoms << ") above x[" << dir << "] = " << cut << endl;
  else cout << "Atom_System::remove(): Removing atoms (now " << natoms << ") below x[" << dir << "] = " << cut << endl;


  // --- remove atoms

  // loop over atoms
  nrem = 0;
  for(i=0;i<natoms;i++) {

    if(above) {
      if(x0[dim*i + dir] > cut) {
        removeat(i);
        i--;
        nrem++;
      }
    }
    else {
      if(x0[dim*i + dir] < cut) {
        removeat(i);
        i--;
        nrem++;
      }
    }
  }

  cout << "Atom_System::remove(): ...done. Removed " << nrem << " atoms." << endl;
}


int Atom_System::largestblock() {

  /*
    Get number of largest block of atoms
  */


  // --- variables
  int i;  // loops
  int largest;  // index of largest cluster
  int nlargest;  // number of atoms in largest cluster


  // --- check
  if(blocknat.size() < 1) {
    cerr << "Atom_System::largestblock(): Do clustering first." << endl;
    exit(-1);
  }


  // --- report
  cout << "Atom_System::largestblock(): Finding largest block. Assuming clustering has been done properly." << endl;


  // --- find largest
  nlargest = -1;
  largest = -1;
  for(i=0;i<blocknat.size();i++) {
    if(blocknat[i] > nlargest) {
      nlargest = blocknat[i];
      largest = i;
    }
  }


  return largest;
}


void Atom_System::leaveblock(int bl) {

  /*
    Remove atoms not in block bl.
  */


  // --- variables
  int i;  // loops


  // --- check
  if(bl > blocknat.size()) {
    cerr << "Atom_System::leaveblock(): bl blocks " << bl << " " << blocknat.size() << endl;
    exit(-1);
  }


  // --- report
  cout << "Only leaving block " << bl << " with " << blocknat[bl] << " atoms." << endl;


  // --- remove
  for(i=0;i<natoms;i++) {
    if(block[i] != bl) {
      removeat(i);
      i--;
    }
  }
}


void Atom_System::removeblock(int bl) {

  /*
    Remove atoms in block bl.
  */


  // --- variables
  int i;  // loops


  // --- check
  if(bl > blocknat.size()) {
    cerr << "Atom_System::removeblock(): bl blocks " << bl << " " << blocknat.size() << endl;
    exit(-1);
  }


  // --- report
  cout << "Atom_System::removeblock(): Removing block " << bl << " with " << blocknat[bl] << " atoms." << endl;


  // --- remove
  for(i=0;i<natoms;i++) {
    if(block[i] == bl) {
      removeat(i);
      i--;
    }
  }
}


void Atom_System::removehalfspace(valarray<double> base, valarray<double> dir) {

  /*
    Remove atoms that have

    (r - base) \cdot dir > 0

    (dir will be the surface normal)
  */


  // --- variables
  int i, j;  // loops
  double dot;  // dot product

  // --- check
  if(base.size() != dim  || dir.size() != dim) {
    cerr << "Atom_System::removehalfspace(): base.size() dir.size() " << base.size() << " " << dir.size() << endl;
    exit(-1);
  }


  // --- report
  cout << "Atom_System::removehalfspace(): Removing a half space." << endl;


  // --- remove
  for(i=0;i<natoms;i++) {
    dot = 0;
    for(j=0;j<dim;j++) {
      dot += (x0[dim*i+j] - base[j])*dir[j];
    }

    if(dot > 0) {
      removeat(i);
      i--;
    }
  }
}


void Atom_System::removeat(int at, int mode) {

  /*
    Removes an atom from the system by moving it to the last
    array position and substracting 1 from natoms. Also updates
    numbers of atoms of different types.

    DOES NOT UPDATE NEIGHBOUR LISTS!

    mode
    1        replace removed atom by last atom (this should be the default in atomsystem.h -> don't change!)
    2        keep order of atoms
  */


  // --- variables
  int i, j;  // loops
  valarray<double> tempda(dim);  // temp
  valarray<int> tempia(dim);  // temp
  double tempd;  // temp
  int tempi;  // temp
  int lastat = natoms - 1;  // last atom


  // --- fix atom type and block count
  atypes.natype[atype[at]]--;
  if(blocknat.size() > block[at]) blocknat[block[at]]--;


  // --- remove if this is the last atom
  if(at==natoms-1) {
    natoms--;
    return;
  }


  // --- remove otherwise
  if(mode==1) {

    // arrays of size x0size
    for(i=0;i<dim;i++) {
      x0[dim*at + i] = x0[dim*lastat + i];
      x1[dim*at + i] = x1[dim*lastat + i];
      F[dim*at + i] = F[dim*lastat + i];
      fixx0[dim*at + i] = fixx0[dim*lastat + i];
      dx0[dim*at + i] = dx0[dim*lastat + i];
    }

    // arrays of size atypesize
    atype[at] = atype[lastat];
    V[at] = V[lastat];
    K[at] = K[lastat];
    block[at] = block[lastat];
  }
  else if(mode==2) {

    for(j=at;j<lastat;j++) {

      // arrays of size x0size
      for(i=0;i<dim;i++) {
	x0[dim*j + i] = x0[dim*(j+1) + i];
	x1[dim*j + i] = x1[dim*(j+1) + i];
	F[dim*j + i] = F[dim*(j+1) + i];
	fixx0[dim*j + i] = fixx0[dim*(j+1) + i];
	dx0[dim*j + i] = dx0[dim*(j+1) + i];
      }

      // arrays of size atypesize
      atype[j] = atype[j+1];
      V[j] = V[j+1];
      K[j] = K[j+1];
      block[j] = block[j+1];
    }

  }
  else {
    cerr << "Atom_System::removeat(): Unknown mode " << mode << endl;
    exit(-1);
  }

  // natoms
  natoms--;


  return;
}


double Atom_System::dist_par_z(double x, double y) {

  /*
    Get distance from line (x,y) || z to closest atom.

    Works only in 3d.
  */


  // --- variables
  int i;  // loops
  valarray<double> line(dim);  // closest point on line
  int closest;  // closest atom
  double r;  // distance


  // --- search
  // (natoms^2 for now)

  line[0] = x;
  line[1] = y;

  // first atom first
  line[2] = x0[3*0 + 2];
  closest = 0;
  r = dr(0, line);

  // then others
  for(i=1;i<natoms;i++) {
    line[2] = x0[3*i + 2];

    if(dr(i, line) < r) {
      closest = i;
      r = dr(i, line);
    }
  }


  return r;
}


bool Atom_System::comment_time(double &time) {

  /*
    Attempt to determine system time from comment
  */


  // -- variables
  double t = 0.0;
  int found = -1;


  // --- report
  //cout << "Atom_System: Trying to determine system time from comment: \"" << comment << "\"" << endl;


  // --- try to find time

  // look for fs etc.
  if(found < 0) {
    int ind = 0;

    {
      istringstream temp(comment);
      string dummy;

      while(!temp.eof()) {
	ind++;
	temp >> dummy;  // read field
	std::transform(dummy.begin(), dummy.end(), dummy.begin(), (int(&)(int))std::tolower);  // to lowercase

	if(dummy=="fs" || dummy=="ps" || dummy=="ns") {
	  found = ind - 1;
	  break;
	}
      }
    }

    if(found > 0) {
      istringstream temp(comment);
      int ind = 0;
      string dummy;

      for(ind=1;ind<found;ind++) {
	temp >> dummy;  // read field
      }
      temp >> t;
    }
  }

  if(found < 0) {
    // implement other possible ways to read time in other sections
  }

  // --- set time

  if(found > 0) {
    //cout << "             found, setting time: " << t << endl;
    time = t;
    return true;
  }
  else {
    //cout << "             not found, setting time: " << 0.0 << endl;
    time = 0.0;
    return false;
  }
}

void Atom_System::add_atom(const Atom_System &other, int at) {

  /*
    Adds an atom at from another system to this system.

    Assumes that dim is consistent.

    Preserves atom type number if possible.
  */


  // --- variables
  int i;  // loops


  // --- resize arrays if needed
  if(atypesize<natoms+1) {
    moreatoms(1);
  }


  // --- add atom

  // atom data, except type
  for(i=0;i<dim;i++) {
    x0[dim*natoms + i] = other.x0[dim*at + i];
    x1[dim*natoms + i] = other.x1[dim*at + i];
    F[dim*natoms + i] = other.F[dim*at + i];
    fixx0[dim*natoms + i] = other.fixx0[dim*at + i];
    dx0[dim*natoms + i] = other.dx0[dim*at + i];
  }
  V[natoms] = other.V[at];
  K[natoms] = other.K[at];
  block[natoms] = other.block[at];

  // atom type, preserve type number if possible
  if(atypes.safeadd(other.atypes.atypename[other.atype[at]], other.atype[at])) {
    atypes.add(other.atypes.atypename[other.atype[at]], other.atype[at]);
    i = other.atype[at];
  }
  else {
    i = -1;
    atypes.add(other.atypes.atypename[other.atype[at]], i);
  }
  atype[natoms] = i;

  // --- natoms
  atypes.natype[atype[natoms]]++;
  natoms++;
}


void Atom_System::join(const Atom_System &other) {

  /*
    Take atoms from other system to this one.
  */


  // --- variables
  int i;  // loops


  // --- report
  cout << "Atom_System::join(): Joining a system of " << other.atoms() << " atoms to this one with " << natoms << " atoms..." << endl;


  // --- resize arrays
  moreatoms(other.atoms());


  // --- take atoms
  for(i=0;i<other.atoms();i++) {
    add_atom(other, i);
  }

  cout << "...done." << endl;
}


double Atom_System::extrcoord(int c, int d) {

  /*
    Get extremal coordinate in direction c.

    c in [0,dim-1].

    if d==0, minimum
    if d==1, maximum
  */


  // --- variables
  int i;  // loops
  double val;  // value found
  int at;  // which atom is this


  // --- check
  if(natoms < 1) {
    cerr << "Atom_System::extrcoord(): Error: No atoms in system." << endl;
    exit(-1);
  }
  if(c < 0 || c > dim-1) {
    cerr << "Atom_System::extrcoord(): Error: Direction = " << c << endl;
    exit(-1);
  }
  if(d != 0 && d != 1) {
    cerr << "Atom_System::extrcoord(): Error: 0 - min, 1 - max, now d = " << d << endl;
    exit(-1);
  }


  // --- report
  if(d==0) cout << "Atom_System::extrcoord(): Finding minimum in direction " << c << "...";
  else cout << "Atom_System::extrcoord(): Finding maximum in direction " << c << "...";

  // --- find minimum
  at = 0;
  val = x0[dim*at + c];
  for(i=1;i<natoms;i++) {
    if(d==0) {
      if(x0[dim*i + c] < val) {
	val = x0[dim*i + c];
	at = i;
      }
    }
    else {
      if(x0[dim*i + c] > val) {
	val = x0[dim*i + c];
	at = i;
      }
    }
  }


  // --- report
  if(d==0) cout << " ...done. min = " << val << " (atom " << at << ")." << endl;
  else cout << " ...done. max = " << val << " (atom " << at << ")." << endl;


  return val;
}


void Atom_System::addEk(double Eat) {

  /*
    Add kinetic energy of Eat per atom.

    Ignores fixed atoms. (But they should be taken into
    account properly since updateT() ignores them. This
    should be checked if used for anything serious,
    however.)
  */


  // --- variables
  int i;  // loops
  double E;  // E_k
  double f;  // v_i -> f v_i


  // --- E_k now
  updateT();

  E = 0;
  for(i=0;i<natoms;i++) {
    E += K[i];  // eV
  }
  E /= (double)natoms;


  // --- factor
  if(E + Eat < 0) {
    cerr << "Atom_System::addEk(): Error: Resulting kinetic energy negative: E, Eat: " << E << ", " << Eat << endl;
    exit(-1);
  }
  f = sqrt( (E + Eat)/E );


  // --- add E
  for(i=0;i<dim*natoms;i++) {
    x1[i] *= f;
  }


  // --- report
  cout << "Atom_System::addEk(): Added " << Eat << " eV/at. T [K]: " << T;
  updateT();
  cout << " -> " << T << ", E [eV/at.]: " << E;
  E = 0;
  for(i=0;i<natoms;i++) {
    E += K[i];  // eV
  }
  E /= (double)natoms;
  cout << " -> " << E << endl;
}


void Atom_System::strain(vector< valarray<double> > &eps) {

  /*
    Strain system according to matrix eps:

    a = (1 + eps) a,

    where a is a matrix of unit cell vectors.

    Work under the following assumptions:

          ( e11 e12 e13 )
    eps = ( e12 e22 e23 )
          ( e13 e23 e33 )

          ( a1 b1 c1 )   ( a1  0  0 )   ( a 0 0 )
    a   = ( a2 b2 c2 ) = (  0 b2  0 ) = ( 0 a 0 )
          ( a3 b2 c3 )   (  0  0 c3 )   ( 0 0 a )

    On input eps[0] = (e11, e12, e13) (first row, not column)
             eps[1] = (e12, e22, e23)
             eps[2] = (e13, e23, e33)

    (We check that it's symmetric so this doesn't matter.)

    eps GET'S REPLACED BY NEW BOX SIZE VECTORS, NOT UNIT CELL VECTORS!

    On input eps[0] = (a1, a2, a3) (first row, not column)
             eps[1] = (b1, b2, b3)
             eps[2] = (c1, c2, c3)
  */


  // --- variables

  int i, j, k;  // loops
  vector< valarray<double> > ap;  // new unit cell vectors, a = ap[0], b = ap[1], etc.
  valarray<double> ta;  // temp for old unit or position vector
  double voli, volf;  // volume of unit cell initially and after transformation


  // --- check

  // dim
  if(dim==0) {
    cerr << "Atom_System::strain: Error: dimension = " << dim << endl;
    exit(-1);
  }

  // eps size
  if(eps.size() != dim) {
    cerr << "Atom_System::strain: Error: Strain matrix rows != dimension: " << eps.size() << " " << dim << endl;
    exit(-1);
  }
  for(i=0;i<dim;i++) {
    if(eps[i].size() != dim) {
      cerr << "Atom_System::strain: Error: Width of strain matrix column " << i << " != dimension: " << eps[i].size() << " " << dim << endl;
      exit(-1);
    }
  }

  // that eps is symmetric
  for(i=0;i<dim;i++) {
    for(j=0;j<dim;j++) {
      if(i==j) continue;
      if(eps[i][j] != eps[j][i]) {
	cerr << "Atom_System::strain: Error: Strain matrix not symmetric: eps(" << i << "," << j << ") != eps(" << j << "," << i << "): " << eps[i][j] << " " << eps[j][i] << endl;
	exit(-1);
      }
    }
  }


  // --- add unity to epsilon
  for(i=0;i<dim;i++) {
    eps[i][i] += 1.0;
  }


  // --- construct new unit cell vectors

  ap.resize(dim);
  for(i=0;i<dim;i++) {
    ap[i].resize(dim);
  }

  ta.resize(dim);

  // loop over unit cell vectors
  for(i=0;i<dim;i++) {

    // old unit cell vector
    for(j=0;j<dim;j++) {
      if(j==i) ta[j] = a;
      else ta[j] = 0;
    }

    // new one
    for(j=0;j<dim;j++) {
      ap[i][j] = 0;

      for(k=0;k<dim;k++) {
	ap[i][j] += eps[j][k]*ta[k];
      }
    }

    // report
    cout << "Atom_System::strain(): Changed unit cell vector " << i << ": ";
    for(k=0;k<dim;k++) {
      cout << ta[k] << " ";
    }
    cout << "-> ";
    for(k=0;k<dim;k++) {
      cout << ap[i][k] << " ";
    }
    cout << endl;

  }  // end of loop over unit cell vectors


  // --- report volume change
  if(dim!=3) {
    cout << "Atom_System::strain(): Warning: Too lazy to calculate unit cell volume change for system since dimension = " << dim << endl;
  }
  else {
    voli = a*a*a;

    volf = ap[0][0]*(ap[1][1]*ap[2][2] - ap[1][2]*ap[2][1]);
    volf -= ap[0][1]*(ap[1][0]*ap[2][2] - ap[1][2]*ap[2][0]);
    volf += ap[2][2]*(ap[1][0]*ap[2][1] - ap[1][1]*ap[2][0]);
  }
  cout << "Atom_System::strain(): Unit cell volume change: " << voli << " -> " << volf << " (change: " << volf-voli << ")" << endl;


  // --- transform system

  // loop over atoms
  for(i=0;i<natoms;i++) {

    // position in basis of unit vectors
    for(j=0;j<dim;j++) {
      ta[j] = x0[dim*i + j]/a;
    }

    // loop over unit coordinates
    for(j=0;j<dim;j++) {
      x0[dim*i + j] = 0;

      // loop over unit vectors
      for(k=0;k<dim;k++) {
	x0[dim*i + j] += ap[k][j]*ta[k];
      }  // end of loop over unit vectors

    }  // end of loop over unit coordinates

  }  // end of loop over atoms


  // --- replace eps with new unit cell vectors

  // loop over unit cell vectors
  for(i=0;i<dim;i++) {

    // old unit cell vector
    for(j=0;j<dim;j++) {
      if(j==i) ta[j] = box[j];
      else ta[j] = 0;
    }

    // new one
    for(j=0;j<dim;j++) {
      ap[i][j] = 0;

      for(k=0;k<dim;k++) {
	ap[i][j] += eps[j][k]*ta[k];
      }
    }
  }

  // fill eps
  for(i=0;i<dim;i++) {
    for(j=0;j<dim;j++) {
      eps[i][j] = ap[i][j];
    }
  }
}
